using System;
using System.IO;
using System.Text.RegularExpressions;
using System.Diagnostics;

using Pen = System.Drawing.Pen;
using Color = System.Drawing.Color;
using Image = System.Drawing.Image;
using DotNetBitmap = System.Drawing.Bitmap;
using DotNetGraphics = System.Drawing.Graphics;

#if !MONO
using Microsoft.Win32;
#endif

namespace Espresso.OMR
{
    public enum CacheMode { UseCache, OverwriteCache };

    /// <summary>
    /// Contains various image conversion methods
    /// </summary>
    public class BitmapConvertUtils
    {
        private static string m_ImageMagickBinPath;

        /// <summary>
        /// Returns the path to the ImageMagick binary folder
        /// </summary>
        public static string ImageMagickBinPath
        {
            get {
                #if MONO
	                return "";
                #else
                    if (m_ImageMagickBinPath == null)
                    {
                        RegistryKey reg_key = Registry.LocalMachine.OpenSubKey(@"Software\ImageMagick\Current");
                        m_ImageMagickBinPath = (string)reg_key.GetValue("BinPath", false); 
                    }

                    return m_ImageMagickBinPath;
                #endif
            }
        }

        /// <summary>
        /// Executes a process and returns all its standard output.
        /// </summary>
        /// <param name="exec_filename">The name of the executable</param>
        /// <param name="arguments">Arguments to pass to the executable</param>
        /// <returns>Standard output of the process</returns>
        protected static string SyncExecuteProcess(string filename, string arguments)
        {
            Process p = new Process();
            p.StartInfo.UseShellExecute = false;
            p.StartInfo.CreateNoWindow = true;
            p.StartInfo.RedirectStandardOutput = true;
            p.StartInfo.FileName = filename;
            p.StartInfo.Arguments = arguments;

            p.Start();
            string output = p.StandardOutput.ReadToEnd();
            p.WaitForExit();

            if (p.ExitCode != 0)
            {
                string mess = "Warning: " + p.StartInfo.FileName + " " + p.StartInfo.Arguments
                                + " terminated with exit status " + p.ExitCode.ToString();
                Console.Error.WriteLine(mess);
                throw new Exception(mess);
            }

            return output;
        }

        /// <summary>
        /// Converts <see cref="sourceFileName"/> into the format specified by 
        /// <see cref="destinationFileName"/>'s extension with given options
        /// </summary>
        /// <param name="sourceFileName">Full path to the original file</param>
        /// <param name="destinationFileName">Full path to the destination file</param>
        /// <param name="options">ImageMagick options</param>
        public static void SyncConvert(string sourceFileName, string destinationFileName, string options)
        {
            string exec_filename = ImageMagickBinPath + "\\" + "convert.exe";
            string arguments = options + " \"" + sourceFileName + "\" \"" + destinationFileName + "\"";

            SyncExecuteProcess(exec_filename, arguments);
        }

        /// <summary>
        /// Converts <see cref="sourceFileName"/> into the format specified by 
        /// <see cref="destinationFileName"/>'s extension with given options
        /// </summary>
        /// <param name="sourceFileName">Full path to the original file</param>
        /// <param name="destinationFileName">Full path to the destination file</param>
        public static void SyncConvert(string sourceFileName, string destinationFileName)
        {
            SyncConvert(sourceFileName, destinationFileName, "");
        }

        /// <summary>
        /// Gets number of pages in a pdf document.
        /// </summary>
        /// <param name="pdf_filename">The path to the pdf file</param>
        /// <returns>Number of pages</returns>
        public static int PdfNumberOfPages(string pdfFilename)
        {
            string exec_filename = ImageMagickBinPath + "\\" + "identify.exe";
            string arguments = "-format \"%n\" \"" + pdfFilename + "\"";
            string output = SyncExecuteProcess(exec_filename, arguments);

            string[] output_lines = Regex.Split(output, Environment.NewLine);

            return Int32.Parse(output_lines[0]);
        }

        /// <summary>
        /// Converts a page of a pdf document into pbm.
        /// </summary>
        /// <param name="pdf_filename">Full path to the pdf file</param>
        /// <param name="page_num">Page number to convert</param>
        /// <param name="cache_mode">Cache mode. When set to UseCache, the pbm will not be created if the page
        ///                          has already been converted and stored in cache on hard drive.</param>
        /// <param name="resolution">Resolution in dpi</param>
        /// <returns>Full path to the converted pbm file</returns>
        public static string ConvertPdfPageToPbm(string pdf_filename, int page_num, CacheMode cache_mode, int resolution)
        {
            int ext_pos = pdf_filename.LastIndexOf('.');
            int slash_pos = pdf_filename.LastIndexOf(Path.DirectorySeparatorChar);

            string filename = pdf_filename.Substring(slash_pos + 1, ext_pos - slash_pos - 1);
            string root_filename = EspressoUtils.GetPathToStorageFolder(filename)+ filename;

            string p_str = page_num.ToString();

            string options = " -type bilevel -density " + resolution;
            string src_filename = pdf_filename + "[" + p_str + "]";
            string dst_filename = root_filename + p_str + ".pbm";

            if ((cache_mode == CacheMode.OverwriteCache) || ((cache_mode == CacheMode.UseCache) && (!File.Exists(dst_filename))))
            {
                BitmapConvertUtils.SyncConvert(src_filename, dst_filename, options);
            }

            return dst_filename;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="pdfFilepath"></param>
        /// <param name="pageNum"></param>
        /// <param name="cacheMode"></param>
        /// <param name="resolution"></param>
        /// <returns>Full path to the resulting png file</returns>
        public static string ConvertPdfPageToPng(string pdfFilepath, int pageNum, CacheMode cacheMode, int resolution)
        {
            int ext_pos = pdfFilepath.LastIndexOf('.');
            int slash_pos = pdfFilepath.LastIndexOf(Path.DirectorySeparatorChar);

            string filename = pdfFilepath.Substring(slash_pos + 1, ext_pos - slash_pos - 1);
            string root_filename = EspressoUtils.GetPathToStorageFolder(filename) + filename;

            string options = " -density " + resolution;
            string src_filename = pdfFilepath + "[" + pageNum + "]";
            string dst_filename = root_filename + pageNum + ".png";

            //TODO: fix this so it actually deletes the old file
            if (cacheMode == CacheMode.OverwriteCache) 
            {
                if (!File.Exists(dst_filename))
                {
                    BitmapConvertUtils.SyncConvert(src_filename, dst_filename, options);
                }

            }
            else if (cacheMode == CacheMode.UseCache)
            {
                //We are using the old cache. If it does not exist, then create it
                if (!File.Exists(dst_filename))
                {
                    BitmapConvertUtils.SyncConvert(src_filename, dst_filename, options);
                }
            }
            
            return dst_filename;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="bitmap"></param>
        /// <returns></returns>
        /*
        public static DotNetBitmap ConvertToDotNetBitmap(Bitmap bitmap)
        {
            int x_dim = bitmap.x_dim();
            int y_dim = bitmap.y_dim();

            int x_dim_rounded16 = (x_dim + 15) / 16 * 16;
            int y_dim_rounded16 = (y_dim + 15) / 16 * 16;

            DotNetBitmap dotnet_bitmap = new DotNetBitmap(x_dim_rounded16, y_dim_rounded16);


            for (int y = 0; y < y_dim; y++)
            {
                // *****
                // can implemement something faster here that directly processing the bits within a byte
                for (int x = 0; x < x_dim; x++)
                {
                    if (bitmap.pixel_is_set(x, y))
                    {
                        dotnet_bitmap.SetPixel(x, y, Color.Black);
                    }
                    else
                    {
                        //dotnet_bitmap.SetPixel(x, y, Color.White);
                    }

                }
            }

            return dotnet_bitmap;
        }
        public static DotNetBitmap ConvertToDotNetBitmapDrawLine(Bitmap bitmap)
        {
            int x_dim = bitmap.x_dim();
            int y_dim = bitmap.y_dim();

            // some pages about DotNet talk about resizing of images being faster if multiple of 16
            int x_dim_rounded16 = (x_dim + 15) / 16 * 16;
            int y_dim_rounded16 = (y_dim + 15) / 16 * 16;

            // background defaults to white
            DotNetBitmap dotnet_bitmap = new DotNetBitmap(x_dim_rounded16, y_dim_rounded16);
            DotNetGraphics dotnet_graphics = DotNetGraphics.FromImage((Image)dotnet_bitmap);
            Pen pen_line = new Pen(Color.Black);

            for (int y = 0; y < y_dim; y++)
            {
                int x = 0;
                while (x < x_dim)
                {
                    // skip over white pixels
                    bool run_start = false;
                    while (true)
                    {
                        if (bitmap.pixel_is_set(x, y))
                        {
                            run_start = true;
                            break;
                        }

                        x++;
                        if (x >= x_dim)
                        {
                            break;
                        }
                    }

                    if (run_start)
                    {
                        // determine length of run of black
                        int x_start = x;
                        while (true)
                        {
                            x++;
                            if (x >= x_dim)
                            {
                                break;
                            }

                            if (!bitmap.pixel_is_set(x, y))
                            {
                                break;
                            }
                        }
                        int x_end = x - 1;

                        dotnet_graphics.DrawLine(pen_line, x_start, y, x_end, y);

                    }
                }
            }

            return dotnet_bitmap;
        }
        
        public static Bitmap ConvertFromDotNetBitmap(DotNetBitmap dotnet_bitmap)
        {
            int x_dim = dotnet_bitmap.Width;
            int y_dim = dotnet_bitmap.Height;
            Bitmap bitmap = new Bitmap(x_dim, y_dim);

            for (int y = 0; y < y_dim; y++)
            {
                for (int x = 0; x < x_dim; x++)
                {
                    // *****
                    // can implemement something faster here that directly processing the bits within a byte
                    string pixel_colour = dotnet_bitmap.GetPixel(x, y).Name;
                    if (pixel_colour == "Black")
                    {
                        bitmap.set_pixel(x, y);
                    }
                    else if (pixel_colour == "White")
                    {
                        //clear_pixel(x, y);
                    }
                    else
                    {
                        Console.Error.WriteLine("Warning: unrecognised colour " + pixel_colour);
                    }
                }
            }

            return bitmap;
        }
         */
    }
}

