using System;
using System.IO;

namespace Espresso.OMR
{
	internal class SepApplication
	{
		public enum StageOptionEnum 
		{
			no_stage_opt = 0,
			skew_opt = 1 + 0,
			deform_opt = 1 + 1 };

		public enum StopAfterStageEnum 
		{
			stop_gcharacters,   stop_staff_systems,
			stop_staff_areas,   stop_staff_segs,
			stop_staff_lines,   stop_correct_skew,
			stop_remove_staffs, stop_correct_deform,
			stop_never };

		protected static bool   gen_without_sl_ = false;
		protected static bool   gen_only_sl_    = false;

		protected static int    tune_no_ = 0;
		protected static int    part_no_ = 0;

		protected static double title_ymul_   = 2.0;
		protected static int    title_thresh_ = 2;

		protected static bool   do_auto_bcrop_  = false;
		protected static int    auto_bcrop_     = 0;
		protected static double abcrop_ymul_    = 2.0;
		protected static int    abcrop_thresh_  = 4;

		protected static bool   do_bcrop_       = false;
		protected static int    bcrop_          = 0;
		protected static bool   do_rcrop_       = false;
		protected static int    rcrop_          = 0;
		protected static bool   do_lclear_      = false;
		protected static int    lclear_         = 0;

		protected static int    restrict_y_dim_ = 0;
		protected static int    restrict_x_dim_ = 0;

		protected PageManager page_manager_;
		protected MObjects mobjects_;
		protected SepGraphics graphics_;

		protected CpuUsage machine_usage_;

		protected ComputingCircle computing_perc_;
		protected VerboseExtra verbose_extra_;

		protected StageOptionEnum stage_options_;
		protected string input_filename_;
		protected StringList input_filenames_;
		protected string output_filename_;

		protected bool gcharacters_done_;
		protected bool staff_systems_done_;
		protected bool staff_areas_done_;
		protected bool staff_segs_done_;
		protected bool staff_lines_done_;
		protected bool correct_skew_done_;
		protected bool remove_staffs_done_;
		protected bool correct_deform_done_;

		protected MObjects.GcharactersModeEnum   gcharacters_mode_;
		protected MObjects.FloodFillModeEnum     staff_systems_mode_;
		protected MObjects.StaffAreasModeEnum    staff_areas_mode_;
		protected Staff.PassEnum                 staff_segs_pass_mode_;
		protected MObjects.StaffSegModeEnum      staff_segs_mode_;
		protected MObjects.StaffLinesModeEnum    staff_lines_mode_;
		protected MObjects.CorrectSkewModeEnum   correct_skew_mode_;
		protected MObjects.RemoveStaffsModeEnum  remove_staffs_mode_;
		protected MObjects.CorrectDeformModeEnum correct_deform_mode_;

		protected double cond_skew_val_;

		protected int sa_prop_numer_;
		protected int sa_prop_denom_;
		protected int sg_min_gap_;

		protected double seg_upt_mul_;
		protected double seg_lowt_strict_mul_;
		protected double seg_lowt_hyster_mul_;

		protected double sl_lines_vt_;
		protected double sl_gaps_vt_;
		protected int sl_lines_rt_;
		protected int sl_gaps_rt_;

		protected StopAfterStageEnum stop_after_stage_;
		protected bool do_view_;

		protected int print_level_;

		protected static int verbose_level_ = 0;


		public SepApplication(int argc, string[] argv)
		{
			// PageManager and BaseMObjects need mutual references to each other
			// Construct PageManager first, then MObjects (with ref to PageManager)
			// finally set reference to MObjects in PageManager.

			page_manager_  = new PageManager(argc, argv);
			mobjects_      = new MObjects(page_manager_);
			page_manager_.set_base_mobjects(mobjects_); 

			graphics_      = new SepGraphics(mobjects_.bitmap_ops().scratch());
			machine_usage_ = new CpuUsage();

			computing_perc_  = null;
			verbose_extra_   = new VerboseExtra();
			stage_options_   = StageOptionEnum.no_stage_opt; 
			input_filename_  = "";
			input_filenames_ = new StringList(); 
			output_filename_ = "";

			// defaults

			FilenameManager.set_file_access("local"); 

			gcharacters_mode_ = MObjects.GcharactersModeEnum.gc_ignore;
			staff_systems_mode_ = MObjects.FloodFillModeEnum.flood_fill_basic; 
			staff_areas_mode_ = MObjects.StaffAreasModeEnum.sa_locate_4_5; 
			staff_segs_pass_mode_ = Staff.PassEnum.first_pass; 
			staff_segs_mode_ = MObjects.StaffSegModeEnum.seg_locate_1_3_4; 
			staff_lines_mode_ = MObjects.StaffLinesModeEnum.sl_locate_average; 
			correct_skew_mode_ = MObjects.CorrectSkewModeEnum.rotate_page; 
			cond_skew_val_ = 0.0;
			remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.remove_straight_clarke;
			correct_deform_mode_ = MObjects.CorrectDeformModeEnum.deform_level_60;

			gcharacters_done_ = false; 
			staff_systems_done_ = false; 
			staff_areas_done_ = false; 
			staff_segs_done_ = false; 
			staff_lines_done_ = false; 
			correct_skew_done_ = false; 
			remove_staffs_done_ = false; 
			correct_deform_done_ = false;

			sa_prop_numer_ = 0; 
			sa_prop_denom_ = 1; 
			sg_min_gap_ = 0;
			seg_upt_mul_ = 1.0; 
			seg_lowt_strict_mul_ = 1.0/3.0; 
			seg_lowt_hyster_mul_ = 1.0/4.0;
			sl_lines_vt_ = 0.0; 
			sl_gaps_vt_ = 0.0;
			sl_lines_rt_ = 0; 
			sl_gaps_rt_ = 0;
			stop_after_stage_ = StopAfterStageEnum.stop_never; 
			do_view_ = false; 
			print_level_ = 0;


			computing_perc_ = new ComputingCircle();

			int no_of_args = argc;
			int c=1;

			StopAfterStageEnum potential_stop = StopAfterStageEnum.stop_never;

			while (c<no_of_args)
			{
				if (argv[c].StartsWith("-"))
				{
					string minus_word = argv[c];
					if (minus_word.Equals("-graphical_char_mode"))
					{
						potential_stop = StopAfterStageEnum.stop_gcharacters;

						string word = argv[++c];
						if (word.Equals("ignore"))
							gcharacters_mode_ = MObjects.GcharactersModeEnum.gc_ignore;
						else if (word.Equals("ocr"))
							gcharacters_mode_ = MObjects.GcharactersModeEnum.gc_ocr;
						else
						{
							Console.Error.WriteLine("Unrecognised graphical chars mode: " + word );
						}
					}
					else if (minus_word.Equals("-staff_system_mode"))
					{
						potential_stop = StopAfterStageEnum.stop_staff_systems;

						string word = argv[++c];
						if (word.Equals("normal"))
							staff_systems_mode_ = MObjects.FloodFillModeEnum.flood_fill_basic;
						else if (word.Equals("2bit"))
							staff_systems_mode_ = MObjects.FloodFillModeEnum.flood_fill_2bit;
						else if (word.Equals("byte"))
							staff_systems_mode_ = MObjects.FloodFillModeEnum.flood_fill_byte;
						else if (word.Equals("trace"))
							staff_systems_mode_ = MObjects.FloodFillModeEnum.contour_trace;
						else
						{
							Console.Error.WriteLine("Unrecognised staff system mode: " + word );
						}
					}
					else if (minus_word.Equals("-staff_area_mode"))
					{
						potential_stop = StopAfterStageEnum.stop_staff_areas;

						string word = argv[++c];

						if (word.Equals("sa_prop_numer"))
						{
							string num_str = argv[++c];
							sa_prop_numer_ = Int32.Parse(num_str);
						}
						else if (word.Equals("sa_prop_denom"))
						{
							string num_str = argv[++c];
							sa_prop_denom_ = Int32.Parse(num_str);
						}
						else if (word.Equals("sg_min_gap"))
						{
							string num_str = argv[++c];
							sg_min_gap_ = Int32.Parse(num_str);
						}
						else if (word.Equals("sa_locate_4_5"))
						{
							staff_areas_mode_ = MObjects.StaffAreasModeEnum.sa_locate_4_5;
							sa_prop_numer_ = 4;
							sa_prop_denom_ = 5;
						}
						else if (word.Equals("sa_locate_3_4"))
						{
							staff_areas_mode_ = MObjects.StaffAreasModeEnum.sa_locate_3_4;
							sa_prop_numer_ = 3;
							sa_prop_denom_ = 4;
						}
						else if (word.Equals("sa_locate_6_10"))
						{
							staff_areas_mode_ = MObjects.StaffAreasModeEnum.sa_locate_3_4;
							sa_prop_numer_ = 6;
							sa_prop_denom_ = 10;
						}
						else if (word.Equals("sg_min_gap_4"))
						{
							staff_areas_mode_ = MObjects.StaffAreasModeEnum.sg_min_gap_4;
							sg_min_gap_ = 4;
						}
						else if (word.Equals("sg_min_gap_6"))
						{
							staff_areas_mode_ = MObjects.StaffAreasModeEnum.sg_min_gap_6;
							sg_min_gap_ = 6;
						}
						else if (word.Equals("sg_min_gap_12"))
						{
							staff_areas_mode_ = MObjects.StaffAreasModeEnum.sg_min_gap_12;
							sg_min_gap_ = 12;
						}
						else if (word.Equals("sg_min_gap_30"))
						{
							staff_areas_mode_ = MObjects.StaffAreasModeEnum.sg_min_gap_30;
							sg_min_gap_ = 30;
						}
						else
						{
							Console.Error.WriteLine("Unrecognised staff area mode: " + word );
						}
					}


					else if (minus_word.Equals("-staff_seg_mode"))
					{
						potential_stop = StopAfterStageEnum.stop_staff_segs;

						string word = argv[++c];

						if (word.Equals("seg_upt_mul"))
						{
							string num_str = argv[++c];
							seg_upt_mul_ = Double.Parse(num_str);
						}
						else if (word.Equals("seg_lowt_strict_mul"))
						{
							string num_str = argv[++c];
							seg_lowt_strict_mul_ = Double.Parse(num_str);
						}
						else if (word.Equals("seg_lowt_hyster_mul"))
						{
							string num_str = argv[++c];
							seg_lowt_hyster_mul_ = Double.Parse(num_str);
						}
						else if (word.Equals("seg_1_2_3"))
						{
							staff_segs_mode_ = MObjects.StaffSegModeEnum.seg_locate_1_2_3;
							seg_upt_mul_ = 1.0;
							seg_lowt_strict_mul_ = 1.0/2.0;
							seg_lowt_hyster_mul_ = 1.0/3.0;
						}
						else if (word.Equals("seg_1_3_4"))
						{
							staff_segs_mode_ = MObjects.StaffSegModeEnum.seg_locate_1_3_4;
							seg_upt_mul_ = 1.0;
							seg_lowt_strict_mul_ = 1.0/3.0;
							seg_lowt_hyster_mul_ = 1.0/4.0;
						}
						else if (word.Equals("seg_1_4_5"))
						{
							staff_segs_mode_ = MObjects.StaffSegModeEnum.seg_locate_1_4_5;
							seg_upt_mul_ = 1.0;
							seg_lowt_strict_mul_ = 1.0/4.0;
							seg_lowt_hyster_mul_ = 1.0/5.0;
						}
						else if (word.Equals("seg_2_0_0"))
						{
							staff_segs_mode_ = MObjects.StaffSegModeEnum.seg_locate_2_0_0;
							seg_upt_mul_ = 2.0;
							seg_lowt_strict_mul_ = 0.0;
							seg_lowt_hyster_mul_ = 0.0;
						}
						else
						{
							Console.Error.WriteLine("Unrecognised staff seg mode: " + word );
						}
					}
					else if (minus_word.Equals("-staff_seg_pass_mode"))
					{
						potential_stop = StopAfterStageEnum.stop_staff_segs;

						string word = argv[++c];

						if (word.Equals("second_pass"))
							staff_segs_pass_mode_ = Staff.PassEnum.second_pass;
					}
					else if (minus_word.Equals("-staff_line_mode"))
					{
						potential_stop = StopAfterStageEnum.stop_staff_lines;

						string word = argv[++c];

						if (word.Equals("sl_lines_vt"))
						{
							string num_str = argv[++c];
							sl_lines_vt_ = Double.Parse(num_str);
						}
						else if (word.Equals("sl_gaps_vt"))
						{
							string num_str = argv[++c];
							sl_gaps_vt_ = Double.Parse(num_str);
						}
						else if (word.Equals("sl_lines_rt"))
						{
							string num_str = argv[++c];
							sl_lines_rt_ = Int32.Parse(num_str);
						}
						else if (word.Equals("sl_gaps_rt"))
						{
							string num_str = argv[++c];
							sl_gaps_rt_ = Int32.Parse(num_str);
						}
						else if (word.Equals("sl_strict"))
						{
							staff_lines_mode_ = MObjects.StaffLinesModeEnum.sl_locate_strict;
							sl_lines_vt_ = 1.0;
							sl_gaps_vt_  = 3.0;
							sl_lines_rt_ = 3;
							sl_gaps_rt_  = 5;
						}
						else if (word.Equals("sl_average"))
						{
							staff_lines_mode_ = MObjects.StaffLinesModeEnum.sl_locate_average;
							sl_lines_vt_ = 1.5;
							sl_gaps_vt_  = 3.5;
							sl_lines_rt_ = 4;
							sl_gaps_rt_  = 7;
						}
						else if (word.Equals("sl_loose"))
						{
							staff_lines_mode_ = MObjects.StaffLinesModeEnum.sl_locate_loose;
							sl_lines_vt_ = 2.0;
							sl_gaps_vt_  = 4.5;
							sl_lines_rt_ = 6;
							sl_gaps_rt_  = 9;
						}
						else
						{
							Console.Error.WriteLine("Unrecognised staff line mode: " + word );
						}
					}
					else if (minus_word.Equals("-correct_skew_mode"))
					{
						potential_stop = StopAfterStageEnum.stop_correct_skew;

						string word = argv[++c];
						if (word.Equals("rotate_pg"))
							correct_skew_mode_ = MObjects.CorrectSkewModeEnum.rotate_page;
						else if (word.Equals("shear_pg"))
							correct_skew_mode_ = MObjects.CorrectSkewModeEnum.shear_page;
						else if (word.Equals("shear_pg_byte"))
							correct_skew_mode_ = MObjects.CorrectSkewModeEnum.shear_page8;
						else if (word.Equals("rotate_ob"))
							correct_skew_mode_ = MObjects.CorrectSkewModeEnum.rotate_systems;
						else if (word.Equals("shear_ob"))
							correct_skew_mode_ = MObjects.CorrectSkewModeEnum.shear_systems;
						else if (word.Equals("shear_ob_byte"))
							correct_skew_mode_ = MObjects.CorrectSkewModeEnum.shear_systems8;
						else if (word.Equals("rotate_ln"))
							correct_skew_mode_ = MObjects.CorrectSkewModeEnum.rotate_lines;
						else if (word.Equals("shear_ln"))
							correct_skew_mode_ = MObjects.CorrectSkewModeEnum.shear_lines;
						else if (word.Equals("shear_ln_byte"))
							correct_skew_mode_ = MObjects.CorrectSkewModeEnum.shear_lines8;
						else
						{
							Console.Error.WriteLine("Unrecognised correct skew mode: " + word );
						}	      
					}
					else if (minus_word.Equals("-remove_staffs_mode"))
					{
						potential_stop = StopAfterStageEnum.stop_remove_staffs;

						string word = argv[++c];
						if (word.Equals("remove_straight_prerau"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.remove_straight_prerau;
						else if (word.Equals("remove_straight_clarke"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.remove_straight_clarke;
						else if (word.Equals("remove_wobble"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.remove_wobble;
						else if (word.Equals("remove_track"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.remove_track;
						else if (word.Equals("remove_track_wide"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.remove_track_wide;
						else if (word.Equals("remove_wobble_simple"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.remove_wobble_simple;
						else if (word.Equals("remove_track_simple"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.remove_track_simple;
						else if (word.Equals("remove_wobble_prerau"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.remove_wobble_prerau;
						else if (word.Equals("remove_track_prerau"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.remove_track_prerau;
						else if (word.Equals("remove_wobble_circle"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.remove_wobble_circle;
						else if (word.Equals("remove_track_circle"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.remove_track_circle;
						else if (word.Equals("cross_straight"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.cross_straight;
						else if (word.Equals("cross_wobble"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.cross_wobble;
						else if (word.Equals("cross_track"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.cross_track;
						else if (word.Equals("cross_wobble_circle"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.cross_wobble_circle;
						else if (word.Equals("cross_track_circle"))
							remove_staffs_mode_ = MObjects.RemoveStaffsModeEnum.cross_track_circle;
						else
						{
							Console.Error.WriteLine("Unrecognised correct skew mode: " + word );
						}	     
					}
					else if (minus_word.Equals("-skew_opt"))
					{
						potential_stop = StopAfterStageEnum.stop_correct_skew;

						stage_options_ = (StageOptionEnum)(stage_options_ | StageOptionEnum.skew_opt);
					}
					else if (minus_word.Equals("-conditional_skew"))
					{
						potential_stop = StopAfterStageEnum.stop_correct_skew;

						stage_options_ = (StageOptionEnum)(stage_options_ | StageOptionEnum.skew_opt);
						cond_skew_val_ = Double.Parse(argv[++c]);

					}
					else if (minus_word.Equals("-deform"))
					{
						potential_stop = StopAfterStageEnum.stop_correct_deform;

						string word = argv[++c];
						if (word.Equals("level_60"))
							correct_deform_mode_ = MObjects.CorrectDeformModeEnum.deform_level_60;
						if (word.Equals("level_70"))
							correct_deform_mode_ = MObjects.CorrectDeformModeEnum.deform_level_70;
						if (word.Equals("level_80"))
							correct_deform_mode_ = MObjects.CorrectDeformModeEnum.deform_level_80;

						if (word.Equals("angle_60"))
							correct_deform_mode_ = MObjects.CorrectDeformModeEnum.deform_angle_60;
						if (word.Equals("angle_70"))
							correct_deform_mode_ = MObjects.CorrectDeformModeEnum.deform_angle_70;
						if (word.Equals("angle_80"))
							correct_deform_mode_ = MObjects.CorrectDeformModeEnum.deform_angle_80;
						else
						{
							Console.Error.WriteLine("Unrecognised correct skew mode: " + word );
							Console.Error.WriteLine("Using default mode: level_60");
						}	     
						stage_options_ = (StageOptionEnum)(stage_options_ | StageOptionEnum.deform_opt);
					}
					else if (minus_word.Equals("-input_filename"))
					{
						c++;
						if (c<argc)
						{
							input_filename_ = argv[c];
							string ifname_copy = argv[c];
							input_filenames_.append(ifname_copy); // respon passed over

							while (c+1<argc)
							{
								if (argv[c+1][0]=='-')
								{
									break;
								}
								string ifn_copy = argv[c+1];
								input_filenames_.append(ifn_copy); // respon passed over

								c++;
							}
						}

					}
					else if (minus_word.Equals("-output_filename"))
					{
						output_filename_ = argv[++c];	      
					}
					else if (minus_word.Equals("-print"))
					{
						print_level_ = Int32.Parse(argv[++c]);
					}
					else if (minus_word.Equals("-stop"))
					{
						stop_after_stage_ = potential_stop;
					}
					else if (minus_word.Equals("-view"))
					{
						do_view_ = true;
					}
					else print_usage(argv[0]);
				}
				else print_usage(argv[0]);
				c++;
			}

			if (print_level_>0)
				Console.Error.WriteLine("Loading");

			if (input_filenames_.length()==1)
			{
				page_manager_.replace_page(input_filename_);
			}
			else
			{
				page_manager_.concat_pages(input_filenames_);      
			}

			FileBitmap current_fbitmap = page_manager_.current_fbitmap();
			if (!current_fbitmap.valid())
			{
				Console.Error.WriteLine("Unable to read file \"" + current_fbitmap.filename() + "\"");
				Environment.Exit(1);
			}

			if (FilenameManager.file_type()==FilenameManager.FileTypeEnum.sep)
				reset_stages_done_sep();

			if (MObjects.definitive_convert())
			{
				reset_stages_done_convert_dsep();
			}

		}


		protected void print_usage(string program_name)
		{
			Console.Error.WriteLine("Usage: ");
			Console.Error.WriteLine();
				  
			string match_prog_name = program_name;
			int match_pn_len = match_prog_name.Length;

			if (match_prog_name.EndsWith("septopbm"))
			{
				Console.Error.WriteLine("septopbm");
				Console.Error.WriteLine("\t [-remote|-local] ");
				Console.Error.WriteLine("\t [-gen_without_sl|-gen_only_sl]");
				Console.Error.WriteLine("\t [-input filename] [-output filename]");
				Console.Error.WriteLine("\t [-print number]");
			}
			else if (match_prog_name.EndsWith("sepsplit"))
			{
				Console.Error.WriteLine("sepsplit");
				Console.Error.WriteLine("\t [-remote|-local] ");
				Console.Error.WriteLine("\t [-title_ymul float-num] [-title_thresh number]");
				Console.Error.WriteLine("\t [-tpinit num:num] [-print number]");
				Console.Error.WriteLine("\t [-auto_bcrop number] "
					+      "[-abcrop_ymul number] [-abcrop_thresh number]");
				Console.Error.WriteLine("\t [-bcrop number] [-rcrop number] [-lclear number]");
				Console.Error.WriteLine("\t [-shrinkwrap_x_dim number] [-shrinkwrap_y_dim number]");
				Console.Error.WriteLine("\t [-input filename] [-output filename]");
			}
			else if (match_prog_name.EndsWith("sepcat"))
			{
				Console.Error.WriteLine("sepcat");
				Console.Error.WriteLine("\t [-remote|-local] ");
				Console.Error.WriteLine("\t [-input filenames] [-output filename]");
			}
			else
			{
				Console.Error.WriteLine("sep");
				Console.Error.WriteLine("\t [-remote|-local] [-def_comp] "
					+ "[-remove_def_text] [-single_staff_lines] ");
				Console.Error.WriteLine("\t[-graphical_char_mode ignore|ocr]");
				Console.Error.WriteLine("\t[-staff_system_mode normal|2bit|byte|trace]");
				Console.Error.WriteLine("\t[-staff_area_mode sa_locate_4_5|sa_locate_3_4"
					+    "|sg_min_gap_{4,6,12,30}]");
				Console.Error.WriteLine("\t[-staff_seg_pass_mode first_pass|second_pass]");
				Console.Error.WriteLine("\t[-staff_seg_mode "
					+ "seg_1_2_3|seg_1_3_4|seg_1_4_5|seg_2_0_0]");
				Console.Error.WriteLine("\t[-staff_line_mode sl_strict|sl_average|sl_loose]");
				Console.Error.WriteLine("\t[-correct_skew_mode rotate_pg|shear_pg|shear_pg_byte|"
					+    "rotate_ob|shear_ob|shear_ob_byte|"
					+    "rotate_ln|shear_ln|shear_ln_byte]");
				Console.Error.WriteLine("\t[-remove_staffs_mode "
					+       "remove_straight_prerau|remove_straight_clarke|"
					+       "remove_wobble|remove_track|remove_track_wide");

				Console.Error.WriteLine("       remove_wobble_simple|remove_track_simple|"
					+       "remove_wobble_prerau|remove_track_prerau|"
					+       "remove_wobble_circle|remove_track_circle|");
				Console.Error.WriteLine("\tcross_straight|cross_wobble|cross_track|"
					+       "cross_wobble_circle|cross_track_circle]");
				Console.Error.WriteLine("\t[-skew_opt]");
				Console.Error.WriteLine("\t[-conditional_skew degrees]");
				Console.Error.WriteLine("\t[-correct_deform level_60|level_70|level_80"
					+ "|angle_60|angle_70|angle_80]");
				Console.Error.WriteLine("\t[-input filename] [-output filename]");
				Console.Error.WriteLine("\t[-stop] [-view] [-print number]");
				Console.Error.WriteLine("");
			}


			Environment.Exit(1);
		}

		public static void set_gen_without_sl(bool gen_without_sl)
		{
			gen_without_sl_ = gen_without_sl;
			gen_only_sl_ = false;
		}

		public static void set_gen_only_sl(bool gen_only_sl)
		{
			gen_only_sl_ = gen_only_sl;
			gen_without_sl_ = false;
		}

		public static void set_tune_part_init(int tune_no, int part_no)
		{
			tune_no_ = tune_no;
			part_no_ = part_no;
		}

		public static void set_title_ymul(double title_ymul)
		{
			title_ymul_ = title_ymul;
		}

		void set_title_thresh(int title_thresh)
		{
			title_thresh_ = title_thresh;
		}

		public static void set_restrict_y_dim(int restrict_y_dim)
		{
			restrict_y_dim_ = restrict_y_dim;
			do_bcrop_ = true;
		}


		public static void set_auto_bcrop(int auto_bcrop)
		{
			double scan_dpi = (double)ShareGlobals.scan_dpi_;
			double auto_bcrop_float = (double)auto_bcrop;
			auto_bcrop_ = MathExtra.irint(auto_bcrop_float*(scan_dpi/300.0));
			do_auto_bcrop_ = true;
		}

		public static bool do_auto_bcrop()
		{
			return do_auto_bcrop_;
		}

		public static void set_abcrop_ymul(double abcrop_ymul)
		{
			abcrop_ymul_ = abcrop_ymul;
		}

		public static void set_abcrop_thresh(int abcrop_thresh)
		{
			abcrop_thresh_ = abcrop_thresh;
		}


		public static void set_bcrop(int bcrop)
		{
			double scan_dpi = (double)ShareGlobals.scan_dpi_;
			double bcrop_float = (double)bcrop;
			bcrop_ = MathExtra.irint(bcrop_float*(scan_dpi/300.0));
			do_bcrop_ = true;
		}

		public static bool do_bcrop()
		{
			return do_bcrop_;
		}

		public static void set_restrict_x_dim(int restrict_x_dim)
		{
			restrict_x_dim_ = restrict_x_dim;
			do_rcrop_ = true;
		}

		public static void set_rcrop(int rcrop)
		{
			double scan_dpi = (double)ShareGlobals.scan_dpi_;
			double rcrop_float = (double)rcrop;
			rcrop_ = MathExtra.irint(rcrop_float*(scan_dpi/300.0));
			do_rcrop_ = true;
		}

		public static bool do_rcrop()
		{
			return do_rcrop_;
		}

		public static void set_lclear(int lclear)
		{
			double scan_dpi = (double)ShareGlobals.scan_dpi_;
			double lclear_float = (double)lclear;
			lclear_ = MathExtra.irint(lclear_float*(scan_dpi/300.0));
			do_lclear_ = true;
		}

		public static bool do_lclear()
		{
			return do_lclear_;
		}


		protected void reset_stages_done_sep()
		{
			staff_systems_done_ = true;
			staff_areas_done_   = true;
			staff_segs_done_    = true;
			staff_lines_done_   = true;
			correct_skew_done_  = false;
			remove_staffs_done_ = true;
			correct_deform_done_= false;
		}

		protected void reset_stages_done_convert_dsep()
		{
			staff_systems_done_ = true;
			staff_areas_done_   = true;
			staff_segs_done_    = true;
			staff_lines_done_   = false;
			correct_skew_done_  = false;
			remove_staffs_done_ = false;
			correct_deform_done_= false; // check ***
		}

		/*--------------------*/


		protected void def_comp_gcharacters()
		{}

		protected void def_comp_staff_systems()
		{
			StaffSystemList dssl = mobjects_.def_staff_systems();
			StaffSystemList ssl = mobjects_.staff_systems();
  
			Console.Out.WriteLine("Number of definitive staff systems = " + dssl.length());

			int no_missed_dss = 0;

			StaffSystemNode current_dss = dssl.head();

			while (current_dss!=null)
			{
				bool found_match = false;

				StaffSystem dss = current_dss.val();
				GObject dss_gobject = dss.gobject();

				StaffSystemNode current_ss = ssl.head();
      
				while (current_ss!=null)
				{
					StaffSystem ss = current_ss.val();
					GObject ss_gobject = ss.gobject();

					if (dss_gobject.rect_intersect(ss_gobject))
					{
						// compute overlap;
						double rt_overlap 
							= (double)dss_gobject.rect_overlap(ss_gobject);
						double dss_gobject_area 
							= (double)(dss_gobject.x_dim() * dss_gobject.y_dim());

						double overlap_percent 
							= (100.0 * rt_overlap)/dss_gobject_area;

						if (overlap_percent>=80.0)
						{
							found_match = true;
							break;
						}
					}

					current_ss = ssl.next();
				}

				if (!found_match)
				{
					no_missed_dss++;
				}

				current_dss = dssl.next();
			}

			Console.Out.WriteLine("Number of missed definitive staff systems = " 
				+ no_missed_dss);

			Console.Out.WriteLine("Number of extra staff systems    = " 
				+ (ssl.length() + no_missed_dss - dssl.length()));

		}

		protected void def_comp_staff_areas()
		{
			// derived staffs in mobjects.potential_staffs();

			StaffSystemList dssl = mobjects_.def_staff_systems();
			StaffSystemList ssl = mobjects_.staff_systems();
  
			int no_def_staffs = mobjects_.no_staffs(dssl);
			int no_staffs = mobjects_.no_potential_staffs(ssl);

			Console.Out.WriteLine("Number of definitive staff areas = " + no_def_staffs);

			int no_missed_dst = 0;

			StaffSystemNode current_dss = dssl.head();

			while (current_dss!=null)
			{
				StaffSystem dss = current_dss.val();

				StaffList dstl = dss.staffs();
				StaffNode current_dst = dstl.head();

				while (current_dst!=null)
				{
					bool found_match = false;

					Staff dst = current_dst.val();
					GObject dst_gobject = dst.staff_area();

					StaffSystemNode current_ss = ssl.head();
      
					while (current_ss!=null)
					{
						StaffSystem ss = current_ss.val();

						StaffList stl = ss.potential_staffs();
						StaffNode current_st = stl.head();

						while (current_st!=null)
						{
							Staff st = current_st.val();
							GObject st_gobject = st.staff_area();

							if (dst_gobject.rect_intersect(st_gobject))
							{
								// compute overlap;
								double rt_overlap 
									= (double)dst_gobject.rect_overlap(st_gobject);

								double dst_gobject_area 
									= (double)(dst_gobject.x_dim() * dst_gobject.y_dim());
								double doverlap_percent 
									= (100.0 * rt_overlap)/dst_gobject_area;

								double st_gobject_area 
									= (double)(st_gobject.x_dim() * st_gobject.y_dim());
								double overlap_percent 
									= (100.0 * rt_overlap)/st_gobject_area;

								if ((doverlap_percent>=60.0) && (overlap_percent>=60.0))
								{
									found_match = true;
									break;
								}
							}

							current_st = stl.next();
						}

						if (found_match) break;
						current_ss = ssl.next();
					}

					if (!found_match)
					{
						no_missed_dst++;
					}

					current_dst = dstl.next();
				}

				current_dss = dssl.next();
			}

			Console.Out.WriteLine("Number of missed definitive staff areas = " 
				+ no_missed_dst);

			Console.Out.WriteLine("Number of extra staff areas    = " 
				+ (no_staffs + no_missed_dst - no_def_staffs));

		}

		protected int def_comp_staff_seg_differs(Bitmap def_staff_lines,
			GObject dsa_gobject,
			StaffSeg seg)
		{
			int gxl = seg.global_x_start();
			int gxr = seg.global_x_end();
			int lxl = seg.x_start();
			int lxr = seg.x_end();

			int gyt = dsa_gobject.y_top();
			int gyb = dsa_gobject.y_bot();

			int diff_count = 0;

			for (int gy=gyt,ly=0; gy<=gyb; gy++,ly++)
				for (int gx=gxl,lx=lxl; gx<=gxr; gx++,lx++)
				{
					if (dsa_gobject[ly,lx]!=def_staff_lines[gy,gx])
					{
						diff_count++;
					}
				}

			return diff_count;
		}

		protected void def_comp_staff_segs_overlap(Bitmap def_st_lines,
			Staff dst, 
			Staff st)
		{
			GObject dsa_gobject = dst.staff_area();

			StaffSeg lseg = st.left_staff_seg();

			int ldiff_pixels 
				= def_comp_staff_seg_differs(def_st_lines,dsa_gobject,lseg);
			Console.Out.WriteLine("Left staff seg differs = " + ldiff_pixels);

			StaffSeg rseg = st.right_staff_seg();

			int rdiff_pixels 
				= def_comp_staff_seg_differs(def_st_lines,dsa_gobject,rseg);
			Console.Out.WriteLine("Right staff seg differs = "+ rdiff_pixels);
		}


		protected void def_comp_staff_segs()
		{
			// derived staffs in mobjects.staffs();

			Bitmap def_staff_lines = mobjects_.inlay_def_staff_lines();

			StaffSystemList dssl = mobjects_.def_staff_systems();
			StaffSystemList ssl = mobjects_.staff_systems();
  
			int no_def_staffs = mobjects_.no_staffs(dssl);
			int no_staffs = mobjects_.no_staffs(ssl);

			Console.Out.WriteLine("Number of definitive staff seg areas = " + no_def_staffs);

			int no_missed_dst = 0;

			StaffSystemNode current_dss = dssl.head();

			while (current_dss!=null)
			{
				StaffSystem dss = current_dss.val();

				StaffList dstl = dss.staffs();
				StaffNode current_dst = dstl.head();

				while (current_dst!=null)
				{
					bool found_match = false;

					Staff dst = current_dst.val();
					GObject dst_gobject = dst.staff_area();

					StaffSystemNode current_ss = ssl.head();
      
					while (current_ss!=null)
					{
						StaffSystem ss = current_ss.val();

						StaffList stl = ss.staffs();
						StaffNode current_st = stl.head();

						while (current_st!=null)
						{
							Staff st = current_st.val();
							GObject st_gobject = st.staff_area();

							if (dst_gobject.rect_intersect(st_gobject))
							{
								// compute overlap;
								double rt_overlap 
									= (double)dst_gobject.rect_overlap(st_gobject);

								double dst_gobject_area 
									= (double)(dst_gobject.x_dim() * dst_gobject.y_dim());
								double doverlap_percent 
									= (100.0 * rt_overlap)/dst_gobject_area;

								double st_gobject_area 
									= (double)(st_gobject.x_dim() * st_gobject.y_dim());
								double overlap_percent 
									= (100.0 * rt_overlap)/st_gobject_area;

								if ((doverlap_percent>=60.0) && (overlap_percent>=60.0))
								{
									def_comp_staff_segs_overlap(def_staff_lines,dst,st);
									found_match = true;
									break;
								}
							}

							current_st = stl.next();
						}

						if (found_match) break;
						current_ss = ssl.next();
					}

					if (!found_match)
					{
						no_missed_dst++;
					}

					current_dst = dstl.next();
				}

				current_dss = dssl.next();
			}

			Console.Out.WriteLine("Number of missed definitive staff seg areas = " 
				+ no_missed_dst);

			Console.Out.WriteLine("Number of extra staff seg areas    = " 
				+ (no_staffs + no_missed_dst - no_def_staffs));


			def_staff_lines.clear_all();
		}

		protected void def_comp_staff_lines()
		{
			// derived staffs in mobjects.staffs();
			StaffSystemList dssl = mobjects_.def_staff_systems();
			StaffSystemList ssl = mobjects_.staff_systems();
  
			int no_def_staffs = mobjects_.no_staffs(dssl);
			int no_staffs = mobjects_.no_staffs(ssl);

			Console.Out.WriteLine("Number of definitive staff line areas = " + no_def_staffs);

			int no_missed_dst = 0;

			StaffSystemNode current_dss = dssl.head();

			while (current_dss!=null)
			{
				StaffSystem dss = current_dss.val();

				StaffList dstl = dss.staffs();
				StaffNode current_dst = dstl.head();

				while (current_dst!=null)
				{
					bool found_match = false;

					Staff dst = current_dst.val();
					GObject dst_gobject = dst.staff_area();

					StaffSystemNode current_ss = ssl.head();
      
					while (current_ss!=null)
					{
						StaffSystem ss = current_ss.val();

						StaffList stl = ss.staffs();
						StaffNode current_st = stl.head();

						while (current_st!=null)
						{
							Staff st = current_st.val();
							GObject st_gobject = st.staff_area();

							if (dst_gobject.rect_intersect(st_gobject))
							{
								// compute overlap;
								double rt_overlap 
									= (double)dst_gobject.rect_overlap(st_gobject);

								double dst_gobject_area 
									= (double)(dst_gobject.x_dim() * dst_gobject.y_dim());
								double doverlap_percent 
									= (100.0 * rt_overlap)/dst_gobject_area;

								double st_gobject_area 
									= (double)(st_gobject.x_dim() * st_gobject.y_dim());
								double overlap_percent 
									= (100.0 * rt_overlap)/st_gobject_area;

								if ((doverlap_percent>=60.0) && (overlap_percent>=60.0))
								{
									found_match = true;

									Console.Out.WriteLine("Number of definitive staff lines = " 
										+ dst.no_staff_lines());
									Console.Out.WriteLine("Number of derived staff lines = " 
										+ st.no_staff_lines());

									break;
								}
							}

							current_st = stl.next();
						}

						if (found_match) break;
						current_ss = ssl.next();
					}

					if (!found_match)
					{
						no_missed_dst++;
					}

					current_dst = dstl.next();
				}

				current_dss = dssl.next();
			}

			Console.Out.WriteLine("Number of missed definitive staff line areas = " 
				+ no_missed_dst);

			Console.Out.WriteLine("Number of extra staff line areas    = " 
				+ (no_staffs + no_missed_dst - no_def_staffs));

		}

		
		protected void save_graphics_bitmap()
		{
			Console.Error.WriteLine("SepApplication.save_graphics_bitmap() not implemented");

			/*
			  
			string full_output_filename
				= FilenameManager.full_pathname(output_filename_);

			ostream ofs = FilenameManager::open_output_pipe(full_output_filename);
  
			if (print_level_>0) Console.Error.WriteLine("Saving");

			if (!FilenameManager::output_is_open())
												  {
													  Console.Error.WriteLine("Unable to open pipe for output");
													  return;
												  } 

			graphics_.save_bitmap(ofs);

			// ofs.close(); // ****?
			FilenameManager::close_output_pipe();

			delete &ofs;
			delete &full_output_filename;
			
			*/
		}



		protected void inlay_gobjects(Bitmap bitmap, GObjectList gol)
		{
			GObjectNode current_go = gol.head();
			while (current_go!=null)
			{
				GObject gobject = current_go.val();
				int gx_org = gobject.x_org();
				int gy_org = gobject.y_org();

				bitmap.inlay_bitmap(gobject,gx_org,gy_org);

				current_go = gol.next();
			}
		}

		protected void inlay_without_staff_lines(Bitmap scratch)
		{
			int page_x_dim = mobjects_.page_x_dim();
			int page_y_dim = mobjects_.page_y_dim();

			StaffSystemList ssl = mobjects_.staff_systems();
			StaffSystemNode ss_current = ssl.head();
			while (ss_current!=null)
			{
				StaffSystem staff_system = ss_current.val();
				GObjectList staff_system_objects = staff_system.staff_system_objects();
				inlay_gobjects(scratch,staff_system_objects);
      
				ss_current = ssl.next();
			}

			GObjectList non_staff_system_objects = mobjects_.non_staff_systems();
			inlay_gobjects(scratch,non_staff_system_objects);
		}


		protected void inlay_def_staff_lines_removed(Bitmap scratch)
		{
			scratch.clear_all();

			// inlay definitive staff lines in scratch;

			StaffSystemList dssl = mobjects_.def_staff_systems();
			StaffSystemNode current_dss = dssl.head();

			while (current_dss!=null)
			{
				StaffSystem dss = current_dss.val();

				StaffList dstl = dss.staffs();
				StaffNode current_dst = dstl.head();

				while (current_dst!=null)
				{
					Staff dst = current_dst.val();
					StaffLines dsl = dst.staff_lines();

					SubBitmapArray1d dsl_bitmaps = dsl.bitmaps();
					int no_staff_lines = dsl_bitmaps.length();

					for (int i=0; i<no_staff_lines; i++)
					{
						scratch.inlay_bitmap(dsl_bitmaps[i]);
					}

					current_dst = dstl.next();
				}

				current_dss = dssl.next();
			}
		}


		protected void inlay_staff_lines_removed(Bitmap scratch)
		{
			scratch.clear_all();

			// inlay staff lines in scratch;

			StaffSystemList ssl = mobjects_.staff_systems();
			StaffSystemNode current_ss = ssl.head();

			while (current_ss!=null)
			{
				StaffSystem ss = current_ss.val();

				StaffList stl = ss.staffs();
				StaffNode current_st = stl.head();

				while (current_st!=null)
				{
					Staff st = current_st.val();
					StaffLines sl = st.staff_lines();

					SubBitmapArray1d sl_bitmaps = sl.bitmaps();
					int no_staff_lines = sl_bitmaps.length();

					for (int i=0; i<no_staff_lines; i++)
					{
						scratch.inlay_bitmap(sl_bitmaps[i]);
					}

					current_st = stl.next();
				}

				current_ss = ssl.next();
			}
		}



		protected void def_comp_removed_staff_lines()
		{
			// derived staffs in mobjects.staffs();

			BitmapOps bitmap_ops = mobjects_.bitmap_ops();
			Bitmap scratch = bitmap_ops.scratch();

			inlay_def_staff_lines_removed(scratch);

			// compute definitive subdivided into 4;

			int scratch_x_dim = scratch.x_dim();
			int scratch_y_dim = scratch.y_dim();

			int x_sub1 = scratch_x_dim/4;
			int x_sub2 = scratch_x_dim/2;
			int x_sub3 = (3*scratch_x_dim)/4;

			int sxdm1 = scratch_x_dim -1;
			int sydm1 = scratch_y_dim -1;

			int def1 
				= bitmap_ops.Count.count_set_pixels(scratch,0,0,x_sub1,sydm1);
			int def2
				= bitmap_ops.Count.count_set_pixels(scratch,x_sub1+1,0,x_sub2,sydm1);
			int def3
				= bitmap_ops.Count.count_set_pixels(scratch,x_sub2+1,0,x_sub3,sydm1);
			int def4
				= bitmap_ops.Count.count_set_pixels(scratch,x_sub3+1,0,sxdm1,sydm1);

			Console.Out.WriteLine("Removed staff line definitive (first quarter)  = " + def1);
			Console.Out.WriteLine("Removed staff line definitive (second quarter) = " + def2);
			Console.Out.WriteLine("Removed staff line definitive (third quarter)  = " + def3);
			Console.Out.WriteLine("Removed staff line definitive (fourth quarter) = " + def4);


			// missing staff lines

			StaffSystemList ssl = mobjects_.staff_systems();
			StaffSystemNode current_ss = ssl.head();

			while (current_ss!=null)
			{
				StaffSystem ss = current_ss.val();

				StaffList stl = ss.staffs();
				StaffNode current_st = stl.head();

				while (current_st!=null)
				{
					Staff st = current_st.val();
					StaffLines sl = st.staff_lines();

					SubBitmapArray1d sl_bitmaps = sl.bitmaps();
					int no_staff_lines = sl_bitmaps.length();

					for (int i=0; i<no_staff_lines; i++)
					{
						scratch.remove_bitmap(sl_bitmaps[i]);
					}
					current_st = stl.next();
				}

				current_ss = ssl.next();
			}


			// compute missed pixels subdivided into 4;

			int miss1 
				= bitmap_ops.Count.count_set_pixels(scratch,0,0,x_sub1,sydm1);
			int miss2
				= bitmap_ops.Count.count_set_pixels(scratch,x_sub1+1,0,x_sub2,sydm1);
			int miss3
				= bitmap_ops.Count.count_set_pixels(scratch,x_sub2+1,0,x_sub3,sydm1);
			int miss4
				= bitmap_ops.Count.count_set_pixels(scratch,x_sub3+1,0,sxdm1,sydm1);

			Console.Out.WriteLine("Removed staff line missed (first quarter)  = " + miss1);
			Console.Out.WriteLine("Removed staff line missed (second quarter) = " + miss2);
			Console.Out.WriteLine("Removed staff line missed (third quarter)  = " + miss3);
			Console.Out.WriteLine("Removed staff line missed (fourth quarter) = " + miss4);


			// extra staff lines

			inlay_staff_lines_removed(scratch);

			StaffSystemList dssl = mobjects_.def_staff_systems();
			StaffSystemNode current_dss = dssl.head();

			while (current_dss!=null)
			{
				StaffSystem dss = current_dss.val();

				StaffList dstl = dss.staffs();
				StaffNode current_dst = dstl.head();

				while (current_dst!=null)
				{
					Staff dst = current_dst.val();
					StaffLines dsl = dst.staff_lines();

					SubBitmapArray1d dsl_bitmaps = dsl.bitmaps();
					int no_staff_lines = dsl_bitmaps.length();

					for (int i=0; i<no_staff_lines; i++)
					{
						scratch.remove_bitmap(dsl_bitmaps[i]);
					}

					current_dst = dstl.next();
				}

				current_dss = dssl.next();
			}

			// compute extra pixels subdivided into 4;

			int extra1 
				= bitmap_ops.Count.count_set_pixels(scratch,0,0,x_sub1,sydm1);
			int extra2
				= bitmap_ops.Count.count_set_pixels(scratch,x_sub1+1,0,x_sub2,sydm1);
			int extra3
				= bitmap_ops.Count.count_set_pixels(scratch,x_sub2+1,0,x_sub3,sydm1);
			int extra4
				= bitmap_ops.Count.count_set_pixels(scratch,x_sub3+1,0,sxdm1,sydm1);

			Console.Out.WriteLine("Removed staff line extra (first quarter)  = " + extra1);
			Console.Out.WriteLine("Removed staff line extra (second quarter) = " + extra2);
			Console.Out.WriteLine("Removed staff line extra (third quarter)  = " + extra3);
			Console.Out.WriteLine("Removed staff line extra (fourth quarter) = " + extra4);

		}


		/*--------------------*/

		protected void plot_gcharacters()
		{
			if (print_level_>0)
				Console.Error.WriteLine("Plotting graphical characters");

			bool is_def = gcharacters_mode_==MObjects.GcharactersModeEnum.gc_definitive;

			graphics_.clear_all();

			GCharList gcharacters 
				= (!is_def) ? mobjects_.gcharacters() : mobjects_.def_gcharacters();

			if (print_level_<4)
				graphics_.plot_gcharacters(gcharacters,GChar.SelectionEnum.gc_bitmap);
			else
				graphics_.plot_gcharacters(gcharacters,GChar.SelectionEnum.gc_library_graymap);

			graphics_.set_state(SepGraphics.StateEnum.gcharacters,is_def);
			save_graphics_bitmap();
		}

		protected void calc_gcharacters()
		{
			if (gcharacters_mode_ == MObjects.GcharactersModeEnum.gc_ignore) return;

			if (print_level_>0)
				Console.Error.WriteLine("Locating graphical characters");
			if (print_level_>=10)
				Console.Out.WriteLine("Locating graphical characters");

			FileBitmap current_fbitmap = page_manager_.current_fbitmap();

			machine_usage_.start_timer();

			GCharList gcharacters = mobjects_.gcharacters();
			mobjects_.generate_gcharacters(current_fbitmap,gcharacters_mode_,
				verbose_level_,computing_perc_,gcharacters);

			int cpu_time = machine_usage_.stop_timer();
			if (print_level_>=10)
			{
				Console.Out.WriteLine("\tCPU time: " + cpu_time);
			}

			gcharacters_done_ = true;

			if (stop_after_stage_ == StopAfterStageEnum.stop_gcharacters) 
			{
				if (do_view_) plot_gcharacters();
				Environment.Exit(0);
			}
		}
      
		protected void plot_staff_systems()
		{
			if (print_level_>0)
				Console.Error.WriteLine("Plotting staff systems");

			bool is_def = staff_systems_mode_==MObjects.FloodFillModeEnum.ff_definitive;

			graphics_.clear_all();

			StaffSystemList staff_systems
				= (!is_def) ? mobjects_.staff_systems() : mobjects_.def_staff_systems();

			graphics_.plot_staff_systems(staff_systems);

			graphics_.set_state(SepGraphics.StateEnum.staff_systems,is_def);
			save_graphics_bitmap();
		}

		protected void calc_staff_systems()
		{
			if (print_level_>0)
				Console.Error.WriteLine("Locating staff systems");
			if (print_level_>=10)
				Console.Out.WriteLine("Locating staff systems");

			machine_usage_.start_timer();

			FileBitmap current_fbitmap = page_manager_.current_fbitmap();
			StaffSystemList staff_systems=mobjects_.staff_systems();
			GObjectList non_staff_systems=mobjects_.non_staff_systems();
			mobjects_.generate_staff_systems(current_fbitmap,staff_systems_mode_,
				verbose_level_,computing_perc_,
				staff_systems, non_staff_systems);

			int cpu_time = machine_usage_.stop_timer();
			if (print_level_>=10)
			{
				Console.Out.WriteLine("\tCPU time: " + cpu_time);
			}

			staff_systems_done_ = true;

			if (MObjects.definitive_compare())
			{
				def_comp_staff_systems();
			}

			if (stop_after_stage_ == StopAfterStageEnum.stop_staff_systems) 
			{
				if (do_view_) plot_staff_systems();
				Environment.Exit(0);
			}
		}

		protected void plot_staff_areas()
		{
			if (print_level_>0)
				Console.Error.WriteLine("Plotting staff areas");

			bool is_def = staff_areas_mode_==MObjects.StaffAreasModeEnum.sa_definitive;

			int verbose_level = print_level_;

			graphics_.clear_all();

			switch(staff_areas_mode_)
			{
				case MObjects.StaffAreasModeEnum.sa_locate_4_5:
				case MObjects.StaffAreasModeEnum.sa_locate_3_4:
				case MObjects.StaffAreasModeEnum.sa_locate_6_10:
				case MObjects.StaffAreasModeEnum.sg_min_gap_4:
				case MObjects.StaffAreasModeEnum.sg_min_gap_6:
				case MObjects.StaffAreasModeEnum.sg_min_gap_12:
				case MObjects.StaffAreasModeEnum.sg_min_gap_30:
					graphics_.plot_staff_areas(mobjects_.staff_systems());
					graphics_.set_state(SepGraphics.StateEnum.staff_areas,is_def);
					break;

				case MObjects.StaffAreasModeEnum.sa_hproject:
				{
					StaffSystemList pj_ssl = mobjects_.staff_systems();
					StaffSystemNode pj_current_ss = pj_ssl.head();
					while (pj_current_ss!=null)
					{
						StaffSystem staff_system = pj_current_ss.val();
						GObject ss_gobject = staff_system.gobject();
						int ss_x_org = ss_gobject.x_org();
						int ss_y_org = ss_gobject.y_org();
						int ss_x_dim = ss_gobject.x_dim();
						int ss_y_dim = ss_gobject.y_dim();
	  
						graphics_.plot_horizontal_count(ss_gobject.y_project(),
							ss_x_org, ss_y_org,
							true);
						graphics_.plot_rectangle(ss_x_org,ss_y_org,ss_x_dim,ss_y_dim);
	  
						if (verbose_level>=8)
						{
							int max_value = ss_gobject.y_project().max_value();
							int ss_x_thresh = ss_x_org + (max_value/2);
	      
							graphics_.plot_line(ss_x_thresh,ss_y_org,
								ss_x_thresh,ss_y_org+ss_y_dim-1);
						}
						pj_current_ss = pj_ssl.next();
					}

					graphics_.set_state(SepGraphics.StateEnum.staff_areas_hproject,is_def);
					break;
				}

				case MObjects.StaffAreasModeEnum.sa_hpeaks:
				{
					StaffSystemList pk_ssl = mobjects_.staff_systems();
					StaffSystemNode pk_current_ss = pk_ssl.head();
					while (pk_current_ss!=null)
					{
						StaffSystem staff_system = pk_current_ss.val();
						GObject ss_gobject = staff_system.gobject();
						int ss_x_org = ss_gobject.x_org();
						int ss_y_org = ss_gobject.y_org();
						int ss_x_dim = ss_gobject.x_dim();
						int ss_y_dim = ss_gobject.y_dim();
	  
						graphics_.plot_horizontal_peaks(staff_system.peaks(),
							ss_x_org,ss_y_org);
						graphics_.plot_rectangle(ss_x_org,ss_y_org,ss_x_dim,ss_y_dim);
	  
						if (verbose_level>=8)
						{
							int max_value = ss_gobject.y_project().max_value();
							int ss_x_thresh = ss_x_org + (max_value/2);
	      
							graphics_.plot_line(ss_x_thresh,ss_y_org,
								ss_x_thresh,ss_y_org+ss_y_dim-1);
						}
	  
						pk_current_ss = pk_ssl.next();
					}

					graphics_.set_state(SepGraphics.StateEnum.staff_areas_hpeaks,is_def);
					break;
				}

				case MObjects.StaffAreasModeEnum.sa_staffs:
					graphics_.plot_staff_systems(mobjects_.staff_systems());
					graphics_.set_state(SepGraphics.StateEnum.staff_systems,is_def);
					break;

				case MObjects.StaffAreasModeEnum.sa_non_staffs:
				{
					GObjectList non_staff_systems = mobjects_.non_staff_systems();
					graphics_.plot_gobjects(non_staff_systems);
					graphics_.set_state(SepGraphics.StateEnum.non_staff_systems,is_def);
					break;
				}

				case MObjects.StaffAreasModeEnum.sa_definitive:
					graphics_.plot_staff_areas(mobjects_.def_staff_systems());
					graphics_.set_state(SepGraphics.StateEnum.staff_areas,is_def);
					break;

			}

			save_graphics_bitmap();
		}


		protected void calc_staff_areas()
		{
			if (print_level_>0)
				Console.Error.WriteLine("Locating staff areas");
			if (print_level_>=10)
				Console.Out.WriteLine("Locating staff areas");

    
			machine_usage_.start_timer();

			FileBitmap current_fbitmap = page_manager_.current_fbitmap();
			StaffSystemList staff_systems=mobjects_.staff_systems();
			GObjectList non_staff_systems=mobjects_.non_staff_systems();
			mobjects_.generate_staff_areas(current_fbitmap,
				staff_systems,non_staff_systems,
				sg_min_gap_,
				sa_prop_numer_,sa_prop_denom_,
				verbose_level_,verbose_extra_,
				computing_perc_);

			int cpu_time = machine_usage_.stop_timer();
			if (print_level_>=10)
			{
				Console.Out.WriteLine("\tCPU time: " + cpu_time);
			}

			staff_areas_done_ = true;

			if (MObjects.definitive_compare())
			{
				def_comp_staff_areas();
			}

			if (stop_after_stage_ == StopAfterStageEnum.stop_staff_areas) 
			{
				if (do_view_) plot_staff_areas();
				Environment.Exit(0);
			}
		}



		protected void calc_staff_segs_first_pass()
		{
			if (print_level_>0)
				Console.Error.WriteLine("Locating staff segs - first pass");
			if (print_level_>=10)
				Console.Out.WriteLine("Locating staff segs - first pass");

			machine_usage_.start_timer();

			FileBitmap current_fbitmap = page_manager_.current_fbitmap();
			StaffSystemList staff_systems=mobjects_.staff_systems();
			GObjectList non_staff_systems=mobjects_.non_staff_systems();
			mobjects_.generate_staff_segs_first_pass(current_fbitmap,
				staff_systems,non_staff_systems,
				seg_upt_mul_,
				seg_lowt_strict_mul_,
				seg_lowt_hyster_mul_,
				verbose_level_,verbose_extra_,
				computing_perc_);

			int cpu_time = machine_usage_.stop_timer();
			if (print_level_>=10)
			{
				Console.Out.WriteLine("\tCPU time: " + cpu_time);
			}

			staff_segs_done_ = true;

			if (MObjects.definitive_compare())
			{
				def_comp_staff_segs();
			}

		}


		protected int median_no_staff_lines()
		{
			StaffSystemList staff_systems = mobjects_.staff_systems();
			int no_staffs = mobjects_.no_staffs(staff_systems);
			IntArray1d no_staff_lines = new IntArray1d(no_staffs);

			int i=0;
			StaffSystemNode current_ss = staff_systems.head();
			while (current_ss!=null)
			{
				StaffSystem staff_system = current_ss.val();

				StaffList staffs = staff_system.staffs();
				StaffNode current_st = staffs.head();
				while (current_st!=null)
				{
					Staff staff = current_st.val();
					no_staff_lines[i++] = staff.no_staff_lines();

					current_st = staffs.next();
				}
      
				current_ss = staff_systems.next();
			}

			Statistics.IntValDelegate extract_int
				= new Statistics.IntValDelegate(Statistics.get_int);

			double no_staff_lines_median
				= Statistics.int_median(no_staff_lines.data(),extract_int);

			return (int)no_staff_lines_median;
		}


		protected void calc_staff_segs_second_pass()
		{
			if (print_level_>0)
				Console.Error.WriteLine("Locating staff areas - second pass");
			if (print_level_>=10)
				Console.Out.WriteLine("Locating staff areas - second pass");

			StaffInfo staff_info = mobjects_.staff_info();

			machine_usage_.start_timer();

			FileBitmap current_fbitmap = page_manager_.current_fbitmap();
			StaffSystemList staff_systems=mobjects_.staff_systems();
			int med_no_staff_lines = median_no_staff_lines();

			mobjects_.generate_staff_segs_second_pass(current_fbitmap,
				staff_systems,
				seg_upt_mul_,
				seg_lowt_strict_mul_,
				seg_lowt_hyster_mul_,
				verbose_level_,verbose_extra_,
				computing_perc_,
				staff_info,
				med_no_staff_lines);  

			int cpu_time = machine_usage_.stop_timer();
			if (print_level_>=10)
			{
				Console.Out.WriteLine("\tCPU time: " + cpu_time);
			}

			staff_segs_done_ = true;

			if (MObjects.definitive_compare())
			{
				def_comp_staff_segs();
			}

		}


		protected void plot_staff_segs()
		{
			if (print_level_>0)
				Console.Error.WriteLine("Plotting staff segs");

			bool is_def = staff_segs_mode_==MObjects.StaffSegModeEnum.seg_definitive;

			int verbose_level = print_level_;

			graphics_.clear_all();

			switch(staff_segs_mode_)
			{
				case MObjects.StaffSegModeEnum.seg_locate_1_2_3:
				case MObjects.StaffSegModeEnum.seg_locate_1_3_4:
				case MObjects.StaffSegModeEnum.seg_locate_1_4_5:
				case MObjects.StaffSegModeEnum.seg_locate_2_0_0:
					if (verbose_level>=4)
					{
						if (verbose_level==4)
						{
							graphics_.plot_staff_areas(mobjects_.staff_systems());
							graphics_.gray_out(1,9);
						}
	      
						StaffSystemList mos_ssl = mobjects_.staff_systems();
						StaffSystemNode mos_current_ss = mos_ssl.head();
						while (mos_current_ss!=null)
						{
							StaffSystem staff_system = mos_current_ss.val();
							StaffList stl = staff_system.staffs();
							StaffNode current_st = stl.head();
							while (current_st!=null)
							{
								Staff staff = current_st.val();
								GObject staff_area = staff.staff_area();
		  
								int sa_x_org = staff_area.x_org();
								int sa_y_org = staff_area.y_org();
								int sa_x_dim = staff_area.x_dim();
								int sa_y_dim = staff_area.y_dim();
		  
								int sa_xl = sa_x_org;
								int sa_yb = sa_y_org + sa_y_dim -1;
		  
								IntArray1d x_project = staff_area.x_project();
								int x_len = x_project.length();
		  
								if (verbose_level>=8)
								{		      
									BoolArray1d is_low_trough 
										= staff.is_low_trough();
		      
									IntArray1d filtered_x_project = new IntArray1d(x_len);
		      
									for (int i=0; i<x_len; i++)
									{
										filtered_x_project[i]
											= (is_low_trough[i]) ? x_project[i] : 0;
									}
		      
									graphics_.plot_vertical_peaks(filtered_x_project,
										sa_xl,sa_yb);
								}
								else if (verbose_level>=7)
								{		      
									IntArray1d filtered_x_project = new IntArray1d(x_len);
		      
									for (int i=0; i<x_len; i++)
										filtered_x_project[i]=0;
		      
									StaffSeg left_seg = staff.left_staff_seg();
									int ls_x_start = left_seg.x_start();
									int ls_x_end = left_seg.x_end();
									for (int lsi=ls_x_start; lsi<=ls_x_end; lsi++)
										filtered_x_project[lsi]=x_project[lsi];
		      
									StaffSeg right_seg = staff.right_staff_seg();
									int rs_x_start = right_seg.x_start();
									int rs_x_end = right_seg.x_end();
									for (int rsi=rs_x_start; rsi<=rs_x_end; rsi++)
										filtered_x_project[rsi]=x_project[rsi];
		      
									graphics_.plot_vertical_peaks(filtered_x_project,
										sa_xl,sa_yb);
								}
								else if (verbose_level>=6)
								{
									int xp_lower_thresh = staff.xp_lower_thresh();
									int xp_upper_thresh = staff.xp_upper_thresh();
		      
									IntArray1d filtered_x_project = new IntArray1d(x_len);
		      
									for (int i=0; i<x_len; i++)
									{
										int yp = x_project[i];
										filtered_x_project[i]
											= ((yp>=xp_lower_thresh) && (yp<=xp_upper_thresh))
											? yp : 0;
									}
		      
									graphics_.plot_vertical_peaks(filtered_x_project,
										sa_xl,sa_yb);
								}
								else
								{
									graphics_.plot_vertical_peaks(x_project,
										sa_xl,sa_yb);
								}
		  
								if ((verbose_level==4) || (verbose_level>=6))
								{
									graphics_.plot_rectangle(sa_x_org,sa_y_org,
										sa_x_dim,sa_y_dim);
								}
		  
								current_st = stl.next();
							}
							mos_current_ss = mos_ssl.next();
						}
	      
	      
						if (verbose_level==5)
						{
							graphics_.gray_out(1,7);
	      
							StaffSystemList ssl = mobjects_.staff_systems();
							StaffSystemNode current_ss = ssl.head();
							while (current_ss!=null)
							{
								StaffSystem staff_system = current_ss.val();
								StaffList stl = staff_system.staffs();
								StaffNode current_st = stl.head();
								while (current_st!=null)
								{
									Staff staff = current_st.val();
									GObject staff_area = staff.staff_area();
		      
									int sa_x_org = staff_area.x_org();
									int sa_y_org = staff_area.y_org();
									int sa_x_dim = staff_area.x_dim();
									int sa_y_dim = staff_area.y_dim();
		      
									int xp_upper_thresh = staff.xp_upper_thresh();
									int xp_lower_thresh = staff.xp_lower_thresh();
		      
									int sa_xl = sa_x_org;
									int sa_xr = sa_x_org+sa_x_dim-1;
									int sa_yb = sa_y_org+sa_y_dim-1;
									int sa_y_thu = sa_yb - xp_upper_thresh;
									int sa_y_thl = sa_yb - xp_lower_thresh;
		      
									graphics_.plot_line(sa_xl,sa_y_thu,sa_xr,sa_y_thu);
									graphics_.plot_line(sa_xl,sa_y_thl,sa_xr,sa_y_thl);
		      
									current_st = stl.next();
								}
								current_ss = ssl.next();
							}
						}
	  
					}
					else
					{
						// plus one on verbose_level causes lhs and rhs segs for vl==3
						graphics_.plot_staff_segs(mobjects_.staff_systems(),
							verbose_level+1);
					}

					graphics_.set_state(SepGraphics.StateEnum.staff_segs,is_def);
					break;

				case MObjects.StaffSegModeEnum.seg_hproject:
				{
					StaffSystemList ssl = mobjects_.staff_systems();
					StaffSystemNode current_ss = ssl.head();
					while (current_ss!=null)
					{
						StaffSystem staff_system = current_ss.val();
						StaffList stl = staff_system.staffs();
						StaffNode current_st = stl.head();
						while (current_st!=null)
						{
							Staff staff = current_st.val();
							GObject staff_area = staff.staff_area();
	      
							int sa_x_org = staff_area.x_org();
							int sa_y_org = staff_area.y_org();
							int sa_x_dim = staff_area.x_dim();
							int sa_y_dim = staff_area.y_dim();
	      
							StaffSeg left_seg = staff.left_staff_seg();
							Projection ls_y_project 
								= left_seg.gobject().y_project();
	      
							int gls_x_start = left_seg.global_x_start();
							int gls_x_end = left_seg.global_x_end();
							int gls_x_org = gls_x_start;
							int gls_x_dim = gls_x_end - gls_x_start +1; // **** ?
		
							graphics_.plot_horizontal_count(ls_y_project,
								gls_x_org,sa_y_org,
								true);
							//            graphics_.plot_rectangle(gls_x_org,sa_y_org,
							//                                     gls_x_dim,sa_y_dim); // **** ?
				
							StaffSeg right_seg = staff.right_staff_seg();
							Projection rs_y_project 
								= right_seg.gobject().y_project();

							int grs_x_start = right_seg.global_x_start();
							int grs_x_end = right_seg.global_x_end();
							int grs_x_org = grs_x_start;
							int grs_x_dim = grs_x_end - grs_x_start +1;
	      
							graphics_.plot_horizontal_count(rs_y_project,
								grs_x_org,sa_y_org,
								true);
							//	      graphics_.plot_rectangle(grs_x_org,sa_y_org,
							//                                     grs_x_dim,sa_y_dim); // **** ?

							graphics_.plot_rectangle(sa_x_org,sa_y_org,
								sa_x_dim,sa_y_dim);

							current_st = stl.next();
						}
						current_ss = ssl.next();
					}

					graphics_.set_state(SepGraphics.StateEnum.staff_segs_hproject,is_def);
					break;
				}

				case MObjects.StaffSegModeEnum.seg_staffs:
					graphics_.plot_staff_systems(mobjects_.staff_systems());
					graphics_.set_state(SepGraphics.StateEnum.staff_systems,is_def);
					break;

				case MObjects.StaffSegModeEnum.seg_non_staffs:
				{
					GObjectList non_staff_systems = mobjects_.non_staff_systems();
					graphics_.plot_gobjects(non_staff_systems);
					graphics_.set_state(SepGraphics.StateEnum.non_staff_systems,is_def);
					break;
				}

				case MObjects.StaffSegModeEnum.seg_definitive:
					int def_verb_level=6;
					graphics_.plot_staff_segs(mobjects_.def_staff_systems(),def_verb_level);
					graphics_.set_state(SepGraphics.StateEnum.staff_segs,is_def);
					break;

			}

			save_graphics_bitmap();
		}


		protected void calc_staff_segs()
		{
			calc_staff_segs_first_pass();

			if ((staff_segs_pass_mode_==Staff.PassEnum.second_pass) && 
				(stop_after_stage_ == StopAfterStageEnum.stop_staff_segs))
			{
				if (do_view_) plot_staff_segs();
				Environment.Exit(0);
			}
		}

		protected void plot_staff_lines()
		{
			if (print_level_>0)
				Console.Error.WriteLine("Plotting staff lines");

			bool is_def = staff_lines_mode_==MObjects.StaffLinesModeEnum.sl_definitive;

			int verbose_level = print_level_;

			graphics_.clear_all();

			switch(staff_lines_mode_)
			{
				case MObjects.StaffLinesModeEnum.sl_locate_strict:
				case MObjects.StaffLinesModeEnum.sl_locate_average:
				case MObjects.StaffLinesModeEnum.sl_locate_loose:
					if (verbose_level>=4)
					{
						if (verbose_level==4)
						{
							graphics_.plot_staff_areas(mobjects_.staff_systems());
							graphics_.gray_out(1,9);
						}
	  
						StaffSystemList mos_ssl = mobjects_.staff_systems();
						StaffSystemNode mos_current_ss = mos_ssl.head();
						while (mos_current_ss!=null)
						{
							StaffSystem staff_system = mos_current_ss.val();
							StaffList stl = staff_system.staffs();
							StaffNode current_st = stl.head();
							while (current_st!=null)
							{
								Staff staff = current_st.val();
								GObject staff_area = staff.staff_area();
		  
								int sa_x_org = staff_area.x_org();
								int sa_y_org = staff_area.y_org();
								int sa_x_dim = staff_area.x_dim();
								int sa_y_dim = staff_area.y_dim();
		  
								int sa_xl = sa_x_org;
								int sa_yb = sa_y_org + sa_y_dim -1;
		  
								IntArray1d x_project = staff_area.x_project();
								int x_len = x_project.length();
		  
								if (verbose_level>=8)
								{		      
									BoolArray1d is_low_trough 
										= staff.is_low_trough();
		      
									IntArray1d filtered_x_project = new IntArray1d(x_len);
		      
									for (int i=0; i<x_len; i++)
									{
										filtered_x_project[i]
											= (is_low_trough[i]) ? x_project[i] : 0;
									}
		      
									graphics_.plot_vertical_peaks(filtered_x_project,
										sa_xl,sa_yb);
								}
								else if (verbose_level>=7)
								{		      
									IntArray1d filtered_x_project = new IntArray1d(x_len);
		      
									for (int i=0; i<x_len; i++)
										filtered_x_project[i]=0;
		      
									StaffSeg left_seg = staff.left_staff_seg();
									int ls_x_start = left_seg.x_start();
									int ls_x_end = left_seg.x_end();
									for (int lsi=ls_x_start; lsi<=ls_x_end; lsi++)
										filtered_x_project[lsi]=x_project[lsi];
		      
									StaffSeg right_seg = staff.right_staff_seg();
									int rs_x_start = right_seg.x_start();
									int rs_x_end = right_seg.x_end();
									for (int rsi=rs_x_start; rsi<=rs_x_end; rsi++)
										filtered_x_project[rsi]=x_project[rsi];
		      
									graphics_.plot_vertical_peaks(filtered_x_project,
										sa_xl,sa_yb);
								}
								else if (verbose_level>=6)
								{
									int xp_lower_thresh = staff.xp_lower_thresh();
									int xp_upper_thresh = staff.xp_upper_thresh();
		      
									IntArray1d filtered_x_project = new IntArray1d(x_len);
		      
									for (int i=0; i<x_len; i++)
									{
										int yp = x_project[i];
										filtered_x_project[i]
											= ((yp>=xp_lower_thresh) && (yp<=xp_upper_thresh))
											? yp : 0;
									}
		      
									graphics_.plot_vertical_peaks(filtered_x_project,
										sa_xl,sa_yb);
								}
								else
								{
									graphics_.plot_vertical_peaks(x_project,
										sa_xl,sa_yb);
								}
		  
								if ((verbose_level==4) || (verbose_level>=6))
								{
									graphics_.plot_rectangle(sa_x_org,sa_y_org,
										sa_x_dim,sa_y_dim);
								}
		  
								current_st = stl.next();
							}
							mos_current_ss = mos_ssl.next();
						}

	  
						if (verbose_level==5)
						{
							graphics_.gray_out(1,7);
	      
							StaffSystemList ssl = mobjects_.staff_systems();
							StaffSystemNode current_ss = ssl.head();
							while (current_ss!=null)
							{
								StaffSystem staff_system = current_ss.val();
								StaffList stl = staff_system.staffs();
								StaffNode current_st = stl.head();
								while (current_st!=null)
								{
									Staff staff = current_st.val();
									GObject staff_area = staff.staff_area();
		      
									int sa_x_org = staff_area.x_org();
									int sa_y_org = staff_area.y_org();
									int sa_x_dim = staff_area.x_dim();
									int sa_y_dim = staff_area.y_dim();
		      
									int xp_upper_thresh = staff.xp_upper_thresh();
									int xp_lower_thresh = staff.xp_lower_thresh();
		      
									int sa_xl = sa_x_org;
									int sa_xr = sa_x_org+sa_x_dim-1;
									int sa_yb = sa_y_org+sa_y_dim-1;
									int sa_y_thu = sa_yb - xp_upper_thresh;
									int sa_y_thl = sa_yb - xp_lower_thresh;
		      
									graphics_.plot_line(sa_xl,sa_y_thu,sa_xr,sa_y_thu);
									graphics_.plot_line(sa_xl,sa_y_thl,sa_xr,sa_y_thl);
		      
									current_st = stl.next();
								}
								current_ss = ssl.next();
							}
						}
	  
					}
					else
					{
						graphics_.plot_staff_lines(mobjects_.staff_systems(),
							verbose_level);
					}

					graphics_.set_state(SepGraphics.StateEnum.staff_lines,is_def);
					break;

				case MObjects.StaffLinesModeEnum.sl_hproject:
				{
					StaffSystemList ssl = mobjects_.staff_systems();
					StaffSystemNode current_ss = ssl.head();
					while (current_ss!=null)
					{
						StaffSystem staff_system = current_ss.val();
						StaffList stl = staff_system.staffs();
						StaffNode current_st = stl.head();
						while (current_st!=null)
						{
							Staff staff = current_st.val();
							GObject staff_area = staff.staff_area();
	      
							int sa_x_org = staff_area.x_org();
							int sa_y_org = staff_area.y_org();
							int sa_x_dim = staff_area.x_dim();
							int sa_y_dim = staff_area.y_dim();
	      
							StaffSeg left_seg = staff.left_staff_seg();
							Projection ls_y_project 
								= left_seg.gobject().y_project();
	      
							int gls_x_start = left_seg.global_x_start();
							int gls_x_end = left_seg.global_x_end();
							int gls_x_org = gls_x_start;
							int gls_x_dim = gls_x_end - gls_x_start +1;
	      
							graphics_.plot_horizontal_count(ls_y_project,
								gls_x_org,sa_y_org,
								true);
							graphics_.plot_rectangle(gls_x_org,sa_y_org,
								gls_x_dim,sa_y_dim);

							StaffSeg right_seg = staff.right_staff_seg();
							Projection rs_y_project 
								= right_seg.gobject().y_project();
	      
							int grs_x_start = right_seg.global_x_start();
							int grs_x_end = right_seg.global_x_end();
							int grs_x_org = grs_x_start;
							int grs_x_dim = grs_x_end - grs_x_start +1;
	      
							graphics_.plot_horizontal_count(rs_y_project,
								grs_x_org,sa_y_org,
								true);
							graphics_.plot_rectangle(grs_x_org,sa_y_org,
								grs_x_dim,sa_y_dim);

							//	      graphics_.plot_rectangle(sa_x_org,sa_y_org,
							//				       sa_x_dim,sa_y_dim);
	      
							current_st = stl.next();
						}
						current_ss = ssl.next();
					}

					graphics_.set_state(SepGraphics.StateEnum.staff_lines_hproject,is_def);
					break;
				}

				case MObjects.StaffLinesModeEnum.sl_staffs:
					graphics_.plot_staff_systems(mobjects_.staff_systems());
					graphics_.set_state(SepGraphics.StateEnum.staff_systems,is_def);
					break;

				case MObjects.StaffLinesModeEnum.sl_non_staffs:
				{
					GObjectList non_staff_systems = mobjects_.non_staff_systems();
					graphics_.plot_gobjects(non_staff_systems);
					graphics_.set_state(SepGraphics.StateEnum.non_staff_systems,is_def);
					break;
				}

				case MObjects.StaffLinesModeEnum.sl_definitive:
					graphics_.plot_removed_staff_lines(mobjects_.def_staff_systems());
					graphics_.set_state(SepGraphics.StateEnum.staff_lines,is_def);
					break;

			}

			save_graphics_bitmap();
		}

		protected void calc_staff_lines()
		{
			if (print_level_>0)
				Console.Error.WriteLine("Locating staff lines");
			if (print_level_>=10)
				Console.Out.WriteLine("Locating staff lines");

			if (staff_segs_pass_mode_==Staff.PassEnum.first_pass)
			{
				machine_usage_.start_timer();
			}

			FileBitmap current_fbitmap = page_manager_.current_fbitmap();
			StaffSystemList staff_systems=mobjects_.staff_systems();
			GObjectList non_staff_systems=mobjects_.non_staff_systems();

			mobjects_.generate_staff_lines(current_fbitmap,
				staff_systems,non_staff_systems,
				sl_lines_vt_,sl_gaps_vt_,
				sl_lines_rt_,sl_gaps_rt_,
				verbose_level_,verbose_extra_,
				computing_perc_);

			if (staff_segs_pass_mode_==Staff.PassEnum.second_pass)
			{
				calc_staff_segs_second_pass();

				if (stop_after_stage_ == StopAfterStageEnum.stop_staff_segs)
				{
					if (do_view_) plot_staff_segs();
					Environment.Exit(0);
				}

				machine_usage_.start_timer();

				mobjects_.generate_staff_lines(current_fbitmap,
					staff_systems,non_staff_systems,
					sl_lines_vt_,sl_gaps_vt_,
					sl_lines_rt_,sl_gaps_rt_,
					verbose_level_,verbose_extra_,
					computing_perc_);
			}

			int cpu_time = machine_usage_.stop_timer();
			if (print_level_>=10)
			{
				Console.Out.WriteLine("\tCPU time: " + cpu_time);
			}

			staff_lines_done_ = true;

			if (MObjects.definitive_compare())
			{
				def_comp_staff_lines();
			}

			if (stop_after_stage_ == StopAfterStageEnum.stop_staff_lines) 
			{
				if (do_view_) plot_staff_lines();
				Environment.Exit(0);
			}
		}


		protected void plot_correct_skew()
		{
			if (print_level_>0)
				Console.Error.WriteLine("Plotting staff systems");


			graphics_.clear_all();

			switch (correct_skew_mode_)
			{
				case MObjects.CorrectSkewModeEnum.rotate_page:
				case MObjects.CorrectSkewModeEnum.shear_page:
				case MObjects.CorrectSkewModeEnum.shear_page8:
				case MObjects.CorrectSkewModeEnum.rotate_lines:
				case MObjects.CorrectSkewModeEnum.shear_lines:
				case MObjects.CorrectSkewModeEnum.shear_lines8:
				{
					Bitmap corrected_page = mobjects_.corrected_page();
					if (corrected_page!=null) 
					{
						graphics_.plot_bitmap(corrected_page);
						graphics_.plot_gcharacters(mobjects_.gcharacters(),GChar.SelectionEnum.gc_bitmap);
					}
					break;
				}

				case MObjects.CorrectSkewModeEnum.rotate_systems:
				case MObjects.CorrectSkewModeEnum.shear_systems:
				case MObjects.CorrectSkewModeEnum.shear_systems8:
					StaffSystemList corrected_staff_systems
						= mobjects_.corrected_staff_systems();
					graphics_.plot_staff_systems(corrected_staff_systems);

					GObjectList corrected_non_staff_systems
						= mobjects_.corrected_non_staff_systems();
					graphics_.plot_gobjects(corrected_non_staff_systems);

					graphics_.plot_gcharacters(mobjects_.gcharacters(),GChar.SelectionEnum.gc_bitmap);

					break;

					// graphics_.set_state(SepGraphics.StateEnum.corrected_bitmap,false); // was placed after break statement.  correct state set (probably) from calling plot_gcharacters
			}
			save_graphics_bitmap();

		}


		protected void calc_correct_skew()
		{
			if (cond_skew_val_>0.0)
			{
				StaffInfo staff_info = mobjects_.staff_info();
				double rad_skew = staff_info.skew();
				double deg_skew = 360.0*Math.Atan(rad_skew)/(2.0*Math.PI);
				double fabs_deg_skew = Math.Abs(deg_skew);
				if (fabs_deg_skew<cond_skew_val_) return;
				if (print_level_>0)
					Console.Error.WriteLine("Skew of " + deg_skew + " exceeds conditional limit");
				if (print_level_>10)
					Console.Out.WriteLine("Skew of " + deg_skew + " exceeds conditional limit");

			}

			MObjects.CorrectSkewModeEnum mode = correct_skew_mode_;
			if ((mode==MObjects.CorrectSkewModeEnum.rotate_lines) || (mode==MObjects.CorrectSkewModeEnum.shear_lines) || 
				(mode==MObjects.CorrectSkewModeEnum.shear_lines8))
			{
				if (!remove_staffs_done_) calc_remove_staffs();
			}

			if (print_level_>0)
				Console.Error.WriteLine("Correcting for skew");
			if (print_level_>=10)
				Console.Out.WriteLine("Correcting for skew");

			machine_usage_.start_timer();

			mobjects_.generate_corrected_skew(correct_skew_mode_,machine_usage_,
				1,computing_perc_);

			int cpu_time = machine_usage_.stop_timer();
			if (print_level_>=10)
			{
				Console.Out.WriteLine("\tCPU time: " + cpu_time);
			}

			correct_skew_done_ = true;

			if (stop_after_stage_ == StopAfterStageEnum.stop_correct_skew) 
			{
				if (do_view_) plot_correct_skew();
				Environment.Exit(0);
			}
		}

		protected void plot_remove_staffs()
		{
			if (print_level_>0)
				Console.Error.WriteLine("Plotting staff objects");


			graphics_.clear_all();

			int verbose_level = print_level_;
			MObjects.CorrectSkewModeEnum skew_mode = correct_skew_mode_;
			switch (skew_mode)
			{
				case MObjects.CorrectSkewModeEnum.rotate_page:
				case MObjects.CorrectSkewModeEnum.shear_page:
				case MObjects.CorrectSkewModeEnum.shear_page8:
				case MObjects.CorrectSkewModeEnum.rotate_lines:
				case MObjects.CorrectSkewModeEnum.shear_lines:
				case MObjects.CorrectSkewModeEnum.shear_lines8:
				{
					Bitmap corrected_page = mobjects_.corrected_page();
					if (corrected_page!=null)
					{
						// corrected staff objects
						StaffSystemList cssl = mobjects_.corrected_staff_systems();
						StaffSystemNode ss_current = cssl.head();
						while (ss_current!=null)
						{
							StaffSystem staff_system = ss_current.val();
							GObjectList staff_system_objects 
								= (verbose_level<=4) ? staff_system.staff_system_objects()
								:  staff_system.processed_staff_system_objects();
							graphics_.plot_gobjects(staff_system_objects);

							ss_current = cssl.next();
						}
						graphics_.set_state(SepGraphics.StateEnum.corrected_staff_objects,false);
					}
					else
					{
						// uncorrected staff objects	  
						StaffSystemList ssl = mobjects_.staff_systems();
						StaffSystemNode ss_current = ssl.head();
						while (ss_current!=null)
						{
							StaffSystem staff_system = ss_current.val();
							GObjectList staff_system_objects 
								= (verbose_level<=4) ? staff_system.staff_system_objects()
								:  staff_system.processed_staff_system_objects();
							graphics_.plot_gobjects(staff_system_objects);

							ss_current = ssl.next();
						}
						graphics_.set_state(SepGraphics.StateEnum.staff_objects,false);
					}
					break;
				}

				case MObjects.CorrectSkewModeEnum.rotate_systems:
				case MObjects.CorrectSkewModeEnum.shear_systems:
				case MObjects.CorrectSkewModeEnum.shear_systems8:
					int css_len = mobjects_.corrected_staff_systems().length();
					if (css_len>0)
					{
						// corrected staff objects
						StaffSystemList cssl = mobjects_.corrected_staff_systems();
						StaffSystemNode ss_current = cssl.head();
						while (ss_current!=null)
						{
							StaffSystem staff_system = ss_current.val();
							GObjectList staff_system_objects 
								= (verbose_level<=4) ? staff_system.staff_system_objects()
								:  staff_system.processed_staff_system_objects();
							graphics_.plot_gobjects(staff_system_objects);

							ss_current = cssl.next();
						}
						graphics_.set_state(SepGraphics.StateEnum.corrected_staff_objects,false);
					}
					else
					{
						// uncorrected staff objects
						StaffSystemList ssl = mobjects_.staff_systems();
						StaffSystemNode ss_current = ssl.head();
						while (ss_current!=null)
						{
							StaffSystem staff_system = ss_current.val();
							GObjectList staff_system_objects 
								= (verbose_level<=4) ? staff_system.staff_system_objects()
								:  staff_system.processed_staff_system_objects();
							graphics_.plot_gobjects(staff_system_objects);

							ss_current = ssl.next();
						}
						graphics_.set_state(SepGraphics.StateEnum.staff_objects,false);
					}

					break;
			}

			save_graphics_bitmap();
		}


		protected void calc_remove_staffs()
		{
			machine_usage_.start_timer();

			switch (correct_skew_mode_)
			{
				case MObjects.CorrectSkewModeEnum.rotate_page:
				case MObjects.CorrectSkewModeEnum.shear_page:
				case MObjects.CorrectSkewModeEnum.shear_page8:
				case MObjects.CorrectSkewModeEnum.rotate_lines:
				case MObjects.CorrectSkewModeEnum.shear_lines:
				case MObjects.CorrectSkewModeEnum.shear_lines8:
				{
					Bitmap corrected_page = mobjects_.corrected_page();
					if ((object)corrected_page!=null)
					{
						StaffSystemList cstaff_systems=mobjects_.corrected_staff_systems();
						GObjectList cnstaff_systems=mobjects_.corrected_non_staff_systems();

						if (print_level_>0)
							Console.Error.WriteLine("Locating corrected staff systems");
						if (print_level_>=10)
							Console.Out.WriteLine("Locating corrected staff systems");
						mobjects_.generate_staff_systems(corrected_page,staff_systems_mode_,
							verbose_level_,computing_perc_,
							cstaff_systems, cnstaff_systems);

						if (print_level_>0)
							Console.Error.WriteLine("Locating corrected staff areas");
						if (print_level_>=10)
							Console.Out.WriteLine("Locating corrected staff areas");
						mobjects_.generate_staff_areas(corrected_page,
							cstaff_systems,cnstaff_systems,
							sg_min_gap_,
							sa_prop_numer_,sa_prop_denom_,
							verbose_level_,verbose_extra_,
							computing_perc_);

						if (print_level_>0)
							Console.Error.WriteLine("Locating corrected staff segs");
						if (print_level_>=10)
							Console.Out.WriteLine("Locating corrected staff segs");
						// change this to second pass since staff line thickness now known !?
						// ****** (and other similar places)
						mobjects_.generate_staff_segs_first_pass(corrected_page,
							cstaff_systems,cnstaff_systems,
							seg_upt_mul_,
							seg_lowt_strict_mul_,
							seg_lowt_hyster_mul_,
							verbose_level_,verbose_extra_,
							computing_perc_);

						if (print_level_>0)
							Console.Error.WriteLine("Locating corrected staff lines");
						if (print_level_>=10)
							Console.Out.WriteLine("Locating corrected staff lines");
						mobjects_.generate_staff_lines(corrected_page,
							cstaff_systems,cnstaff_systems,
							sl_lines_vt_,sl_gaps_vt_,
							sl_lines_rt_,sl_gaps_rt_,
							verbose_level_,verbose_extra_,
							computing_perc_);

						if (print_level_>0)
							Console.Error.WriteLine("Removing false staff systems");
						if (print_level_>=10)
							Console.Out.WriteLine("Removing false staff systems");
						mobjects_.filter_staff_systems_with_no_staffs(corrected_page,
							cstaff_systems,
							cnstaff_systems);

						if (print_level_>0)
							Console.Error.WriteLine("Removing staffs");
						if (print_level_>=10)
							Console.Out.WriteLine("Removing staffs");
						mobjects_.generate_all_staff_lines(cstaff_systems,cnstaff_systems,
							remove_staffs_mode_,
							verbose_level_,computing_perc_);
					}
					else
					{
						FileBitmap current_fbitmap = page_manager_.current_fbitmap();
						StaffSystemList staff_systems = mobjects_.staff_systems();
						GObjectList non_staff_systems = mobjects_.non_staff_systems();

						if (print_level_>0)
							Console.Error.WriteLine("Removing false staff systems");
						if (print_level_>=10)
							Console.Out.WriteLine("Removing false staff systems");
						mobjects_.filter_staff_systems_with_no_staffs(current_fbitmap,
							staff_systems,
							non_staff_systems);

						if (print_level_>0)
							Console.Error.WriteLine("Removing staffs");
						if (print_level_>=10)
							Console.Out.WriteLine("Removing staffs");
						mobjects_.generate_all_staff_lines(staff_systems,non_staff_systems,
							remove_staffs_mode_,
							verbose_level_,computing_perc_);
					}
					break;
				}

				case MObjects.CorrectSkewModeEnum.rotate_systems:
				case MObjects.CorrectSkewModeEnum.shear_systems:
				case MObjects.CorrectSkewModeEnum.shear_systems8:
				{
					int css_len = mobjects_.corrected_staff_systems().length();
					if (css_len>0)
					{
						Bitmap corrected_page = mobjects_.corrected_page();
						StaffSystemList cstaff_systems=mobjects_.corrected_staff_systems();
						GObjectList cnstaff_systems=mobjects_.corrected_non_staff_systems();

						if (print_level_>0)
							Console.Error.WriteLine("Locating corrected staff areas");
						if (print_level_>=10)
							Console.Out.WriteLine("Locating corrected staff areas");
						mobjects_.generate_staff_areas(corrected_page,
							cstaff_systems,cnstaff_systems,
							sg_min_gap_,
							sa_prop_numer_,sa_prop_denom_,
							verbose_level_,verbose_extra_,
							computing_perc_);

						if (print_level_>0)
							Console.Error.WriteLine("Locating corrected staff segs");
						if (print_level_>=0)
							Console.Out.WriteLine("Locating corrected staff segs");
						mobjects_.generate_staff_segs_first_pass(corrected_page,
							cstaff_systems,cnstaff_systems,
							seg_upt_mul_,
							seg_lowt_strict_mul_,
							seg_lowt_hyster_mul_,
							verbose_level_,verbose_extra_,
							computing_perc_);

						if (print_level_>0)
							Console.Error.WriteLine("Locating corrected staff lines");
						if (print_level_>=10)
							Console.Out.WriteLine("Locating corrected staff lines");
						mobjects_.generate_staff_lines(corrected_page,
							cstaff_systems,cnstaff_systems,
							sl_lines_vt_,sl_gaps_vt_,
							sl_lines_rt_,sl_gaps_rt_,
							verbose_level_,verbose_extra_,
							computing_perc_);

						if (print_level_>0)
							Console.Error.WriteLine("Removing false staff systems");
						if (print_level_>=10)
							Console.Out.WriteLine("Removing false staff systems");
						mobjects_.filter_staff_systems_with_no_staffs(corrected_page,
							cstaff_systems,
							cnstaff_systems);

						if (print_level_>0)
							Console.Error.WriteLine("Removing staffs");
						if (print_level_>=10)
							Console.Out.WriteLine("Removing staffs");
						mobjects_.generate_all_staff_lines(cstaff_systems,cnstaff_systems,
							remove_staffs_mode_,
							verbose_level_,computing_perc_);
					}
					else
					{
						FileBitmap current_fbitmap = page_manager_.current_fbitmap();
						StaffSystemList staff_systems=mobjects_.staff_systems();
						GObjectList non_staff_systems=mobjects_.non_staff_systems();

						if (print_level_>0)
							Console.Error.WriteLine("Removing false staff systems");
						if (print_level_>=10)
							Console.Out.WriteLine("Removing false staff systems");
						mobjects_.filter_staff_systems_with_no_staffs(current_fbitmap,
							staff_systems,
							non_staff_systems);
						if (print_level_>0)
							Console.Error.WriteLine("Removing staffs");
						if (print_level_>=10)
							Console.Out.WriteLine("Removing staffs");
						mobjects_.generate_all_staff_lines(staff_systems,non_staff_systems,
							remove_staffs_mode_,
							verbose_level_,computing_perc_);
					}
					break;
				}
			}

			int cpu_time = machine_usage_.stop_timer();
			if (print_level_>=10)
			{
				Console.Out.WriteLine("\tCPU time: " + cpu_time);
			}

			remove_staffs_done_ = true;

			if (MObjects.definitive_compare())
			{
				def_comp_removed_staff_lines();
			}

			if (stop_after_stage_ == StopAfterStageEnum.stop_remove_staffs) 
			{
				if (do_view_) plot_remove_staffs();
				Environment.Exit(0);
			}
		}

		void plot_correct_deform()
		{
			if (print_level_>0)
				Console.Error.WriteLine("Plotting corrected deformation");


			graphics_.clear_all();

			int verbose_level = print_level_;
			int half_y_dim = page_manager_.current_fbitmap().y_dim()/2;

			if (verbose_level>=6)
				graphics_.plot_bitmap(mobjects_.deform().thick_line(),0,half_y_dim);
			else if (verbose_level>=4)
				graphics_.plot_bitmap(mobjects_.deform().thin_line(),0,half_y_dim);
			else if (verbose_level>=3)
				graphics_.plot_bitmap(mobjects_.deform().fthin_line(),0,half_y_dim);
			else if (verbose_level>=2)
				graphics_.plot_bitmap(mobjects_.deform().efthin_line(),0,half_y_dim);
			else if (verbose_level>=1)
				graphics_.plot_bitmap(mobjects_.corrected_bitmap());
			else
			{
				Bitmap corrected_page = mobjects_.corrected_page();
				if (corrected_page==null) 
				{
					Console.Error.WriteLine("Warning: No correction for deform has been performed." 
						);
				}
				else
				{
					graphics_.plot_bitmap(mobjects_.corrected_bitmap());
					graphics_.plot_gcharacters(mobjects_.gcharacters(),
						GChar.SelectionEnum.gc_bitmap);
				}
			}

			save_graphics_bitmap();
		}


		protected void calc_correct_deform()
		{
			if (print_level_>0)
				Console.Error.WriteLine("Correcting for deform");
			if (print_level_>=10)
				Console.Out.WriteLine("Correcting for deform");

			double cd_max_acc_thresh=0.0;
			switch (correct_deform_mode_)
			{
				case MObjects.CorrectDeformModeEnum.deform_level_60:
				case MObjects.CorrectDeformModeEnum.deform_angle_60:
					cd_max_acc_thresh = 0.6;
					break;

				case MObjects.CorrectDeformModeEnum.deform_level_70:
				case MObjects.CorrectDeformModeEnum.deform_angle_70:
					cd_max_acc_thresh = 0.7;
					break;

				case MObjects.CorrectDeformModeEnum.deform_level_80:
				case MObjects.CorrectDeformModeEnum.deform_angle_80:
					cd_max_acc_thresh = 0.8;
					break;
			}

			machine_usage_.start_timer();

			switch (correct_skew_mode_)
			{
				case MObjects.CorrectSkewModeEnum.rotate_page:
				case MObjects.CorrectSkewModeEnum.shear_page:
				case MObjects.CorrectSkewModeEnum.shear_page8:
				case MObjects.CorrectSkewModeEnum.rotate_lines:
				case MObjects.CorrectSkewModeEnum.shear_lines:
				case MObjects.CorrectSkewModeEnum.shear_lines8:
				{
					Bitmap corrected_page = mobjects_.corrected_page();
					if (corrected_page!=null)
					{
						StaffSystemList cstaff_systems=mobjects_.corrected_staff_systems();
						GObjectList cnstaff_systems=mobjects_.corrected_non_staff_systems();

						if (print_level_>0)
							Console.Error.WriteLine("Locating corrected staff systems");
						if (print_level_>=10)
							Console.Out.WriteLine("Locating corrected staff systems");
						mobjects_.generate_staff_systems(corrected_page,staff_systems_mode_,
							verbose_level_,computing_perc_,
							cstaff_systems, cnstaff_systems);

						if (print_level_>0)
							Console.Error.WriteLine("Locating corrected staff areas");
						if (print_level_>=10)
							Console.Out.WriteLine("Locating corrected staff areas");
						mobjects_.generate_staff_areas(corrected_page,
							cstaff_systems,cnstaff_systems,
							sg_min_gap_,
							sa_prop_numer_,sa_prop_denom_,
							verbose_level_,verbose_extra_,
							computing_perc_);

						if (print_level_>0)
							Console.Error.WriteLine("Locating corrected staff segs");
						if (print_level_>=10)
							Console.Out.WriteLine("Locating corrected staff segs");
						mobjects_.generate_staff_segs_first_pass(corrected_page,
							cstaff_systems,cnstaff_systems,
							seg_upt_mul_,
							seg_lowt_strict_mul_,
							seg_lowt_hyster_mul_,
							verbose_level_,verbose_extra_,
							computing_perc_);

						if (print_level_>0)
							Console.Error.WriteLine("Locating corrected staff lines");
						if (print_level_>=10)
							Console.Out.WriteLine("Locating corrected staff lines");
						mobjects_.generate_staff_lines(corrected_page,
							cstaff_systems,cnstaff_systems,
							sl_lines_vt_,sl_gaps_vt_,
							sl_lines_rt_,sl_gaps_rt_,
							verbose_level_,verbose_extra_,
							computing_perc_);

						if (print_level_>0)
							Console.Error.WriteLine("Removing false staff systems");
						if (print_level_>=10)
							Console.Out.WriteLine("Removing false staff systems");
						mobjects_.filter_staff_systems_with_no_staffs(corrected_page,
							cstaff_systems,
							cnstaff_systems);

						if (print_level_>0)
							Console.Error.WriteLine("Removing staffs");
						if (print_level_>=10)
							Console.Out.WriteLine("Removing staffs");
						mobjects_.generate_all_staff_lines(cstaff_systems,cnstaff_systems,
							remove_staffs_mode_,
							verbose_level_,computing_perc_);

						if (print_level_>0)
							Console.Error.WriteLine("Correcting deformation");
						if (print_level_>=10)
							Console.Out.WriteLine("Correcting deformation");
						int max_x_dim = corrected_page.x_dim();
						mobjects_.generate_deform(cstaff_systems,max_x_dim,
							correct_deform_mode_,cd_max_acc_thresh,
							verbose_level_,computing_perc_);
						mobjects_.correct_deform(corrected_page,computing_perc_);
					}
					else
					{
						// page has not been rotated, use original
						StaffSystemList staff_systems = mobjects_.staff_systems();
						GObjectList non_staff_systems = mobjects_.non_staff_systems();

						if (print_level_>0)
							Console.Error.WriteLine("Correcting for deform");
						if (print_level_>=10)
							Console.Out.WriteLine("Correcting for deform");
						FileBitmap current_fbitmap = page_manager_.current_fbitmap();
						int max_x_dim = current_fbitmap.x_dim();
						mobjects_.generate_deform(staff_systems,max_x_dim,
							correct_deform_mode_,cd_max_acc_thresh,
							verbose_level_,computing_perc_);
						mobjects_.correct_deform(current_fbitmap,computing_perc_);
					}
					break;
				}

				case MObjects.CorrectSkewModeEnum.rotate_systems:
				case MObjects.CorrectSkewModeEnum.shear_systems:
				case MObjects.CorrectSkewModeEnum.shear_systems8:
				{
					int css_len = mobjects_.corrected_staff_systems().length();
					if (css_len>0)
					{
						Bitmap corrected_page = mobjects_.corrected_page();
						StaffSystemList cstaff_systems=mobjects_.corrected_staff_systems();
						GObjectList cnstaff_systems=mobjects_.corrected_non_staff_systems();

						if (print_level_>0)
							Console.Error.WriteLine("Locating corrected staff areas");
						if (print_level_>=10)
							Console.Out.WriteLine("Locating corrected staff areas");
						mobjects_.generate_staff_areas(corrected_page,
							cstaff_systems,cnstaff_systems,
							sg_min_gap_,
							sa_prop_numer_,sa_prop_denom_,
							verbose_level_,verbose_extra_,
							computing_perc_);

						if (print_level_>0)
							Console.Error.WriteLine("Locating corrected staff segs");
						if (print_level_>=10)
							Console.Out.WriteLine("Locating corrected staff segs");
						mobjects_.generate_staff_segs_first_pass(corrected_page,
							cstaff_systems,cnstaff_systems,
							seg_upt_mul_,
							seg_lowt_strict_mul_,
							seg_lowt_hyster_mul_,
							verbose_level_,verbose_extra_,
							computing_perc_);

						if (print_level_>0)
							Console.Error.WriteLine("Locating corrected staff lines");
						if (print_level_>=10)
							Console.Out.WriteLine("Locating corrected staff lines");
						mobjects_.generate_staff_lines(corrected_page,
							cstaff_systems,cnstaff_systems,
							sl_lines_vt_,sl_gaps_vt_,
							sl_lines_rt_,sl_gaps_rt_,
							verbose_level_,verbose_extra_,
							computing_perc_);

						if (print_level_>0)
							Console.Error.WriteLine("Removing staffs");
						if (print_level_>=10)
							Console.Out.WriteLine("Removing staffs");
						mobjects_.generate_all_staff_lines(cstaff_systems,cnstaff_systems,
							remove_staffs_mode_,
							verbose_level_,computing_perc_);

						if (print_level_>0)
							Console.Error.WriteLine("Correcting for deform");
						if (print_level_>=10)
							Console.Out.WriteLine("Correcting for deform");
						FileBitmap current_fbitmap = page_manager_.current_fbitmap();
						int max_x_dim = current_fbitmap.x_dim();
						mobjects_.generate_deform(cstaff_systems,max_x_dim,
							correct_deform_mode_,cd_max_acc_thresh,
							verbose_level_,computing_perc_);
						mobjects_.correct_deform(current_fbitmap,computing_perc_);
					}
					else
					{
						StaffSystemList staff_systems=mobjects_.staff_systems();
						GObjectList non_staff_systems=mobjects_.non_staff_systems();

						if (print_level_>0)
							Console.Error.WriteLine("Correcting for deform");
						if (print_level_>=10)
							Console.Out.WriteLine("Correcting for deform");
						FileBitmap current_fbitmap = page_manager_.current_fbitmap();
						int max_x_dim = current_fbitmap.x_dim();
						mobjects_.generate_deform(staff_systems,max_x_dim,
							correct_deform_mode_,cd_max_acc_thresh,
							verbose_level_,computing_perc_);
						mobjects_.correct_deform(current_fbitmap,computing_perc_);
					}
					break;
				}
			}

			int cpu_time = machine_usage_.stop_timer();
			if (print_level_>=10)
			{
				Console.Out.WriteLine("\tCPU time: " + cpu_time);
			}

			correct_deform_done_ = true;

			if (stop_after_stage_ == StopAfterStageEnum.stop_correct_deform) 
			{
				if (do_view_) plot_correct_deform();
				Environment.Exit(0);
			}
		}

		/*--------------------*/


		protected void apply_correct_skew()
		{
			FileBitmap current_fbitmap = page_manager_.current_fbitmap();
			if (!current_fbitmap.valid()) return;
  
			if (!gcharacters_done_) calc_gcharacters();
			if (!staff_systems_done_) calc_staff_systems();
			if (!staff_areas_done_) calc_staff_areas();
			if (!staff_segs_done_) calc_staff_segs();
			if (!staff_lines_done_) calc_staff_lines();

			calc_correct_skew();
		}

		protected void apply_remove_staffs()
		{
			FileBitmap current_fbitmap = page_manager_.current_fbitmap();
			if (!current_fbitmap.valid()) return;

			if (!gcharacters_done_) calc_gcharacters();
			if (!staff_systems_done_) calc_staff_systems();
			if (!staff_areas_done_) calc_staff_areas();
			if (!staff_segs_done_) calc_staff_segs();
			if (!staff_lines_done_) calc_staff_lines();

			calc_remove_staffs();
		}

		protected void apply_correct_deform()
		{
			FileBitmap current_fbitmap = page_manager_.current_fbitmap();
			if (!current_fbitmap.valid()) return;

			if (!gcharacters_done_) calc_gcharacters();
			if (!staff_systems_done_) calc_staff_systems();
			if (!staff_areas_done_) calc_staff_areas();
			if (!staff_segs_done_) calc_staff_segs();
			if (!staff_lines_done_) calc_staff_lines();
			if (!remove_staffs_done_) calc_remove_staffs();

			calc_correct_deform();
		}

		/*--------------------*/


		public void default_run()
		{
			if (FilenameManager.file_type()!=FilenameManager.FileTypeEnum.sep)
			{
				if ((stage_options_ & StageOptionEnum.skew_opt) != 0) apply_correct_skew();
				apply_remove_staffs();
				if ((stage_options_ & StageOptionEnum.deform_opt) != 0) apply_correct_deform();
			}
			else
			{
				// if sep file, do OCR stage if option is switched on
				if (!gcharacters_done_) calc_gcharacters(); 
			}
    
		}


		
		public void customised_run()
		{
			// Console.Error.WriteLine("Warning: SepApplication.pracscore_run() partially implemented!");

			
			if (FilenameManager.file_type()!=FilenameManager.FileTypeEnum.sep)
			{
				if ((stage_options_ & StageOptionEnum.skew_opt) != 0) apply_correct_skew();
				apply_remove_staffs();
				if ((stage_options_ & StageOptionEnum.deform_opt) != 0) apply_correct_deform();
			}
			else
			{
				// if sep file, do OCR stage if option is switched on
				if (!gcharacters_done_) calc_gcharacters(); 
			}

			
			if (FilenameManager.file_type()!=FilenameManager.FileTypeEnum.sep)
			{

				string stored_filename = output_filename_;
				string root_filename = FilenameManager.root_filename();

      
				// generate without stafflines
				gen_without_sl_ = true;
      
				string wsl_filename = root_filename + "_w.pbm";
				output_filename_ = wsl_filename;

				Console.Out.WriteLine("Without staff-lines: ");

				write_pbm();
				gen_without_sl_ = false;

				// generate corrected page if active
				Bitmap corrected_page = mobjects_.corrected_page();
				if ((object)corrected_page!=null) 
				{
					string rot_filename =  root_filename + "_r.pbm";
					output_filename_ = rot_filename;
	  
					plot_correct_skew();
				}

				output_filename_ = stored_filename;
			}
			else 
			{
				Console.Out.WriteLine("");
			}
      
		}


		public void default_write()
		{
			string full_output_filename
				= FilenameManager.full_pathname(output_filename_);
  
			if (print_level_>0) Console.Error.WriteLine("Saving");

			switch (correct_skew_mode_)
			{
				case MObjects.CorrectSkewModeEnum.rotate_page:
				case MObjects.CorrectSkewModeEnum.shear_page:
				case MObjects.CorrectSkewModeEnum.shear_page8:
				case MObjects.CorrectSkewModeEnum.rotate_lines:
				case MObjects.CorrectSkewModeEnum.shear_lines:
				case MObjects.CorrectSkewModeEnum.shear_lines8:
				{
					Bitmap corrected_page = mobjects_.corrected_page();
					if ((object)corrected_page!=null)
					{
						StaffSystemList cstaff_systems=mobjects_.corrected_staff_systems();
						GObjectList cnstaff_systems=mobjects_.corrected_non_staff_systems();
						mobjects_.output_separated_objects(full_output_filename,
							cstaff_systems,cnstaff_systems);
					}
					else
					{
						StaffSystemList staff_systems = mobjects_.staff_systems();
						GObjectList non_staff_systems = mobjects_.non_staff_systems();
						mobjects_.output_separated_objects(full_output_filename,
							staff_systems,non_staff_systems);
					}
					break;
				}

				case MObjects.CorrectSkewModeEnum.rotate_systems:
				case MObjects.CorrectSkewModeEnum.shear_systems:
				case MObjects.CorrectSkewModeEnum.shear_systems8:
				{
					int css_len = mobjects_.corrected_staff_systems().length();
					if (css_len>0)
					{
						StaffSystemList cstaff_systems=mobjects_.corrected_staff_systems();
						GObjectList cnstaff_systems=mobjects_.corrected_non_staff_systems();
						mobjects_.output_separated_objects(full_output_filename,
							cstaff_systems,cnstaff_systems);
					}
					else
					{
						StaffSystemList staff_systems=mobjects_.staff_systems();
						GObjectList non_staff_systems=mobjects_.non_staff_systems();
						mobjects_.output_separated_objects(full_output_filename,
							staff_systems,non_staff_systems);
					}

					break;
				}
			}
		}


		public void customised_write()
		{
			if (FilenameManager.file_type()!=FilenameManager.FileTypeEnum.sep)
			{
				string stored_filename = output_filename_;

				FilenameManager.override_separate_filename("tmp.sep");
				output_filename_ = FilenameManager.separate_filename();

				default_write();

				output_filename_ = stored_filename;
			}
		}



		public void write_pbm()
		{
			string full_output_filename
				= FilenameManager.full_pathname(output_filename_);

			BinaryWriter bwout = FilenameManager.open_output_pipe(full_output_filename);
  
			if (print_level_>0) Console.Error.WriteLine("Saving");


			if (!FilenameManager.output_is_open())
												  {
													  Console.Error.WriteLine("Unable to open pipe for output");
													  return;
												  } 

			if (gen_without_sl_)
			{
				if (print_level_>0) Console.Error.WriteLine("Generating without staff lines");

				BitmapOps bitmap_ops = mobjects_.bitmap_ops();
				Bitmap scratch = bitmap_ops.scratch();
				inlay_without_staff_lines(scratch);
				scratch.fwrite(bwout);
			}
			else if (gen_only_sl_)
			{
				if (print_level_>0) Console.Error.WriteLine("Generating only staff lines");

				BitmapOps bitmap_ops = mobjects_.bitmap_ops();
				Bitmap scratch = bitmap_ops.scratch();
				inlay_staff_lines_removed(scratch);
				scratch.fwrite(bwout);
			}
			else
			{
				FileBitmap current_fbitmap = page_manager_.current_fbitmap();
				current_fbitmap.fwrite(bwout);
			}

			FilenameManager.close_output_pipe();

		}




		public void split_on_title()
		{
			StaffSystemList staff_systems=mobjects_.staff_systems();
			GObjectList non_staff_systems=mobjects_.non_staff_systems();

			if (print_level_>=1)
			{
				Console.Error.WriteLine("Removing false staff systems");
			}
			Bitmap current_bitmap = page_manager_.current_fbitmap();
			mobjects_.filter_staff_systems_with_no_staffs(current_bitmap,
				staff_systems,
				non_staff_systems);

			if (print_level_>=1)
			{
				Console.Error.WriteLine("Grouping titles");
			}
			mobjects_.split_on_title(title_ymul_,title_thresh_,print_level_);

			if (print_level_>=1)
			{
				Console.Error.WriteLine("Subdividing on titles");
			}
			mobjects_.subdivide_staff_systems_on_title(staff_systems,
				non_staff_systems,print_level_);

			if (print_level_>=1)
			{
				Console.Error.WriteLine("Checking again for false staff systems");
			}
			mobjects_.filter_staff_systems_with_no_staffs(current_bitmap,
				staff_systems,
				non_staff_systems);


			string base_sep_filename = output_filename_;

			if (output_filename_ == "")
			{
				output_filename_ = input_filename_;

				int root_break = output_filename_.LastIndexOf('/');
				string only_filename = output_filename_.Substring(root_break+1);
				int dot_break  = output_filename_.IndexOf('.',root_break+2); 
				// 'root_break+2' avoids dot files

				base_sep_filename = output_filename_.Substring(0,dot_break);
			}

			if (do_auto_bcrop_)
			{
				if (print_level_>=1)
				{
					Console.Error.WriteLine("Doing Auto BCrop");
				}
				mobjects_.filter_on_auto_bcrop(staff_systems,non_staff_systems,
					abcrop_ymul_,abcrop_thresh_,auto_bcrop_);
			}
			if (do_bcrop_)
			{
				if (print_level_>=1)
				{
					Console.Error.WriteLine("Doing BCrop");
				}
				mobjects_.filter_on_bcrop(staff_systems,non_staff_systems,
					bcrop_,restrict_y_dim_);
			}
			if (do_rcrop_)
			{
				if (print_level_>=1)
				{
					Console.Error.WriteLine("Doing RCrop");
				}
				mobjects_.filter_on_rcrop(staff_systems,non_staff_systems,
					rcrop_,restrict_x_dim_);
			}
			if (do_lclear_)
			{
				if (print_level_>=1)
				{
					Console.Error.WriteLine("Doing Lclear");
				}
				mobjects_.filter_on_lclear(staff_systems,non_staff_systems,lclear_);
			}

			if (print_level_>=1)
			{
				Console.Error.WriteLine("Saving");
			}
			bool do_crop = do_auto_bcrop_|do_bcrop_|do_lclear_|do_rcrop_;
			mobjects_.output_split_on_title(base_sep_filename,
				staff_systems,non_staff_systems,
				do_crop,
				ref tune_no_, ref part_no_);
  
			Console.Out.Write( tune_no_ + ":" + part_no_);
		}


		public MObjects mobjects() { return mobjects_; }
	}
}
