
using System;
using System.IO;

namespace Espresso.OMR
{
	class Staff : BaseStaff
	{
		protected int xp_lower_thresh_;
		protected int xp_upper_thresh_;
		protected BoolArray1d is_low_trough_;

		protected StaffSeg left_staff_seg_;
		protected StaffSeg right_staff_seg_;

		public enum DirectionEnum { forwards, backwards };
		public enum PassEnum { first_pass, second_pass };


		public Staff()
			: base()
		{
			xp_upper_thresh_ = 0;
			xp_lower_thresh_ = 0;
			is_low_trough_ = new BoolArray1d();

			left_staff_seg_ = null;
			right_staff_seg_ = null;
		}

		public Staff(SubBitmap bitmap, int x_org, int y_org,
			int x_dim, int y_dim)
			: base(bitmap,x_org,y_org,x_dim,y_dim)
		{
			xp_upper_thresh_ = 0;
			xp_lower_thresh_ = 0;
			is_low_trough_ = new BoolArray1d();

			left_staff_seg_ = null;
			right_staff_seg_ = null;
		}

		public Staff(int x_org, int y_org, int x_dim, int y_dim)
			: base(x_org,y_org,x_dim,y_dim)

		{
			xp_upper_thresh_ = 0;
			xp_lower_thresh_ = 0;
			is_low_trough_ = new BoolArray1d();

			left_staff_seg_ = null; 
			right_staff_seg_ = null;
		}

		private StaffSeg staff_seg_consistency_check(StaffSeg staff_seg, 
			double sl_lines_vt,
			double sl_gaps_vt,
			int sl_lines_rt,
			int sl_gaps_rt,
			int verbose_level)
		{
			// compute arverages and standard deviation
			StaffRegionArray1d lines = staff_seg.lines();
	  
			Statistics.IntValDelegate get_staff_region_thickness 
				= new Statistics.IntValDelegate(Statistics.get_staff_region_thickness);

			double lines_variance
				=Statistics.int_variance(lines.data(),get_staff_region_thickness);
			int lines_range
				=Statistics.int_range(lines.data(),get_staff_region_thickness);

			StaffRegionArray1d gaps = staff_seg.gaps();
			double gaps_variance
				=Statistics.int_variance(gaps.data(),get_staff_region_thickness);
			int gaps_range
				=Statistics.int_range(gaps.data(),get_staff_region_thickness);

			if (verbose_level>=5)
			{
				Console.Out.Write("line var = {0:f3} line range = {1:f3} ", lines_variance, lines_range);
				Console.Out.WriteLine("gap var = {0:f3} gap range = {1}",gaps_variance,gaps_range);
			}

			// values decided upon by observation
			// used to be lv=1.5, lr=4, gv=3.5 gr=7;

			if (((lines_variance>=sl_lines_vt) || (lines_range>sl_lines_rt)) ||
				((gaps_variance>=sl_gaps_vt) || (gaps_range>sl_gaps_rt)))
			{
				if (verbose_level>=5)
				{
					Console.Out.Write("Rejecting : ");
					Console.Out.WriteLine("Segment does not match consistency requirments.");
				}
				staff_seg=null;
			} 

			return staff_seg;
		}




		private StaffSeg locate_staff_seg(PassEnum pass,
			int x_start,int x_end,
			int dpi,
			DirectionEnum direction,
			int average_thresh,
			double upper_thresh_mul,
			double lower_thresh_strict_mul,
			double lower_thresh_hyster_mul,
			int verbose_level)
		{
			int max_dpi_len=dpi/6;
			int min_dpi_len=dpi/30;

			bool found_seg=false;
			int min_low_avg = 0;
			int run_start = 0;
			int run_end = 0;

			Projection x_project = staff_area_.x_project();

			int upper_thresh 
				= (pass==PassEnum.first_pass) ? x_project.average() : (6*average_thresh)/4;

			int lower_thresh1 
				= MathExtra.irint(((double)upper_thresh)*lower_thresh_strict_mul);
			int lower_thresh2
				= MathExtra.irint(((double)upper_thresh)*lower_thresh_hyster_mul);

			upper_thresh 
				= MathExtra.irint(((double)upper_thresh)*upper_thresh_mul);

			// store values for later (when graphically displaying info)
			xp_upper_thresh_ = upper_thresh;
			xp_lower_thresh_ = lower_thresh1;

			int p,p_move;

			if (direction==DirectionEnum.forwards)
			{
				p=x_start;
				p_move=1;
			}
			else
			{
				p=x_end;
				p_move=-1;
			}

			int c_range = x_end - x_start+1;
			int c=0;

			while (c<c_range)
			{
				while ((x_project[p]>upper_thresh) || (x_project[p]<lower_thresh1))
				{
					if (c==c_range) break;
					c++; p+=p_move;
				}

				int low_run_len=0;
				int abs_low_run_len=0;
				int low_run_avg=0;

				while ((x_project[p]<=upper_thresh) && (x_project[p]>=lower_thresh2))
				{
					if (c==c_range) break;
					low_run_avg += x_project[p];
					abs_low_run_len++; low_run_len+=p_move;
					c++; p+=p_move;
					if (abs_low_run_len==max_dpi_len) break;
				}

				if (abs_low_run_len>=min_dpi_len) // bottom limit for noise
				{
					low_run_avg=low_run_avg/abs_low_run_len;
					if (!found_seg) //first low run found
					{
						min_low_avg = low_run_avg;
						run_start = p-low_run_len;
						run_end = p;
						found_seg=true;
					}
					else if (low_run_avg<min_low_avg)
					{
						min_low_avg = low_run_avg;
						run_start = p-low_run_len;
						run_end = p;
					}

					int i = run_start;
					while (i!=run_end)
					{
						is_low_trough_[i]=true;
						i += p_move;
					}
					is_low_trough_[i]=true; // safe
				}
			}

			if (direction==DirectionEnum.backwards)
			{
				int temp = run_start;
				run_start=run_end;
				run_end=temp;
			}

			StaffSeg staff_seg;
			if (found_seg)
			{
				staff_seg = new StaffSeg(staff_area_,run_start,run_end);
			}
			else 
			{
				if (verbose_level>=5)
				{
					Console.Out.Write("Rejecting : ");
					Console.Out.WriteLine("Could not find a suitable segment");
				}
				staff_seg=null;
			}

			return staff_seg;
		}


		private StaffSeg locate_staff_seg_insist(PassEnum pass,
			int x_start,int x_end,
			int dpi,
			DirectionEnum direction,
			int must_have_no_lines,
			int average_thresh,
			double upper_thresh_mul,
			double lower_thresh_strict_mul,
			double lower_thresh_hyster_mul,
			int verbose_level)
		{
			int max_dpi_len=dpi/6;
			int min_dpi_len=dpi/30;

			Projection x_project = staff_area_.x_project();

			int upper_thresh 
				= (pass==PassEnum.first_pass) ? x_project.average() : (6*average_thresh)/4;

			int lower_thresh1 
				= MathExtra.irint(((double)upper_thresh)*lower_thresh_strict_mul);
			int lower_thresh2
				= MathExtra.irint(((double)upper_thresh)*lower_thresh_hyster_mul);

			upper_thresh 
				= MathExtra.irint(((double)upper_thresh)*upper_thresh_mul);

			int p,p_move;

			if (direction==DirectionEnum.forwards)
			{
				p=x_start;
				p_move=1;
			}
			else
			{
				p=x_end;
				p_move=-1;
			}

			int c_range = x_end - x_start+1;
			int c=0;

			StaffSeg staff_seg = null;
			int run_start, run_end;

			while (c<c_range)
			{
				while ((x_project[p]>upper_thresh) || (x_project[p]<lower_thresh1))
				{
					if (c==c_range) break;
					c++; p+=p_move;
				}

				int low_run_len=0;
				int abs_low_run_len=0;
				int low_run_avg=0;

				while ((x_project[p]<=upper_thresh) && (x_project[p]>=lower_thresh2))
				{
					if (c==c_range) break;
					low_run_avg += x_project[p];
					abs_low_run_len++; low_run_len+=p_move;
					c++; p+=p_move;
					if (abs_low_run_len==max_dpi_len) break;
				}

				run_start = p-low_run_len;
				run_end = p;

				if (direction==DirectionEnum.backwards)
				{
					int temp = run_start;
					run_start=run_end;
					run_end=temp;
				}

				for (int i=run_start; i<=run_end; i++)
				{
					is_low_trough_[i]=true;
				}

				staff_seg = new StaffSeg(staff_area_,run_start,run_end);
				int no_lines = staff_seg.lines().length();
				if (no_lines != must_have_no_lines)
				{
					staff_seg = null;
				}
				else break;
			}

			if (staff_seg==null)
			{
				if (verbose_level>=5)
				{
					Console.Out.WriteLine("Could not find a insisted segment");
				}
			}

			return staff_seg;
		}



		private StaffSeg locate_staff_seg_insist2(PassEnum pass,
			int x_start,int x_end,
			int dpi,
			DirectionEnum direction,
			int must_have_no_lines,
			int average_thresh,
			double upper_thresh_mul,
			double lower_thresh_strict_mul,
			double lower_thresh_hyster_mul,
			int verbose_level)
		{
			int max_dpi_len=dpi/6;
			int min_dpi_len=dpi/30;

			bool found_seg=false;
			StaffSeg min_staff_seg = null;

			int min_low_avg = 0;

			Projection x_project = staff_area_.x_project();

			int upper_thresh 
				= (pass==PassEnum.first_pass) ? x_project.average() : (6*average_thresh)/4;

			int lower_thresh1 
				= MathExtra.irint(((double)upper_thresh)*lower_thresh_strict_mul);
			int lower_thresh2
				= MathExtra.irint(((double)upper_thresh)*lower_thresh_hyster_mul);

			upper_thresh 
				= MathExtra.irint(((double)upper_thresh)*upper_thresh_mul);

			// store values for later (when graphically displaying info)
			xp_upper_thresh_ = upper_thresh;
			xp_lower_thresh_ = lower_thresh1;

			int p,p_move;

			if (direction==DirectionEnum.forwards)
			{
				p=x_start;
				p_move=1;
			}
			else
			{
				p=x_end;
				p_move=-1;
			}

			int c_range = x_end - x_start+1;
			int c=0;

			while (c<c_range)
			{
				while ((x_project[p]>upper_thresh) || (x_project[p]<lower_thresh1))
				{
					if (c==c_range) break;
					c++; p+=p_move;
				}

				int low_run_len=0;
				int abs_low_run_len=0;
				int low_run_avg=0;

				while ((x_project[p]<=upper_thresh) && (x_project[p]>=lower_thresh2))
				{
					if (c==c_range) break;
					low_run_avg += x_project[p];
					abs_low_run_len++; low_run_len+=p_move;
					c++; p+=p_move;
					if (abs_low_run_len==max_dpi_len) break;
				}

				if (abs_low_run_len>=min_dpi_len) // bottom limit for noise
				{
					low_run_avg=low_run_avg/abs_low_run_len;

					bool poss_run = false;
					int poss_run_start = 0;
					int poss_run_end = 0;

					if (!found_seg) //first poss low run found
					{
						poss_run_start = p-low_run_len;
						poss_run_end = p;
						poss_run = true;
					}
					else if (low_run_avg<min_low_avg)
					{
						poss_run_start = p-low_run_len;
						poss_run_end = p;
						poss_run = true;
					}

					if (poss_run)
					{
						if (direction==DirectionEnum.backwards)
						{
							int temp = poss_run_start;
							poss_run_start=poss_run_end;
							poss_run_end=temp;
						}

						StaffSeg poss_staff_seg 
							= new StaffSeg(staff_area_,poss_run_start,poss_run_end);

						int no_lines = poss_staff_seg.lines().length();
						if (no_lines == must_have_no_lines)
						{		  
							if (found_seg)
							{
								min_staff_seg = null;
							}
							else
							{
								found_seg=true;
							}

							min_staff_seg = poss_staff_seg;
							min_low_avg = low_run_avg;

							for (int i=poss_run_start; i<=poss_run_end; i++)
							{
								is_low_trough_[i]=true;
							}
						}
						else
						{
							poss_staff_seg = null;
						}
					}
				}
			}

			if (!found_seg)
			{
				if (verbose_level>=5)
				{
					Console.Out.Write("Rejecting : ");
					Console.Out.WriteLine("Could not find a suitable segment");
				}
			}

			return min_staff_seg;
		}





		public bool staff_segs_consistency_check(double sl_lines_vt,
			double sl_gaps_vt,
			int sl_lines_rt,
			int sl_gaps_rt,
			int verbose_level)
		{
			left_staff_seg_ 
				= staff_seg_consistency_check(left_staff_seg_,
				sl_lines_vt,sl_gaps_vt,
				sl_lines_rt,sl_gaps_rt,
				verbose_level);
			right_staff_seg_ 
				= staff_seg_consistency_check(right_staff_seg_,
				sl_lines_vt,sl_gaps_vt,
				sl_lines_rt,sl_gaps_rt,
				verbose_level);

			return ((left_staff_seg_!=null) && (right_staff_seg_!=null));
		}


		private double calc_staff_skew(StaffSeg left_seg, StaffSeg right_seg)
		{
			int left_x_mid = (left_seg.x_start()+left_seg.x_start())/2;
			int right_x_mid = (right_seg.x_start()+right_seg.x_start())/2;
			double delta_x = (double)(left_x_mid - right_x_mid);

			StaffRegionArray1d left_lines = left_seg.lines();
			StaffRegionArray1d right_lines = right_seg.lines();

			int no_lines = left_lines.length();
			DoubleArray1d grads = new DoubleArray1d(no_lines);

			for (int i=0; i<no_lines; i++)
			{
				double delta_y // compensate for origin top left corner
					= (double)(right_lines[i].top_edge() - left_lines[i].top_edge());
				grads[i] = delta_y/delta_x;
			}

			Statistics.DoubleValDelegate get_double 
				= new Statistics.DoubleValDelegate(Statistics.get_double);
			double grad_avg
				= Statistics.double_mean(grads.data(),get_double);
			return grad_avg;
		}



		private bool calc_staff_info(bool consistency_check, 
			int verbose_level)
		{
			StaffRegionArray1d left_lines = left_staff_seg_.lines();
			StaffRegionArray1d right_lines =  right_staff_seg_.lines();
			StaffRegionArray1d left_gaps = left_staff_seg_.gaps();
			StaffRegionArray1d right_gaps = right_staff_seg_.gaps();

			int left_num_lines = left_lines.length();
			int right_num_lines = right_lines.length();
			int left_num_gaps = left_gaps.length();
			int right_num_gaps = right_gaps.length();

			Statistics.IntValDelegate get_staff_region_thickness
				= new Statistics.IntValDelegate(Statistics.get_staff_region_thickness);

			double left_lines_median 
				= Statistics.int_median(left_lines.data(),get_staff_region_thickness);

			double right_lines_median 
				= Statistics.int_median(right_lines.data(),get_staff_region_thickness);

			double left_gaps_median 
				= Statistics.int_median(left_gaps.data(),get_staff_region_thickness);

			double right_gaps_median 
				= Statistics.int_median(right_gaps.data(),get_staff_region_thickness);

			double lines_median = (left_lines_median + right_lines_median)/2.0;
			double gaps_median = (left_gaps_median + right_gaps_median)/2.0;

			if (consistency_check)
			{
				if ((left_num_gaps>0) && (right_num_gaps>0))
				{
					if (gaps_median<lines_median*1.7) // used to be 2.0
					{
						if (verbose_level>=5)
						{
							Console.Out.WriteLine("Rejecting: median between lines too large.");
							Console.Out.WriteLine("gap median = " +  gaps_median.ToString() + "must be less than");
							Console.Out.WriteLine("line_median*1.7 = " +  (lines_median*1.7).ToString());
							Console.Out.WriteLine("Consider 'loosing' either top or bottom gap?");
						}
						return false;
					}
				}
			}

			int top_pos = 0;
			int bot_pos = left_lines.length()-1;

			int left_height 
				= left_lines[bot_pos].bot_edge() - left_lines[top_pos].top_edge() +1;
			int right_height 
				= right_lines[bot_pos].bot_edge()-right_lines[top_pos].top_edge() +1;

			int staff_height = (left_height + right_height)/2;

			double staff_rad=calc_staff_skew(left_staff_seg_,right_staff_seg_);

			info_.set_staff_height(staff_height);
			info_.set_gap_height(MathExtra.irint(gaps_median));
			info_.set_line_height(MathExtra.irint(lines_median));
			info_.set_skew(staff_rad);

			return true;
		}

		public bool locate_staff_segs(PassEnum pass, int dpi,
			int average_thresh,
			double upper_thresh_mul,
			double lower_thresh_strict_mul,
			double lower_thresh_hyster_mul,
			int verbose_level,VerboseExtra verb_extra)
		{
			AppGraphics graphics = verb_extra.graphics();
			ivWindow main_window = verb_extra.main_window();

			staff_area_.generate_x_project();
			int x_len = staff_area_.x_project().length();

			is_low_trough_.resize(x_len);
			for (int i=0; i<x_len; i++)
				is_low_trough_[i]=false;

			int lx_search_start = 1*x_len/16; // == x_len/16
			int lx_search_end = 8*x_len/16; // used to be 5, then 7

			if (verbose_level>=6)
				IvLink.app_graphics_plot_sub_bitmap(graphics,staff_area_);


			if (verbose_level>=5)
			{  
				Console.Out.Write("left : ");
			}

			left_staff_seg_
				= locate_staff_seg(pass,lx_search_start,lx_search_end,dpi,DirectionEnum.forwards,average_thresh,
				upper_thresh_mul,
				lower_thresh_strict_mul,
				lower_thresh_hyster_mul,
				verbose_level);

			if ((verbose_level>=6) && (left_staff_seg_!=null))
			{
				int y_top = staff_area_.y_org();
				int y_bot = y_top + staff_area_.y_dim();
				int lx_start = staff_area_.x_org() + left_staff_seg_.x_start();
				int lx_end = staff_area_.x_org() + left_staff_seg_.x_end();

				IvLink.app_graphics_plot_line(graphics,lx_start,y_top,lx_start,y_bot);
				IvLink.app_graphics_plot_line(graphics,lx_end,y_top,lx_end,y_bot);
				IvLink.app_graphics_plot_line(graphics,lx_start,y_top,lx_end,y_top);
				IvLink.app_graphics_plot_line(graphics,lx_start,y_bot,lx_end,y_bot);
			}

			int rx_search_start = 9*x_len/16; // used to be 11
			int rx_search_end = 16*x_len/16;  // used to be 15

			if (verbose_level>=5) 
			{
				Console.Out.Write("right: ");
			}

			right_staff_seg_
				=locate_staff_seg(pass,rx_search_start,rx_search_end,dpi,DirectionEnum.backwards,average_thresh,
				upper_thresh_mul,
				lower_thresh_strict_mul,
				lower_thresh_hyster_mul,verbose_level); 

			if ((verbose_level>=6) && (right_staff_seg_!=null))
			{
				int y_top = staff_area_.y_org();
				int y_bot = y_top + staff_area_.y_dim();
				int rx_start = staff_area_.x_org() + right_staff_seg_.x_start();
				int rx_end = staff_area_.x_org() + right_staff_seg_.x_end();

				IvLink.app_graphics_plot_line(graphics,rx_start,y_top,rx_start,y_bot);
				IvLink.app_graphics_plot_line(graphics,rx_end,y_top,rx_end,y_bot);
				IvLink.app_graphics_plot_line(graphics,rx_start,y_top,rx_end,y_top);
				IvLink.app_graphics_plot_line(graphics,rx_start,y_bot,rx_end,y_bot);
			}

			if (verbose_level>=7)
			{
				// popup to cause delay
				IvLink.app_graphics_flush(graphics);
				BaseIvLink.warning_dialog(main_window,"Showing Staff Segments");
			}

			if ((left_staff_seg_==null) && (right_staff_seg_==null)) return false;
			// check both staff segs have same number of lines

			int left_num_lines = 0;
			int right_num_lines = 0;
			StaffRegionArray1d left_lines = null;
			StaffRegionArray1d right_lines = null;
			StaffRegionArray1d left_gaps = null;
			StaffRegionArray1d right_gaps = null;

			if (left_staff_seg_!=null) 
			{
				left_lines = left_staff_seg_.lines();
				left_gaps = left_staff_seg_.gaps();
				left_num_lines = left_lines.length();
			}

			if (right_staff_seg_!=null)
			{
				right_lines = right_staff_seg_.lines();
				right_gaps = right_staff_seg_.gaps();
				right_num_lines = right_lines.length();
			}

			if (verbose_level>=4)
			{
				Console.Out.Write("no lines (left): " + left_num_lines.ToString() + "  ");
				Console.Out.WriteLine("no lines (right): " + right_num_lines.ToString());
			}

			if (left_num_lines!=right_num_lines) 
			{
				if (verbose_level>=5)
				{
					Console.Out.WriteLine("Segments have different number of lines.");
					Console.Out.WriteLine("Attempting to reselect higher value");
				}
				if ((left_num_lines>right_num_lines) && (right_num_lines>0))
				{
					left_staff_seg_
						=locate_staff_seg_insist2(pass,lx_search_start,lx_search_end,dpi,DirectionEnum.forwards,
						right_num_lines,
						average_thresh,
						upper_thresh_mul,
						lower_thresh_strict_mul,
						lower_thresh_hyster_mul,
						verbose_level); 
					if (left_staff_seg_!=null)
					{
						left_lines = left_staff_seg_.lines();
						left_gaps = left_staff_seg_.gaps();
						left_num_lines = left_lines.length();
					}
				}
				else if ((right_num_lines>left_num_lines) && (left_num_lines>0))
				{
					right_staff_seg_
						=locate_staff_seg_insist2(pass,rx_search_start,rx_search_end,dpi,DirectionEnum.backwards,
						left_num_lines,
						average_thresh,
						upper_thresh_mul,
						lower_thresh_strict_mul,
						lower_thresh_hyster_mul,
						verbose_level);
					if (right_staff_seg_!=null)
					{
						right_lines = right_staff_seg_.lines();
						right_gaps = right_staff_seg_.gaps();
						right_num_lines = right_staff_seg_.lines().length();
					}
				}
			}

			if (left_num_lines!=right_num_lines)
			{
				if (verbose_level>=5)
				{
					Console.Out.WriteLine("Segments STILL have different number of lines.");
					Console.Out.WriteLine("Attempting to reselect lower value");
				}

				if (left_num_lines<right_num_lines)
				{
					left_staff_seg_
						=locate_staff_seg_insist2(pass,lx_search_start,lx_search_end,dpi,DirectionEnum.forwards,
						right_num_lines,
						average_thresh,
						upper_thresh_mul,
						lower_thresh_strict_mul,
						lower_thresh_hyster_mul,
						verbose_level); 
					if (left_staff_seg_!=null)
					{
						left_lines = left_staff_seg_.lines();
						left_gaps = left_staff_seg_.gaps();
						left_num_lines = left_lines.length();
					}
				}
				else if (right_num_lines<left_num_lines)
				{
					right_staff_seg_
						=locate_staff_seg_insist2(pass,rx_search_start,rx_search_end,dpi,DirectionEnum.backwards,
						left_num_lines,
						average_thresh,
						upper_thresh_mul,
						lower_thresh_strict_mul,
						lower_thresh_hyster_mul,
						verbose_level);
					if (right_staff_seg_!=null)
					{
						right_lines = right_staff_seg_.lines();
						right_gaps = right_staff_seg_.gaps();
						right_num_lines = right_staff_seg_.lines().length();
					}
				}
			}

			// if either null, then could not find requested match;
			if ((left_staff_seg_ == null) || (right_staff_seg_ == null))
			{
				if (verbose_level>=5)
				{
					Console.Out.Write("Rejecting : ");
					Console.Out.WriteLine("Unable to find compatible staff segment");
				}
				return false;
			}


			if ((!BaseMObjects.single_staff_lines()) && (left_num_lines<4))
			{
				if (verbose_level>=5)
				{
					Console.Out.Write("Rejecting : ");
					Console.Out.WriteLine("Segments have less than 4 staff lines.");
				}
				return false; 
			}

			no_staff_lines_ = left_num_lines;
			staff_lines_.resize(no_staff_lines_);

			// initialise staff info
			bool consistency_check = true;

			return calc_staff_info(consistency_check,verbose_level);
		}

		public bool locate_staff_segs_insist(PassEnum pass, int dpi,
			int insist_num_lines,
			int average_thresh,
			double upper_thresh_mul,
			double lower_thresh_strict_mul,
			double lower_thresh_hyster_mul,
			int verbose_level,
			VerboseExtra verb_extra)
		{
			AppGraphics graphics = verb_extra.graphics();
			ivWindow main_window = verb_extra.main_window();

			staff_area_.generate_x_project();
			int x_len = staff_area_.x_project().length();

			is_low_trough_.resize(x_len);
			for (int i=0; i<x_len; i++)
				is_low_trough_[i]=false;

			int lx_search_start = 1*x_len/16; // == x_len/16
			int lx_search_end = 8*x_len/16; // used to be 5, then 7

			if (verbose_level>=6)
				IvLink.app_graphics_plot_sub_bitmap(graphics,staff_area_);


			if (verbose_level>=5)  
			{
				Console.Out.Write("left : ");
			}

			left_staff_seg_
				= locate_staff_seg_insist2(pass,lx_search_start,lx_search_end,dpi,DirectionEnum.forwards,
				insist_num_lines,
				average_thresh,
				upper_thresh_mul,
				lower_thresh_strict_mul,
				lower_thresh_hyster_mul,
				verbose_level);

			if ((verbose_level>=6) && (left_staff_seg_!=null))
			{
				int y_top = staff_area_.y_org();
				int y_bot = y_top + staff_area_.y_dim();
				int lx_start = staff_area_.x_org() + left_staff_seg_.x_start();
				int lx_end = staff_area_.x_org() + left_staff_seg_.x_end();

				IvLink.app_graphics_plot_line(graphics,lx_start,y_top,lx_start,y_bot);
				IvLink.app_graphics_plot_line(graphics,lx_end,y_top,lx_end,y_bot);
				IvLink.app_graphics_plot_line(graphics,lx_start,y_top,lx_end,y_top);
				IvLink.app_graphics_plot_line(graphics,lx_start,y_bot,lx_end,y_bot);
			}

			int rx_search_start = 9*x_len/16; // used to be 11
			int rx_search_end = 16*x_len/16 -1;  // used to be 15

			if (verbose_level>=5)
			{
				Console.Out.Write("right: ");
			}

			right_staff_seg_
				=locate_staff_seg_insist2(pass,rx_search_start,rx_search_end,dpi,DirectionEnum.backwards,
				insist_num_lines,
				average_thresh,
				upper_thresh_mul,
				lower_thresh_strict_mul,
				lower_thresh_hyster_mul,verbose_level); 

			if ((verbose_level>=6) && (right_staff_seg_!=null))
			{
				int y_top = staff_area_.y_org();
				int y_bot = y_top + staff_area_.y_dim();
				int rx_start = staff_area_.x_org() + right_staff_seg_.x_start();
				int rx_end = staff_area_.x_org() + right_staff_seg_.x_end();

				IvLink.app_graphics_plot_line(graphics,rx_start,y_top,rx_start,y_bot);
				IvLink.app_graphics_plot_line(graphics,rx_end,y_top,rx_end,y_bot);
				IvLink.app_graphics_plot_line(graphics,rx_start,y_top,rx_end,y_top);
				IvLink.app_graphics_plot_line(graphics,rx_start,y_bot,rx_end,y_bot);
			}

			if (verbose_level>=7)
			{
				// popup to cause delay
				IvLink.app_graphics_flush(graphics);
				BaseIvLink.warning_dialog(main_window,"Showing Staff Segments");
			}

			if ((left_staff_seg_==null) && (right_staff_seg_==null)) return false;
			// check both staff segs have same number of lines

			int left_num_lines = 0;
			int right_num_lines = 0;
			StaffRegionArray1d left_lines = null;
			StaffRegionArray1d right_lines = null;
			StaffRegionArray1d left_gaps = null;
			StaffRegionArray1d right_gaps = null;

			if (left_staff_seg_!=null) 
			{
				left_lines = left_staff_seg_.lines();
				left_gaps = left_staff_seg_.gaps();
				left_num_lines = left_lines.length();
			}

			if (right_staff_seg_!=null)
			{
				right_lines = right_staff_seg_.lines();
				right_gaps = right_staff_seg_.gaps();
				right_num_lines = right_lines.length();
			}

			if (verbose_level>=4)
			{
				Console.Out.Write("no lines (left): " + left_num_lines.ToString() + "  ");
				Console.Out.WriteLine("no lines (right): " + right_num_lines.ToString());
			}

			if (left_num_lines!=right_num_lines) 
			{
				if (verbose_level>=5)
				{
					Console.Out.WriteLine("Segments have different number of lines.");
					Console.Out.WriteLine("Attempting to reselect incorrect staff seg +tol");
				}
				if ((left_num_lines!=insist_num_lines) && 
					(right_num_lines==insist_num_lines))
				{
					int thresh_tol = dpi/50;

					left_staff_seg_
						=locate_staff_seg_insist2(pass,lx_search_start,lx_search_end,dpi,DirectionEnum.forwards,
						right_num_lines,
						average_thresh,
						upper_thresh_mul+thresh_tol,
						lower_thresh_strict_mul-thresh_tol,
						lower_thresh_hyster_mul-thresh_tol,
						verbose_level); 
					if (left_staff_seg_!=null)
					{
						left_lines = left_staff_seg_.lines();
						left_gaps = left_staff_seg_.gaps();
						left_num_lines = left_lines.length();
					}
				}
				else if ((right_num_lines!=insist_num_lines) && 
					(left_num_lines==insist_num_lines))
				{
					int thresh_tol = dpi/50;

					right_staff_seg_
						=locate_staff_seg_insist2(pass,rx_search_start,rx_search_end,dpi,DirectionEnum.backwards,
						left_num_lines,
						average_thresh,
						upper_thresh_mul+thresh_tol,
						lower_thresh_strict_mul-thresh_tol,
						lower_thresh_hyster_mul-thresh_tol,
						verbose_level);
					if (right_staff_seg_!=null)
					{
						right_lines = right_staff_seg_.lines();
						right_gaps = right_staff_seg_.gaps();
						right_num_lines = right_staff_seg_.lines().length();
					}
				}
			}

			if (left_num_lines!=right_num_lines)
			{
				if (verbose_level>=5)
				{
					Console.Out.WriteLine("Segments STILL have different number of lines.");
					Console.Out.WriteLine("Attempting to reselect incorrect seg -tol");
				}

				if (left_num_lines<right_num_lines)
				{
					int thresh_tol = dpi/50;

					left_staff_seg_
						=locate_staff_seg_insist2(pass,lx_search_start,lx_search_end,dpi,DirectionEnum.forwards,
						right_num_lines,
						average_thresh,
						upper_thresh_mul-thresh_tol,
						lower_thresh_strict_mul+thresh_tol,
						lower_thresh_hyster_mul+thresh_tol,
						verbose_level); 
					if (left_staff_seg_!=null)
					{
						left_lines = left_staff_seg_.lines();
						left_gaps = left_staff_seg_.gaps();
						left_num_lines = left_lines.length();
					}
				}
				else if (right_num_lines<left_num_lines)
				{
					int thresh_tol = dpi/50;

					right_staff_seg_
						=locate_staff_seg_insist2(pass,rx_search_start,rx_search_end,dpi,DirectionEnum.backwards,
						left_num_lines,
						average_thresh,
						upper_thresh_mul-thresh_tol,
						lower_thresh_strict_mul+thresh_tol,
						lower_thresh_hyster_mul+thresh_tol,
						verbose_level);
					if (right_staff_seg_!=null)
					{
						right_lines = right_staff_seg_.lines();
						right_gaps = right_staff_seg_.gaps();
						right_num_lines = right_staff_seg_.lines().length();
					}
				}
			}

			// if either null, then could not find requested match;
			if ((left_staff_seg_ == null) || (right_staff_seg_ == null))
			{
				if (verbose_level>=5)
				{
					Console.Out.Write("Rejecting : ");
					Console.Out.WriteLine("Unable to find compatible staff segment");
				}
				return false;
			}


			if ((!BaseMObjects.single_staff_lines()) && (left_num_lines<4))
			{
				if (verbose_level>=5)
				{
					Console.Out.Write("Rejecting : ");
					Console.Out.WriteLine("Segments have less than 4 staff lines.");
				}
				return false; 
			}

			no_staff_lines_ = left_num_lines;
			staff_lines_.resize(no_staff_lines_);

			// initialise staff info
			bool consistency_check = true;

			return calc_staff_info(consistency_check,verbose_level);
		}




		public void calc_def_staff_lines()
		{
			StaffRegionArray1d left_lines = left_staff_seg_.lines();
			int left_num_lines = left_lines.length();

			no_staff_lines_ = left_num_lines;
			staff_lines_.resize(no_staff_lines_);

			calc_staff_info(false,0);
		}


		public void delete_left_staff_seg() 
		{ 
			left_staff_seg_=null; 
		}

		public void delete_right_staff_seg()
		{ 
			right_staff_seg_=null; 
		}


		public int xp_lower_thresh() { return xp_lower_thresh_; }
		public int xp_upper_thresh() { return xp_upper_thresh_; }

		public BoolArray1d is_low_trough() { return is_low_trough_; }

		public bool has_left_staff_seg()
		{ return left_staff_seg_!=null; }
		public bool has_right_staff_seg()
		{ return right_staff_seg_!=null; }

		public void set_left_staff_seg(StaffSeg left_sseg)
		{ left_staff_seg_ = left_sseg; }
		public void set_right_staff_seg(StaffSeg right_sseg)
		{ right_staff_seg_ = right_sseg; }

		public StaffSeg left_staff_seg()
		{ return left_staff_seg_; } // these were actually return copies in the C++ code, possibly unnecessary *****
		public StaffSeg right_staff_seg()
		{ return right_staff_seg_; } // these were actually return copies in the C++ code, possibly unnecessary *****



		/*---------------------------------------------------------------------------*/

		protected override void write_staff_segs(BinaryWriter bwout)
		  {
		    IOSupport.write_line(bwout,"left_staff_seg :");
		    if (left_staff_seg_!=null)
		      {
			left_staff_seg_.write(bwout);
		      }
		    else
		      {
			StaffSeg.write_empty_staff_seg(bwout);
		      }

		    IOSupport.write_line(bwout,"right_staff_seg :");
		    if (right_staff_seg_!=null)
		      {
			right_staff_seg_.write(bwout);
		      }
		    else
		      {
			StaffSeg.write_empty_staff_seg(bwout);
		      }
		  }


		protected override bool read_staff_segs(BinaryReader brin)
		  {
		    int sa_x_org = staff_area_.x_org();
		    int sa_y_org = staff_area_.y_org();

		    left_staff_seg_ = new StaffSeg(sa_x_org,sa_y_org);

		    if (!IOSupport.read_token(brin,"left_staff_seg")) return false;
		    if (!IOSupport.read_token(brin,":")) return false;
		    if (!left_staff_seg_.read_staff_seg(brin)) return false;

		    if ((left_staff_seg_.global_x_start()==0) &&
			(left_staff_seg_.global_x_end()==0))
		      {
			// null entry
			// only possible if dsep file
			left_staff_seg_=null;
		      }
		    else
		      {
			GObject sseg_gobject = left_staff_seg_.gobject();

			int sseg_x_start = left_staff_seg_.x_start();
			int sseg_x_end = left_staff_seg_.x_end();
			int sseg_y_top = staff_area_.y_top();

			int sseg_x_dim = sseg_x_end - sseg_x_start +1;
			int sseg_y_dim = staff_area_.y_dim();

			sseg_gobject.set_x_off(sseg_x_start%8);
			sseg_gobject.resize(sseg_x_dim,sseg_y_dim);

			sseg_gobject.extract(staff_area_,sseg_x_start,0,sseg_x_dim,sseg_y_dim);

			sseg_gobject.set_x_org(sseg_x_start);
      
			left_staff_seg_.generate_data();
		      }


		    right_staff_seg_ = new StaffSeg(sa_x_org,sa_y_org);

		    if (!IOSupport.read_token(brin,"right_staff_seg")) return false;
		    if (!IOSupport.read_token(brin,":")) return false;
		    if (!right_staff_seg_.read_staff_seg(brin)) return false;

		    if ((right_staff_seg_.global_x_start()==0) &&
			(right_staff_seg_.global_x_end()==0))
		      {
			// null entry
			// only possible if dsep file
			right_staff_seg_=null;
		      }
		    else
		      {
			GObject sseg_gobject = right_staff_seg_.gobject();

			int sseg_x_start = right_staff_seg_.x_start();
			int sseg_x_end = right_staff_seg_.x_end();
			int sseg_y_top = staff_area_.y_top();

			int sseg_x_dim = sseg_x_end - sseg_x_start +1;
			int sseg_y_dim = staff_area_.y_dim();

			sseg_gobject.set_x_off(sseg_x_start%8);
			sseg_gobject.resize(sseg_x_dim,sseg_y_dim);

			sseg_gobject.extract(staff_area_,sseg_x_start,0,sseg_x_dim,sseg_y_dim);

			sseg_gobject.set_x_org(sseg_x_start);

			right_staff_seg_.generate_data();
		      }

		    return true;
		  }

		public override bool read_staff(BinaryReader brin, int y_offset)
		  {
		    // staff assumed to be empty

		    if (!read_staff_area(brin,y_offset)) return false;
		    if (!read_staff_segs(brin)) return false;
		    if (!read_staff_lines(brin,y_offset)) return false;

		    if (!info_.read(brin)) return false;

		    return true;
		  }

	}
}
