
using System;
using System.IO;

namespace Espresso.OMR
{

	class StaffSystem : BaseStaffSystem
	{
		private GObject without_staffs_;
		private StaffList potential_staffs_;
		private StaffList second_pass_staffs_;
    
		private GObjectList processed_staff_system_objects_;
    
		private int peaks_no_;
		private int diff_no_;
		private int staff_jump_;
    
		private IntArray1d peak_;
		private IntArray1d peak_pos_;
		private IntArray1d peak_pos_diff_;
		private IntArray1d sorted_peak_pos_diff_;
    
		public StaffSystem()
			: base()
		{
			without_staffs_ = new GObject();
			potential_staffs_ = new StaffList();
			second_pass_staffs_ = new StaffList();

			processed_staff_system_objects_ = new GObjectList();

			peak_ = new IntArray1d();
			peak_pos_ = new IntArray1d();
			peak_pos_diff_ = new IntArray1d();
			sorted_peak_pos_diff_ = new IntArray1d();

			peaks_no_ = 0;
			diff_no_ = 0;
			staff_jump_ = 0;
		}

		public StaffSystem(GObject gobject)
			: base(gobject)
		{
			without_staffs_ = new GObject((SubBitmap)gobject,0,0,
				gobject.x_dim(),gobject.y_dim());
			potential_staffs_ = new StaffList();
			second_pass_staffs_ = new StaffList();

			processed_staff_system_objects_ = new GObjectList();

			peak_ = new IntArray1d();
			peak_pos_ = new IntArray1d();
			peak_pos_diff_ = new IntArray1d();
			sorted_peak_pos_diff_ = new IntArray1d();

			peaks_no_ = 0;
			diff_no_ = 0;
			staff_jump_ = 0;

			without_staffs_.add_to_x_org(gobject.x_org());
			without_staffs_.add_to_y_org(gobject.y_org());
		}


		public void generate_peaks()
		{
			int y;
	
			gobject_.generate_y_project();
			Projection y_project = gobject_.y_project();
			int max_value = y_project.max_value();
			int thresh = 9*max_value/20; // set empirically 

			int y_dim = gobject_.y_dim();
			peak_.resize(y_dim);

			for (y=0; y<y_dim; y++)
				peak_[y]=0;

			y=1;
			while (y<y_dim-1)
			{
				if (y_project[y]>thresh)
				{
					y = y_project.find_max_tp(y,Projection.DirectionEnum.forwards);
					if (y_project[y]>thresh) // rule out last request if not max tp
					{
						peak_[y]=y_project[y];
						peaks_no_++;
						while (y_project[y]>thresh)
						{
							if (y==y_dim-1) break;
							y++;
						}
					}
				}
				y++;
			}

			diff_no_=peaks_no_-1;
		}


		public void generate_sorted_peak_diff(int verbose_level)
		{
			if (peaks_no_>0)
			{
				peak_pos_.resize(peaks_no_);
				peak_pos_diff_.resize(diff_no_);
				sorted_peak_pos_diff_.resize(diff_no_);
      
				int y_dim = gobject_.y_dim();
				int p=0;
				for (int y=0; y<y_dim; y++)
					if (peak_[y]>0) peak_pos_[p++]=y;
      
				for (p=0; p<diff_no_; p++)
				{
					peak_pos_diff_[p]=peak_pos_[p+1]-peak_pos_[p];
					sorted_peak_pos_diff_[p]=peak_pos_diff_[p];
				}
		
				Array.Sort(sorted_peak_pos_diff_.data()); 

				if (verbose_level>=8)
				{
					Console.Out.Write("[ ");
					for (int i=0; i<diff_no_; i++)
					{
						Console.Out.Write(sorted_peak_pos_diff_[i].ToString());
						if (i<(diff_no_-1))
						{
							Console.Out.Write(", ");
						}

					}
					Console.Out.WriteLine(" ]");
				}
			}
		}

		public void find_staff_jump(int sg_min_gap,
			int sa_prop_numer, 
			int sa_prop_denom)
		{
			int len = sorted_peak_pos_diff_.length();

			int y_start = 0;
			if (sg_min_gap==0)
			{
				if (sa_prop_denom!=0)
				{
					// assign with safety check
					y_start = (len > 2) ? (sa_prop_numer*len)/sa_prop_denom : 2;
				}
			}
			else
			{
				while (y_start<len)
				{
					if (sorted_peak_pos_diff_[y_start]>sg_min_gap) 
					{
						break;
					}
					y_start++;
				}
			}
  
			for (int y=y_start; y<len; y++)
			{
				// compare vals separated by 2.1,1.9
				// # used to be 2,2 then 1.8,1.8 then 1.9,1.8 then 2.1 1.9
				if ((sorted_peak_pos_diff_[y]
					>MathExtra.irint(2.1*(double)sorted_peak_pos_diff_[y-2])) ||
					(sorted_peak_pos_diff_[y]
					>MathExtra.irint(1.9*(double)sorted_peak_pos_diff_[y-1])))
				{
					staff_jump_ = sorted_peak_pos_diff_[y];
					break;
				}
			}
		}

		private void calc_staff_info(int num_staffs,int verbose_level)
		{
			IntArray1d staff_heights = new IntArray1d(num_staffs);
			IntArray1d gap_heights = new IntArray1d(num_staffs);
			IntArray1d line_heights = new IntArray1d(num_staffs);
			DoubleArray1d skews = new DoubleArray1d(num_staffs);

			int n=0;
			StaffNode st_current = staffs_.head();
			while (st_current!=null)
			{
				StaffInfo info = st_current.val().info();

				staff_heights[n]=info.staff_height();
				gap_heights[n]=info.gap_height();
				line_heights[n]=info.line_height();
				skews[n]=info.skew();

				if (verbose_level>=3)
				{
					Console.Out.Write("staff_system: ");
					Console.Out.Write("staff height = " + staff_heights[n].ToString() + " ");
					Console.Out.Write("gap height = "   + gap_heights[n].ToString() + " ");
					Console.Out.Write("line height = "  + line_heights[n].ToString() + " ");

					Console.Out.Write("skew = ");
					if (!MathExtra.sign_bit(skews[n]))
					{
						Console.Out.Write("+");
					}

					Console.Out.WriteLine("{0%f4}",skews[n]);
	 
				}

				n++;
				st_current = staffs_.next();
			} 

			Statistics.IntValDelegate get_int
				= new Statistics.IntValDelegate(Statistics.get_int);
			Statistics.DoubleValDelegate get_double
				= new Statistics.DoubleValDelegate(Statistics.get_double);

			double glob_staff_height 
				= Statistics.int_median(staff_heights.data(),get_int);

			double glob_gap_height 
				= Statistics.int_median(gap_heights.data(),get_int);

			double glob_line_height 
				= Statistics.int_median(line_heights.data(),get_int);

			double glob_skew 
				= Statistics.double_mean(skews.data(),get_double);

			staff_info_.set_staff_height(MathExtra.irint(glob_staff_height));
			staff_info_.set_gap_height(MathExtra.irint(glob_gap_height));
			staff_info_.set_line_height(MathExtra.irint(glob_line_height));
			staff_info_.set_skew(glob_skew);
		}

		public bool generate_staff_areas(Bitmap current_page,
			int sg_min_gap,
			int sa_prop_numer, 
			int sa_prop_denom,
			int verbose_level,
			VerboseExtra verb_extra,
			ComputingCircle computing, 
			int computed_base,
			int computing_range)
		{
			if (peaks_no_==0) return false;

			int ss_x_org = gobject_.x_org();
			int ss_y_org = gobject_.y_org();
			int x_start = 0;
			int go_x_dim = gobject_.x_dim();
			int go_y_dim = gobject_.y_dim();

			Projection y_project = gobject_.y_project();
			int py_dim = y_project.length();
			int max_value = y_project.max_value();
			int thresh = max_value/2; // set empircally

			int computing_step1 = 1*computing_range/10;
			int computing_step2 = 9*computing_range/10;

			find_staff_jump(sg_min_gap,sa_prop_numer,sa_prop_denom);

			int computed=computed_base+computing_step1;
			computing.computed_so_far(computed);

			if (staff_jump_==0)
			{
				if (verbose_level>0)
				{
					Console.Out.WriteLine("staff system has only one staff");
				}

				// staff system has only one staff
				int y_pos=peak_pos_[0];
				while (y_project[y_pos]>thresh/3) // used to be 3
				{
					if (y_pos==0) break;
					y_pos--;
				}
				int y_start=y_project.find_min_tp(y_pos,Projection.DirectionEnum.backwards);

				y_pos=peak_pos_[diff_no_];
				while (y_project[y_pos]>thresh/3) // used to be 3
				{
					if (y_pos==py_dim-1) break;
					y_pos++;
				}
				int y_end=y_project.find_min_tp(y_pos,Projection.DirectionEnum.forwards);
				int y_dim=y_end-y_start;

				// extend y direction of staff area by 5%
				int extend_y = (5*y_dim)/100;
				int ey_start = y_start-extend_y;
				int ey_end = y_end+extend_y;
				if (ey_start<0) ey_start=0;
				if (ey_end>=go_y_dim) ey_end=go_y_dim-1;
				int ey_dim = ey_end - ey_start;

				Staff new_staff = new Staff(gobject_,x_start,ey_start,go_x_dim,ey_dim);
				new_staff.reset_org(ss_x_org,ss_y_org);

				potential_staffs_.append(new_staff);

				computed+=computing_step2;
				computing.computed_so_far(computed);
			}
			else
			{
				int p=0;
				int computed_base1 = computed_base+computing_step1;

				while (p<diff_no_)
				{
					computed=computed_base1+(p*computing_step2/diff_no_);
					computing.computed_so_far(computed);

					int y_pos=peak_pos_[p];
					while (y_project[y_pos]>thresh/3)
					{
						if (y_pos==0) break;
						y_pos--;
					}

					int y_start=y_project.find_min_tp(y_pos,Projection.DirectionEnum.backwards);

					int no_violations = 0, no_gaps = 0;
					while (p<diff_no_)
					{
						if (peak_pos_diff_[p]>=staff_jump_) 
						{
							// redundent
							if (no_gaps>=0) break; // used to be 4, then 1 // ****
							no_gaps=0; no_violations++;
						}
						p++; no_gaps++;
						computed=computed_base1+(p*computing_step2/diff_no_);
						computing.computed_so_far(computed);
					}

					y_pos=peak_pos_[p]; // max(p)==diff_no_==peaks_no_-1
					while (y_project[y_pos]>thresh/3)
					{
						if (y_pos==py_dim-1) break;
						y_pos++;
					}
					int y_end=y_project.find_min_tp(y_pos,Projection.DirectionEnum.forwards);
					int y_dim=y_end-y_start;

					// extend y direction of staff area by 5%
					int extend_y = (5*y_dim)/100;
					int ey_start = y_start-extend_y;
					int ey_end = y_end+extend_y;
					if (ey_start<0) ey_start=0;
					if (ey_end>=go_y_dim) ey_end=go_y_dim-1;
					int ey_dim = ey_end - ey_start;

					Staff new_staff
						= new Staff(gobject_,x_start,ey_start,go_x_dim,ey_dim);

					new_staff.reset_org(ss_x_org,ss_y_org);

					potential_staffs_.append(new_staff);

					p++;
				}
			}

			computed=computed_base+computing_range;
			computing.computed_so_far(computed);

			return potential_staffs_.length()>0;
		}


		public void generate_staff_segs_first_pass(Bitmap current_page,
			double upper_thresh_mul,
			double lower_thresh_strict_mul,
			double lower_thresh_hyster_mul,
			int verbose_level,
			VerboseExtra verb_extra,
			ComputingCircle computing, 
			int computed_base,
			int computing_range, int dpi)
		{
			StaffList stl = potential_staffs_;
			int no_pot_staff_areas = stl.length();

			int computing_step 
				= (no_pot_staff_areas>0) ? computing_range/no_pot_staff_areas
				: computing_range;

			int computed=computed_base;
			computing.computed_so_far(computed);

			StaffNode current_st = stl.head();
			while (current_st!=null)
			{
				Staff staff = current_st.val(); // this should be a copy! // ****

				int insist_sl_num = BaseMObjects.insist_sl_num();
				if (insist_sl_num>0)
				{
					if (staff.locate_staff_segs_insist(Staff.PassEnum.first_pass,dpi,
						insist_sl_num,0,
						upper_thresh_mul,
						lower_thresh_strict_mul,
						lower_thresh_hyster_mul,
						verbose_level,verb_extra))
					{
						staffs_.append(staff); // resonsibility passed over
					}
					else 
					{
						// store for second pass of staff locate algorithm
						second_pass_staffs_.append(staff);
					}
				}
				else
				{
					if (staff.locate_staff_segs(Staff.PassEnum.first_pass,dpi,0,
						upper_thresh_mul,
						lower_thresh_strict_mul,
						lower_thresh_hyster_mul,
						verbose_level,verb_extra))
					{
						staffs_.append(staff); // resonsibility passed over
					}
					else 
					{
						// store for second pass of staff locate algorithm
						second_pass_staffs_.append(staff);
					}
				}

				computed+=computing_step;
				computing.computed_so_far(computed);

				current_st = stl.next();
			}

			computed=computed_base+computing_range;
			computing.computed_so_far(computed);
		}



		public void generate_staff_lines(Bitmap current_page,
			double sl_lines_vt,
			double sl_gaps_vt,
			int sl_lines_rt,
			int sl_gaps_rt,
			int verbose_level,
			VerboseExtra verb_extra,
			ComputingCircle computing, 
			int computed_base,
			int computing_range)
		{
			StaffList stl = staffs_;
			int no_pot_staff_areas = stl.length();
  
			int computing_step 
				= (no_pot_staff_areas>0) ? computing_range/no_pot_staff_areas
				: computing_range;
			int computed=computed_base;
			computing.computed_so_far(computed);

			int no_staff_areas=0;

			StaffNode current_st = stl.head();
			while (current_st!=null)
			{
				Staff staff = current_st.val();

				if (staff.staff_segs_consistency_check(sl_lines_vt,sl_gaps_vt,
					sl_lines_rt,sl_gaps_rt,
					verbose_level))
				{
					current_st = stl.next();
					no_staff_areas++;
				}
				else 
				{
					// store for second pass of staff locate algorithm
					current_st = stl.detach_forward();
					second_pass_staffs_.append(staff);
				}

				computed+=computing_step;
				computing.computed_so_far(computed);
			}

			if (no_staff_areas>0) calc_staff_info(no_staff_areas,verbose_level);

			computed=computed_base+computing_range;
			computing.computed_so_far(computed);

		}


		public void generate_staff_segs_second_pass(Bitmap current_page,
			double upper_thresh_mul,
			double lower_thresh_strict_mul,
			double lower_thresh_hyster_mul,
			int verbose_level,
			VerboseExtra verb_extra, ComputingCircle cir, 
			int cbase, int range, int dpi,
			StaffInfo global_staff_info,
			int med_no_staff_lines)
		{
			int average_thresh 
				= global_staff_info.line_height()*med_no_staff_lines;
  
			StaffNode current_sps = second_pass_staffs_.head();
			while (current_sps!=null)
			{
				Staff sp_staff = current_sps.val();
				int insist_sl_num = BaseMObjects.insist_sl_num();
      
				bool located_staff_segs
					= (insist_sl_num>0)
					? sp_staff.locate_staff_segs_insist(Staff.PassEnum.second_pass,dpi,
					insist_sl_num,
					average_thresh,
					upper_thresh_mul,
					lower_thresh_strict_mul,
					lower_thresh_hyster_mul,
					verbose_level,verb_extra)
					: sp_staff.locate_staff_segs(Staff.PassEnum.second_pass,dpi,
					average_thresh,
					upper_thresh_mul,
					lower_thresh_strict_mul,
					lower_thresh_hyster_mul,
					verbose_level,verb_extra);

				if (located_staff_segs)
				{
					// insert in order
					GObject sp_staff_area = sp_staff.staff_area();
					int sp_x_org = sp_staff_area.x_org();
					int sp_y_org = sp_staff_area.y_org();

					StaffNode current_fps = staffs_.head();
					while (current_fps!=null)
					{
						Staff fp_staff = current_fps.val();
						GObject fp_staff_area = fp_staff.staff_area();
						int fp_x_org = fp_staff_area.x_org();
						int fp_y_org = fp_staff_area.y_org();

						if ((sp_y_org < fp_y_org) ||
							((sp_y_org == fp_y_org) && (sp_x_org < fp_x_org)))
						{
							staffs_.insert_before(sp_staff); // resonsibility passed over
							break;
						}
	      
						current_fps = staffs_.next();
					}

					current_sps=second_pass_staffs_.detach_forward();
				}
				else 
				{
					current_sps = second_pass_staffs_.delete_forward();
				}
			}
		}


		public StaffList potential_staffs() 
		{ return potential_staffs_; }
    
		public StaffList second_pass_staffs() 
		{ return second_pass_staffs_; }
    
		public GObject without_staffs() { return without_staffs_; }
    
		public IntArray1d peaks() { return peak_; }
		public int peaks_no() { return peaks_no_; }
    
		public GObjectList processed_staff_system_objects() 
		{ return processed_staff_system_objects_; }


		/*---------------------------------------------------------------------------*/
 
		protected override void write_staff_system_objects(BinaryWriter bwout, 
								 int y_offset)
		  {
		    GObjectList ss_obs = processed_staff_system_objects_; // shorthand
		    int no_ss_obs = ss_obs.length();
		    IOSupport.write_line(bwout,"no_staff_system_objects = " + no_ss_obs.ToString());

		    for (GObjectNode ss_ob=ss_obs.head(); ss_ob!=null; ss_ob=ss_obs.next())
		      {
			SubBitmap sub_bitmap = (SubBitmap)(ss_ob.val());
			sub_bitmap.add_to_y_org(-y_offset);

			sub_bitmap.write(bwout);
			sub_bitmap.add_to_y_org(+y_offset);
		      }
		  }

		protected override bool read_staffs(BinaryReader brin, int y_offset)
		  {
		    int no_staffs;
		    if (!IOSupport.read_token(brin,"no_staffs")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out no_staffs)) return false;

		    for (int i=0; i<no_staffs; i++)
		      {
			Staff staff = new Staff();
			if (!staff.read_staff(brin,y_offset)) return false;
			potential_staffs_.append(staff); // responisbilty passed over to list
			staffs_.append(staff); // this should be a copy! // ****
		      }

		    return true;
		  }

		protected override bool read_staff_system_objects(BinaryReader brin, int y_offset)
		  {
		    int no_ss_obs;
		    if (!IOSupport.read_token(brin,"no_staff_system_objects")) return false;
		    if (!IOSupport.read_token(brin,"=")) return false;
		    if (!IOSupport.read_cardinal(brin,out no_ss_obs)) return false;

		    GObjectList ss_obs = staff_system_objects_; // shorthand
		    GObjectList pss_obs = processed_staff_system_objects_; // shorthand
		    for (int i=0; i<no_ss_obs; i++)
		      {
			GObject ss_ob = new GObject();
			if (!ss_ob.read(brin)) return false;
			ss_ob.add_to_y_org(y_offset);
			ss_obs.append(ss_ob); // responisbilty passed over to list

			SubBitmap ss_ob_subbitmap = ss_ob;
			GObject pss_ob 
			  = new GObject(ss_ob_subbitmap,0,0,ss_ob.x_dim(),ss_ob.y_dim());
			pss_ob.set_x_org(ss_ob.x_org());
			pss_ob.set_y_org(ss_ob.y_org());

			pss_obs.append(pss_ob); // responisbilty passed over to list
		      }

		    return true;
		  }


		public override bool read_staff_system(BinaryReader brin, int y_offset)
		  {
		    // staff_system assumed to be empty

		    if (!read_gobject(brin,y_offset)) return false;
		    if (!read_staffs(brin,y_offset)) return false;
		    if (!read_staff_system_objects(brin,y_offset)) return false;
		    if (!staff_info_.read(brin)) return false;

		    return true;
		  }



	}
}
