using System;
using System.Collections.Generic;
using System.Text;
using Espresso;

namespace Espresso.Utils
{
    /// <summary>
    /// The class that updates a field of type T on an object automatically as some preference changes
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public class FieldUpdater<T>
    {
        /// <summary>
        /// The class that contains the field that will be updated
        /// </summary>
        protected IFieldContainer<T> container;

        /// <summary>
        /// The builder that is used to build the objects once they are changed
        /// </summary>
        protected IBuilder<T> builder;

        /// <summary>
        /// The name of the preference that should be watched.
        /// </summary>
        protected string preferenceToWatch;

        /// <summary>
        /// Constructs a new FieldUpdater. Subscribes to preferences and watches the updates
        /// on <see cref="preferenceToWatch"/>. If an update is triggered, calls <see cref="Update()"/>.
        /// </summary>
        /// <param name="preferenceToWatch">The name of the preference that will be watched.</param>
        /// <param name="container">The object that contains the field that will be updated</param>
        /// <param name="builder">The builder that will be used to build the new value of the field</param>
        public FieldUpdater(string preferenceToWatch, IFieldContainer<T> container, IBuilder<T> builder)
        {
            this.container = container;
            this.builder = builder;
            this.preferenceToWatch = preferenceToWatch;

            Preferences.Instance.PreferencesChanged += delegate(PreferencesChangedEventArgs args)
            {
                if (args.Key.Equals(preferenceToWatch))
                {
                    try
                    {
                        string newDesc = args.newValue as string;

                        Update(newDesc);
                    }
                    catch (InvalidCastException)
                    {
                        //TODO Do some useful logging here
                    }
                }
            };

            Update((string)Preferences.Instance.GetProperty(preferenceToWatch));
        }

        /// <summary>
        /// Updates the field in the container with an 
        /// object built by builder based on the new description.
        /// </summary>
        /// <param name="newDescription"></param>
        public void Update(string newDescription)
        {
            T newValue = builder.Build(newDescription);

            container.SetField(newValue);
        }
    }
}
