using System;
using System.Collections.Generic;
using System.Text;
using System.Xml.Serialization;
using System.Xml;
using System.IO;

namespace Espresso.Utils
{
    /// <summary>
    /// A class that stores various user-defined settings
    /// </summary>
    public class Preferences
    {
        private static Preferences m_Instance;

        /// <summary>
        /// The singleton instance
        /// </summary>
        public static Preferences Instance
        {
            get
            {
                if (m_Instance == null)
                    m_Instance = new Preferences();

                return m_Instance;
            }
        }

        [XmlElement("dictionary")]
        private List<Setting> m_List;

        /// <summary>
        /// The handler that is used for handling of the <see cref="PreferencesChanged"/> event.
        /// </summary>
        /// <param name="args"></param>
        public delegate void PreferencesChangedEventHandler(PreferencesChangedEventArgs args);

        /// <summary>
        /// Fired when a preference is changed.
        /// </summary>
        public event PreferencesChangedEventHandler PreferencesChanged;

        private Preferences() { }

        private Setting FindSetting(string key, IEnumerable<Setting> list)
        {
            foreach (Setting s in list)
                if (s.Key.Equals(key))
                    return s;

            return null;
        }

        /// <summary>
        /// Sets the property <see cref="key"/> to the value of <see cref="value"/>.
        /// If the property is not in the dictionary, creates one.
        /// </summary>
        /// <param name="key"></param>
        /// <param name="value"></param>
        public void SetProperty(string key, object value)
        {
            if (m_List == null)
                Load();

            Setting s = FindSetting(key, m_List);
            
            if (s == null)
            {
                s = new Setting(key, value);
                m_List.Add(s);
            }
            else
            {
                object oldValue = s.Value;
                s.Value = value;

                if (!(oldValue.Equals(value) || oldValue == value))
                    if (PreferencesChanged != null)
                        PreferencesChanged(new PreferencesChangedEventArgs(key, value));
            }
        }

        /// <summary>
        /// Gets a property by the key. If the dictionary is not set, it is loaded
        /// from the hard disk.
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public object GetProperty(string key)
        {
            if (m_List == null)
                Load();

            Setting s = FindSetting(key, m_List);

            if (s == null)
            {
                //If the preference has not been found, look it up in the defaults.
                Setting def = FindSetting(key, Defaults);

                if (def == null)
                {
                    //If it is not even in defaults, just return null
                    return null;
                }
                else
                {
                    //If it is in the defaults, add it to out set
                    SetProperty(key, def.Value);

                    return def.Value;
                }
            }

            return s.Value;
        }

        /// <summary>
        /// Writes the preferences to the disk.
        /// </summary>
        public void Save()
        {
            if (m_List == null)
                return;

            string path = EspressoUtils.PathToPreferenceFileStorageFolder + "preferences.xml";

            XmlSerializer s = new XmlSerializer(typeof(List<Setting>));

            TextWriter w = new StreamWriter(path);
            s.Serialize(w, m_List);
            w.Close();
        }

        /// <summary>
        /// Loads the preferences from the disk. If loading fails, loads defaults
        /// </summary>
        public void Load()
        {
            TextReader r = null;
            XmlSerializer s = new XmlSerializer(typeof(List<Setting>));

            string path = EspressoUtils.PathToPreferenceFileStorageFolder + "preferences.xml";

            try
            {
                r = new StreamReader(path);
                List<Setting> list = (List<Setting>)s.Deserialize(r);

                m_List = list;
            }
            catch (Exception ex)
            {
                Logger.Error("Encountered an error while deserialising preferences");
                Logger.Error(ex.ToString());

                m_List = Defaults;
            }
            finally
            {
                if (r != null)
                    r.Close();
            }
        }

        private List<Setting> Defaults
        {
            get
            {
                List<Setting> retVal = new List<Setting>();

                retVal.Add(new Setting("TransitionDelay", 3.0m));
                retVal.Add(new Setting("PageFoldAngle", 15));
                retVal.Add(new Setting("RollHeight", 50));
                retVal.Add(new Setting("TransitionType", "Page roll"));
                retVal.Add(new Setting("StopHalfWayThrough", false));
                retVal.Add(new Setting("ControlStrategy", "LeftNextRightPrevious"));
                retVal.Add(new Setting("InstantTransition", false));
                retVal.Add(new Setting("ChangeSpeedHalfWayThrough", false));
                retVal.Add(new Setting("FirstHalfDelay", 1.0m));
                retVal.Add(new Setting("SecondHalfDelay", 2.0m));
                retVal.Add(new Setting("UseDefaultPathToStorageFolder", true));
                retVal.Add(new Setting("PathToStorageFolder", "C:\\Temp"));
                retVal.Add(new Setting("DPI", 100));
                retVal.Add(new Setting("DocumentPageLimit", 10));
                retVal.Add(new Setting("LimitFoldMaxHeight", false));
                retVal.Add(new Setting("ViewStrategy", "SingleView"));

                return retVal;
            }
        }
    }

    /// <summary>
    /// The wrapper for the setting.
    /// </summary>
    public class Setting
    {
        [XmlElement("key")]
        public string Key;

        [XmlElement("value")]
        public object Value;

        public Setting(string key, object value)
        {
            Key = key;
            Value = value;
        }

        public Setting() : this(null, null) { }
    }

    /// <summary>
    /// The arguments passed when a property changes.
    /// </summary>
    public class PreferencesChangedEventArgs
    {
        public string Key;

        public object newValue;

        public PreferencesChangedEventArgs(string key, object newValue)
        {
            this.Key = key;
            this.newValue = newValue;
        }
    }
}
