using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using Espresso.Utils;

namespace Espresso.Viewer.Controller
{
    /// <summary>
    /// A mouse handler class that does not fire Click in case of DoubleClick. 
    /// Should be subscribed to Click and DoubleClick events of a control. It will then
    /// throw Click event only when DOUBLE_CLICK_DELAY milliseconds pass and user does not
    /// do a second click; if they do, DoubleClick is invoked.
    /// </summary>
    public class DoubleClickMouseHandler : EspressoMouseHandler
    {
        public override event MouseDelegate Click;
        public override event MouseDelegate DoubleClick;
        public override event MouseDelegate RightClick;

        private MouseButtons m_Buttons = MouseButtons.Left;
        private ClickState m_ClickState = ClickState.NotClicked;

        private Thread m_ClickThread = null;

        public override void ProcessClick(object sender, EventArgs e)
        {
            MouseEventArgs args = null;
            
            try
            {
                args = e as MouseEventArgs;
            }
            catch
            {
                return;
            }

            lock (this)
            {
                m_Buttons = args.Button;

                switch (m_ClickState)
                {
                    case ClickState.NotClicked:
                        //Start the single click thread
                        m_ClickThread = new Thread(new ThreadStart(TryToDoClick));
                        m_ClickThread.Start();
                        m_ClickState = ClickState.ClickedOnce;
                        break;
                    case ClickState.ClickedOnce:
                        //Abort the single click thread and do the double click instead
                        m_ClickThread.Abort();
                        Logger.Info("Executing double click");

                        if (DoubleClick != null)
                        {
                            DoubleClick();
                        }

                        m_ClickState = ClickState.NotClicked;
                        break;
                    case ClickState.ClickedTwice:

                        break;
                }
            }
        }

        private void TryToDoClick()
        {
            bool finishedWaiting = false;

            try
            {
                Thread.Sleep(DOUBLE_CLICK_DELAY);
                finishedWaiting = true;
            }
            catch (ThreadAbortException)
            {
                //MSDN says this point is unreachable
            }
            finally
            {
                if (finishedWaiting)
                {
                    lock (this)
                    {
                        if (m_Buttons == MouseButtons.Left)
                        {
                            Logger.Info("Executing click");
                            if (Click != null)
                            {
                                Click();
                            }
                        }
                        else
                        {
                            Logger.Info("Executing right click");
                            if (RightClick != null)
                            {
                                RightClick();
                            }
                        }

                        m_ClickState = ClickState.NotClicked;
                    }
                }
                else
                {
                    Logger.Info("Aborting click");
                }
            }
        }
    }
    
    public enum ClickState
    {
        NotClicked,
        ClickedOnce,
        ClickedTwice
    }
}
