using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using Espresso.Utils;

namespace Espresso.Viewer.Controller
{
    public class Timer
    {
        /// <summary>
        /// When paused holds the number of ticks elpsed from start to pause.
        /// </summary>
        private long m_TicksPausedAt;

        /// <summary>
        /// The time when the transition started
        /// </summary>
        private DateTime m_StartTime;

        private bool m_hasStarted;

        private bool m_Reversed;

        public bool IsReversed
        {
            get { return m_Reversed; }
            set { m_Reversed = value; }
        }	

        /// <summary>
        /// Returns the ratio in interval [0, 1] that represents how much of the transition is
        /// complete
        /// </summary>
        public float CompleteRatio
        {
            get
            {
                float retVal = 0;

                if (!m_hasStarted || m_StartTime.Ticks == 0 || m_Delay == null)
                {
                    retVal = 0;
                }
                else if (this.IsPaused)
                {
                    retVal = (float)m_TicksPausedAt / (float)m_Delay.Ticks;
                }
                else if ((m_StartTime + m_Delay) < DateTime.Now)
                {
                    retVal = 1;
                }
                else
                {
                    TimeSpan span = DateTime.Now - m_StartTime;

                    float f = (float)span.Ticks / (float)m_Delay.Ticks;
                    retVal = f;
                }

                return (IsReversed ? 1 - retVal : retVal);
            }
        }

        /// <summary>
        /// Shows whether the transition is in action
        /// </summary>
        public bool InTransition
        {
            get
            {
                return m_hasStarted && (IsReversed ? (CompleteRatio > 0) : (CompleteRatio < 1));
            }
        }
        
        private TimeSpan m_Delay;

        /// <summary>
        /// Time that this transition should take to complete in milliseconds. 
        /// The setter resets the transition.
        /// </summary>
        public decimal Delay
        {
            get { return (decimal)m_Delay.TotalMilliseconds; }
            set
            {
                m_Delay = new TimeSpan(0, 0, 0, 0, (int)value);
                Reset();
            }
        }

        private bool m_Paused;

        /// <summary>
        /// Shows whether the transition is paused
        /// </summary>
        public bool IsPaused
        {
            get { return m_Paused; }
        }

        /// <summary>
        /// Starts the transition with the current delay time
        /// </summary>
        public void Start()
        {
            m_StartTime = DateTime.Now;
            m_hasStarted = true;
        }

        /// <summary>
        /// Pauses the transition
        /// </summary>
        public void Pause()
        {
            m_Paused = true;
            m_TicksPausedAt = DateTime.Now.Ticks - m_StartTime.Ticks;
        }

        /// <summary>
        /// Resets the transition
        /// </summary>
        public void Reset()
        {
            m_hasStarted = false;
            m_StartTime = new DateTime(0);
        }

        /// <summary>
        /// Resumes the transition when paused.
        /// </summary>
        public void Resume()
        {
            if (!this.IsPaused)
                return;

            m_StartTime = DateTime.Now - new TimeSpan(m_TicksPausedAt);
            m_Paused = false;
        }


        /// <summary>
        /// Changes the speed of the transition by value represented by <see cref="factor"/>
        /// E.g. 0.5 would make it twice as slow and 2 would make it twice as fast.
        /// Currenty does not support paused transition
        /// </summary>
        /// <param name="factor"></param>
        public void AlterSpeed(float factor)
        {
            if (this.IsPaused || !m_hasStarted)
                return;

            DateTime now = DateTime.Now;
            TimeSpan newDelay = new TimeSpan((long)(m_Delay.Ticks / factor));
            TimeSpan timePassed = now - m_StartTime;

            float completeRatio = (float)timePassed.Ticks / (float)m_Delay.Ticks;
            TimeSpan completeSpan = new TimeSpan((long)(newDelay.Ticks * completeRatio));
            DateTime newStart = now - completeSpan;

            m_StartTime = newStart;
            m_Delay = newDelay;
        }

        /// <summary>
        /// Creates a new transition with the delay value specified by
        /// <see cref="milliseconds"/>
        /// </summary>
        /// <param name="milliseconds">Delay in milliseconds</param>
        public Timer(int milliseconds)
        {
            this.Delay = milliseconds;

            m_Reversed = false;

            this.Reset();
        }

        /// <summary>
        /// Creates a transition with 0 delay.
        /// </summary>
        public Timer() : this(0) { }
    }
}
