using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;
using Espresso.Viewer.View;
using Espresso.Viewer.Model;
using EspressoMenu = Espresso.Viewer.Model.Menu;
using EspressoMenuItem = Espresso.Viewer.Model.MenuItem;
using Espresso.Utils;

namespace Espresso.Viewer.Controller
{
    public class ViewerController : IViewerController, 
                                    IFieldContainer<IPainter>, 
                                    IFieldContainer<MouseStrategy>
    {
        private decimal DELAY
        {
            get
            {
                bool instant = (bool)Preferences.Instance.GetProperty("InstantTransition");
                
                if (instant)
                    return 0;
                else
                    return (decimal)Preferences.Instance.GetProperty("TransitionDelay") * 1000;
            }
        }

        //A hack to allow for half-page preview technique
        //TODO Get rid of this
        private bool stoppedHalfWayThrough = false;

        private int REVERSE_DELAY = 1000;

        private int MENU_SELECTION_DELAY = 1000;

        private const int FPS = 30;

        private Timer m_Timer;

        public Timer Timer
        {
            get { return m_Timer; }
            private set { m_Timer = value; }
        }

        private IPainter m_TransitionPainter;

        public IPainter TransitionPainter
        {
            get
            {
                return m_TransitionPainter;
            }
        }

        private MouseStrategy m_MouseStrategy;

        private IEspressoMouseHandler m_MouseHandler = new DoubleClickMouseHandler();

        public Document Document;

        public EspressoViewer Viewer;

        private Page m_CurrentPage;

        private Thread m_MenuSelectionThread = null;

        /// <summary>
        /// The page that is currently displayed.
        /// </summary>
        public Page CurrentPage
        {
            get { return m_CurrentPage; }
            
            private set 
            { 
                m_CurrentPage = value;
                this.Document.CurrentPageNumber = value.PageNumber +1;
            }
        }

        private EspressoMenu m_Menu;

        public EspressoMenu Menu
        {
            get
            {
                if (m_Menu == null)
                    m_Menu = DefaultMenu;

                return m_Menu;
            }
        }

        //TODO: ADDED BY JOJO, PagesToBeDisplayed
        private int m_PagesToBeDisplayed;

        public int PagesToBeDisplayed
        {
            get
            {
                return m_PagesToBeDisplayed;
            }
            set
            {
                m_PagesToBeDisplayed = value;
            }
        }

        public ViewerController(Document document, EspressoViewer viewer)
        {
            this.Timer = new Timer(0);
            
            this.Document = document;
            this.Viewer = viewer;

            // TODO: ADDED BY JOJO, instantiate PagesToBeDisplayed
            m_PagesToBeDisplayed = 1;

            //Set up controls
            //Register these only once, because since they are anonymous delegates,
            //they will not depend on which m_MouseHandler is currently set.
            this.Viewer.MouseClick += delegate(object sender, MouseEventArgs args)
            {
                m_MouseHandler.ProcessClick(sender, args);
            };

            this.Viewer.MouseDoubleClick += delegate(object sender, MouseEventArgs args)
            {
                m_MouseHandler.ProcessClick(sender, args);
            };

            this.Viewer.Paint += delegate(object sender, PaintEventArgs e)
            {
                TransitionPainter.Paint(sender, e);
            };
        }

        /// <summary>
        /// Redraws the screen at about 30 fps until the transtition is over.
        /// Must be done in a separate thread.
        /// </summary>
        private bool DefaultDraw(decimal delay)
        {
            bool shouldStop = (bool)Preferences.Instance.GetProperty("StopHalfWayThrough");
            bool shouldChangeSpeed = (bool)Preferences.Instance.GetProperty("ChangeSpeedHalfWayThrough");

            decimal firstHalf = (decimal)Preferences.Instance.GetProperty("FirstHalfDelay") * 1000;
            decimal secondHalf = (decimal)Preferences.Instance.GetProperty("SecondHalfDelay") * 1000;

            if (!shouldChangeSpeed)
            {
                this.Timer.Delay = delay;
            }
            else
            {
                this.Timer.Delay = firstHalf * 2;
            }

            if (Timer.IsReversed)
            {
                this.Timer.Delay = REVERSE_DELAY;
            }

            //this.TransitionPainter.Reset();
            this.Timer.Start();
            
            this.TransitionPainter.CompleteRatio = this.Timer.CompleteRatio;
            this.Viewer.Invalidate();

            while (Timer.InTransition)
            {
                float r = this.Timer.CompleteRatio;
                this.TransitionPainter.CompleteRatio = r;

                if (shouldStop && r > 0.5 && !Timer.IsPaused && !Timer.IsReversed)
                {
                    Timer.Pause();
                    shouldStop = false;
                }

                if (shouldChangeSpeed && r > 0.5 && !Timer.IsReversed)
                {
                    shouldChangeSpeed = false;
                    Timer.AlterSpeed((float)(firstHalf/secondHalf));
                }

                if (!Timer.IsPaused)
                {
                    this.Viewer.Invalidate();

                    Thread.Sleep((int)1000 / FPS);
                }
            }

            this.Viewer.Invalidate();

            //Only returns a value to allow a delegate signature in DoTransition
            return true;
        }

        private void DoTransition(Bitmap oldPage, Bitmap newPage)
        {
            this.TransitionPainter.OldImage = oldPage;
            this.TransitionPainter.NewImage = newPage;

            //Spawn the drawing method in a separate thread
            MethodCall<bool, decimal> asyncDrawing = this.DefaultDraw;

            if (Timer.IsReversed)
            {
                asyncDrawing.BeginInvoke(REVERSE_DELAY, delegate(IAsyncResult result)
                   {
                       asyncDrawing.EndInvoke(result);
                   }, null);
            }
            else
            {
                asyncDrawing.BeginInvoke(DELAY, delegate(IAsyncResult result)
                  {
                      asyncDrawing.EndInvoke(result);
                  }, null);
            }
        }

        private void DrawPageInstantly(Page page)
        {
            if (page == null || this.TransitionPainter == null)
                return;

            this.TransitionPainter.NewImage = page.Bitmap;
            this.TransitionPainter.OldImage = page.Bitmap;

            this.TransitionPainter.CompleteRatio = 0;

            this.Viewer.Invalidate();

            this.CurrentPage = page;
        }

        /// <summary>
        /// Draws the first page.
        /// </summary>
        public void Reset()
        {
            DrawPageInstantly(this.Document.Pages[0]);
        }
        
        #region Commands

        /// <summary>
        /// Makes a transition from the current page to page n (starting at 1).
        /// </summary>
        /// <param name="n"></param>
        /// <param name="doTransition">When true, a transition will be made to the new page
        ///                            Otherwise, the new page is simply drawn.</param>
        public void GoToPage(int n, bool doTransition)
        {
            bool shouldStop = (bool)Preferences.Instance.GetProperty("StopHalfWayThrough");

            if (!this.Document.CanGoToPage(n))
                return;

            Page newPage = this.Document.Pages[n-1];

            if (doTransition && DELAY != 0)
            {
                if (Timer.IsReversed)
                    DoTransition(newPage.Bitmap, this.CurrentPage.Bitmap);
                else
                    DoTransition(this.CurrentPage.Bitmap, newPage.Bitmap);

                this.CurrentPage = newPage;
            }
            else if (doTransition && DELAY == 0 && shouldStop && !Timer.IsReversed)
            {//Patented half-page preview technique
                if (!stoppedHalfWayThrough)
                {
                    stoppedHalfWayThrough = true;
                    this.TransitionPainter.CompleteRatio = 0.5f;
                    this.TransitionPainter.NewImage = newPage.Bitmap;
                    this.TransitionPainter.OldImage = CurrentPage.Bitmap;

                    this.Viewer.Invalidate();
                }
                else
                {
                    stoppedHalfWayThrough = false;
                    this.TransitionPainter.CompleteRatio = 1.0f;

                    this.Viewer.Invalidate();

                    this.CurrentPage = newPage;
                }
            }
            else
            {
                DrawPageInstantly(newPage);
            }

        }

        /// <summary>
        /// Makes a transition to the next page unless the last page is displayed
        /// </summary>
        public void GoToNextPage(bool doTransition)
        {
            if (Timer.IsPaused)
            {
                Timer.Resume();
            }
            else if (this.Document.CurrentPageNumber != this.Document.Pages.Count)
            {
                Timer.Reset();
                Timer.IsReversed = false;
                GoToPage(this.Document.CurrentPageNumber + PagesToBeDisplayed, doTransition); // TODO: JOJO
            }
        }

        /// <summary>
        /// Makes transition to the previous page unless the first page is displayed.
        /// </summary>
        public void GoToPreviousPage(bool doTransition)
        {
            if (Timer.IsPaused)
            {
                Timer.Resume();
            }
            else if (this.Document.CurrentPageNumber != 1)
            {
                Timer.Reset();
                Timer.IsReversed = true;
                GoToPage(this.Document.CurrentPageNumber - PagesToBeDisplayed, doTransition); // TODO: JOJO
            }
        }

        /// <summary>
        /// Instantly display the first page of the document
        /// </summary>
        public void GoToFirstPage(bool doTransition)
        {
            GoToPage(1, doTransition);
        }

        /// <summary>
        /// Instantly display the last page of the document
        /// </summary>
        public void GoToLastPage(bool doTransition)
        {
            GoToPage(Document.Pages.Count, doTransition);
        }
        #endregion

        #region Menu functions
        /// <summary>
        /// Brings menu on the screen and starts the timer.
        /// </summary>
        /// <param name="visible"></param>
        public void ToggleMenu()
        {
            if (Menu.IsVisible)
                return;

            Menu.IsVisible = true;

            m_MenuSelectionThread = new Thread(new ThreadStart(DelaySelection));
            m_MenuSelectionThread.Start();

            this.Viewer.Invalidate();
        }

        public void ChooseNextMenuItem()
        {
            Menu.ChooseNextItem();

            if (m_MenuSelectionThread != null)
            {
                m_MenuSelectionThread.Abort();
                m_MenuSelectionThread = new Thread(new ThreadStart(DelaySelection));
                m_MenuSelectionThread.Start();
            }

            this.Viewer.Invalidate();
        }

        public void SelectMenuItem()
        {
            Menu.CurrentItem.Select();
            Menu.IsVisible = false;
            Menu.Reset();

            this.Viewer.Invalidate();
        }

        private void DelaySelection()
        {
            Thread.Sleep(MENU_SELECTION_DELAY);

            this.SelectMenuItem();
        } 
        
        private EspressoMenu DefaultMenu
        {
            get
            {
                EspressoMenu m = new EspressoMenu();
                EspressoMenuItem mi = new EspressoMenuItem(delegate { GoToNextPage(true); }, "Next page");
                m.AddItem(mi);
                m.CurrentItem = mi;
                m.AddItem(new EspressoMenuItem(delegate { GoToPreviousPage(true); }, "Previous page"));
                m.AddItem(new EspressoMenuItem(delegate { GoToPreviousPage(false); }, "Jump to previous"));

                return m;
            }
        }

        #endregion

        public void SetField(IPainter newValue)
        {
            m_TransitionPainter = newValue;
        }

        public void SetField(MouseStrategy newValue)
        {
            m_MouseStrategy = newValue;

            if (typeof(DoubleClickBackMouseStrategy).Equals(newValue.GetType()))
            {
                m_MouseHandler = new DoubleClickMouseHandler();
            }
            else
            {
                m_MouseHandler = new SimpleMouseHandler();
            }

            m_MouseHandler.Click += delegate()
            {
                m_MouseStrategy.LeftClick();
            };

            m_MouseHandler.DoubleClick += delegate()
            {
                m_MouseStrategy.DoubleClick();
            };

            m_MouseHandler.RightClick += delegate()
            {
                m_MouseStrategy.RightClick();
            };
        }
    }
}
