using System;
using System.Collections.Generic;
using System.Text;
using Espresso.Utils;
using Espresso.OMR;
using Espresso.Utils.View;

namespace Espresso.Viewer.Model
{
    /// <summary>
    /// Contains all the pages of the document and other information such as the speed of the
    /// piece.
    /// </summary>
    public class Document
    {
        /// <summary>
        /// Document conversion resolution
        /// </summary>
        public int DPI = (int)Preferences.Instance.GetProperty("DPI");

        private List<Page> m_Pages;

        /// <summary>
        /// All the pages in the document
        /// </summary>
        public List<Page> Pages
        {
            get { return m_Pages; }
            private set { m_Pages = value; }
        }

        private TimeSignature m_Signature;

        public TimeSignature TimeSignature
        {
            get { return m_Signature; }
            set { m_Signature = value; }
        }

        private int m_BarsPerPage;

        public int BarsPerPage
        {
            get { return m_BarsPerPage; }
            set { m_BarsPerPage = value; }
        }

        private int m_CurrentPageNumber;

        public int CurrentPageNumber
        {
            get { return m_CurrentPageNumber; }
            
            set
            {
                m_CurrentPageNumber = value;

                if (CurrentPageChanged != null)
                {
                    CurrentPageChanged(new CurrentPageChangedEventArgs(value));
                }
            }
        }

        public delegate void CurrentPageChangedEventHandler(CurrentPageChangedEventArgs args);

        public event CurrentPageChangedEventHandler CurrentPageChanged;

        /// <summary>
        /// Can we go to a given page number?
        /// </summary>
        /// <param name="pageNumber"></param>
        /// <returns></returns>
        public bool CanGoToPage(int pageNumber)
        {
            if (pageNumber < 1 || pageNumber > this.Pages.Count)
            {
                return false;
            }

            return true;
        }

        /// <summary>
        /// Returns all staff system rectangles in current document
        /// </summary>
        public IEnumerable<SystemRectangle> AllSystems
        {
            get 
            {
                foreach (Page page in this.Pages)
                    foreach (SystemRectangle rectangle in page.StaffSystemRects)
                        yield return rectangle;
            }
        }

        /// <summary>
        /// Default constructor. Does nothing. Exists so that the ProcessedPage event can be registered
        /// only after that the actual document is opened.
        /// </summary>
        public Document()
        {
            
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="pdfFilePath"></param>
        /// <param name="overWriteCache"></param>
        /// <returns></returns>
        /// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!@#$%&#%$$
        /// TODO: Remove the gui code from document creation. Remove the Utils.View dependency
        public bool OpenAPdf(string pdfFilePath, bool overWriteCache)
        {
            int first = 1;
            int last = 1;

            int pdfNumberOfPages = BitmapConvertUtils.PdfNumberOfPages(pdfFilePath);
            last = pdfNumberOfPages;

            this.Pages = new List<Page>();

            if (pdfNumberOfPages > (int)Preferences.Instance.GetProperty("DocumentPageLimit"))
            {
                OpenRequestForm.AskForOpenConfirmation(out first, out last, pdfNumberOfPages);

                if (first < 0)
                {
                    //User cancelled the dialog
                    return false;
                }
            }

            for (int n = first - 1; n < last; n++)
            {
                Logger.Info("Processing page {0} of {1}", n, pdfFilePath);

                //string pbmFilename = BitmapConvertUtils.ConvertPdfPageToPbm(pdfFilePath, n, CacheMode.UseCache, DPI);

                string pngFilename = "";
                if (overWriteCache)
                {
                    pngFilename = BitmapConvertUtils.ConvertPdfPageToPng(pdfFilePath, n, CacheMode.OverwriteCache, DPI);
                }
                else
                {
                    pngFilename = BitmapConvertUtils.ConvertPdfPageToPng(pdfFilePath, n, CacheMode.UseCache, DPI);
                }

                try
                {
                    Page page = new Page("", pngFilename);
                    page.PageNumber = n;

                    m_Pages.Add(page);
                }
                catch (Exception e)
                {
                    Logger.Error(e.Message);
                }

                if (ProcessedAPage != null)
                    ProcessedAPage(this, new ProcessedPageEventArgs(n + 1, pdfNumberOfPages));
            }

            CurrentPageNumber = 1;

            return true;
        }

        /// <summary>
        /// A very simplistic way to let others (for example a progress bar) know that we have 
        /// processed a page. 
        /// </summary>
        public event EventHandler<ProcessedPageEventArgs> ProcessedAPage;
    }

    public class ProcessedPageEventArgs : EventArgs
    {
        public int ProcessedPageNumber;

        public int TotalPages;

        public ProcessedPageEventArgs(int pageNum, int totalPages)
        {
            ProcessedPageNumber = pageNum;
            TotalPages = totalPages;
        }
    }

    public class CurrentPageChangedEventArgs : EventArgs
    {
        public int currentPageNumber;

        public CurrentPageChangedEventArgs(int pageNumber)
        {
            currentPageNumber = pageNumber;
        }
    }
}
