using System;
using System.Collections.Generic;
using System.Text;

using DotNetBitmap = System.Drawing.Bitmap;
using System.IO;
using Espresso.OMR;
using System.Xml.Serialization;
using Espresso.Utils;
using System.Drawing;

namespace Espresso.Viewer.Model
{
    /// <summary>
    /// A class that reprensents a page in a document
    /// </summary>
    public class Page
    {
        private int m_PageNumber;

        /// <summary>
        /// Page number
        /// </summary>
        public int PageNumber
        {
            get { return m_PageNumber; }
            set { m_PageNumber = value; }
        }

        protected DotNetBitmap m_Bitmap;

        /// <summary>
        /// The bitmap of the page
        /// </summary>
        public DotNetBitmap Bitmap
        {
            get { return m_Bitmap; }
            private set { m_Bitmap = value; }
        }

        protected SystemRectangle[] m_StaffSystemRects;

        /// <summary>
        /// System rectangles that this page contains
        /// </summary>
        public SystemRectangle[] StaffSystemRects
        {
            get { return m_StaffSystemRects; }
            private set { m_StaffSystemRects = value; }
        }

        private int m_Width;

        /// <summary>
        /// Page width
        /// </summary>
        public int Width
        {
            get { return m_Width; }
            private set { m_Width = value; }
        }

        private int m_Height;

        /// <summary>
        /// Page height
        /// </summary>
        public int Height
        {
            get { return m_Height; }
            private set { m_Height = value; }
        }

        /// <summary>
        /// Uses the pbm file for displaying the page.
        /// </summary>
        /// <param name="pbmFilepath"></param>
        public Page(string pbmFilepath) : this(pbmFilepath, null) { }

        /// <summary>
        /// Constructs a new page from a pbm file pointed to by <see cref="pbmFilepath"/>
        /// </summary>
        /// <param name="pbmFilepath">The path to the pbm file that contains the page</param>
        /// <param name="imageFilepath">The path to the image file that will be used in displaying 
        ///                              the page. If empty or null the pbm file will be used.</param>
        public Page(string pbmFilepath, string imageFilepath)
        {
            try
            {
                if (String.IsNullOrEmpty(imageFilepath))
                {
                    //GeneratePageBitmap(pbmFilepath);
                }
                else
                {
                    GenerateBitmapFromFile(imageFilepath);
                }
                
                /*if (!TryDeserialiseStaffSystemRects(pbmFilepath))
                {
                    //Could not read cached data from the disk
                    //SepApplication sepApp = SepLibrary.custom_process(pbmFilepath, 2.0, 2);

                    //GenerateStaffSystemRects(sepApp);

                    //ExpandStaffSystemRects();

                    //SerialiseStaffSystemRects(pbmFilepath);
                }*/
            }
            catch (InvalidPbmFileException e)
            {
                throw e;
            }
        }
        
        /// <summary>
        /// Generates a .NET bitmap object from the pbm file <see cref="pbmFilePath"/> points to. 
        /// Sets the <see cref="Bitmap"/> field to the generated bitmap. Sets the <see cref="Width"/>
        /// and <see cref="Height"/> fields.
        /// </summary>
        /// <returns>True on successful generation and set, false otherwise</returns>
        /*
        protected void GeneratePageBitmap(string pbmPath)
        {
            FileStream fsin = new FileStream(pbmPath, FileMode.Open, FileAccess.Read);
            BinaryReader brin = new BinaryReader(fsin);

            FileBitmap fileBitmap = new FileBitmap();
            fileBitmap.set_filename(pbmPath);
            fileBitmap.set_valid(true);

            fileBitmap.replace(brin);

            if (fileBitmap.valid())
            {
                this.Width = fileBitmap.x_dim();
                this.Height = fileBitmap.y_dim();

                this.Bitmap = BitmapConvertUtils.ConvertToDotNetBitmapDrawLine(fileBitmap);
            }
            else
            {
                throw new InvalidPbmFileException("Could not read the pbm file: " + pbmPath);
            }

            brin.Close();
            fsin.Close();
        }
        */
        protected void GenerateBitmapFromFile(string filePath)
        {
            this.Bitmap = new DotNetBitmap(filePath);

            this.Width = this.Bitmap.Width;
            this.Height = this.Bitmap.Height;
        }

        /// <summary>
        /// Runs OMR over the current bitmap.
        /// </summary>
        /// <param name="sepApp">The OMR application</param>
        /*
        private void GenerateStaffSystemRects(Espresso.OMR.SepApplication sepApp)
        {
            MObjects mobjects = sepApp.mobjects();
            Espresso.OMR.Bitmap correctedPage = mobjects.corrected_page();
            StaffSystemList activeSS = null;
            StaffSystemNode currentNode = null;

            if ((object)correctedPage != null)
            {
                activeSS = mobjects.corrected_staff_systems();
            }
            else
            {
                activeSS = mobjects.staff_systems();
            }

            if (activeSS.length() == 0)
            {
                // OMR code failed to find anything.
                // Set up single staff system rect the same size as the page
                this.StaffSystemRects = new SystemRectangle[1];
                this.StaffSystemRects[0] = new SystemRectangle(this);

                return;
            }
            else
            {
                this.StaffSystemRects = new SystemRectangle[activeSS.length()];

                currentNode = activeSS.head();
                int pos = 0;

                Logger.Info("Staff Systems:");
                
                while (currentNode != null)
                {
                    StaffSystem ss = currentNode.val();
                    GObject go = ss.gobject();
                    
                    Rect rect = new Rect(go.x_org(), go.y_org(), go.x_dim(), go.y_dim());
                    SystemRectangle sysRect = new SystemRectangle(this, rect);
                    
                    Logger.Info("\t" + sysRect);

                    this.StaffSystemRects[pos] = sysRect;

                    currentNode = activeSS.next();
                    pos++;
                }
            }
        }
        */

        /// <summary>
        /// Expands system rectangles to fill the whole document
        /// </summary>
        private void ExpandStaffSystemRects()
        {
            int n = this.StaffSystemRects.Length;
            
            if (n == 0) { return; }

            Rect topRect = this.StaffSystemRects[0].Rectangle;
            topRect.Top = topRect.Left = 0;
            topRect.Right = this.Width;

            for (int i = 0; i < n - 1; i++)
            {
                Rect current = this.StaffSystemRects[i].Rectangle;
                Rect next = this.StaffSystemRects[i + 1].Rectangle;

                current.Left = next.Left = 0;
                current.Right = next.Right = this.Width;

                int yGap = next.Top - current.Bottom;
                int ySplit = current.Bottom + (yGap * 7 / 10);

                current.Bottom = next.Top = ySplit;
            }

            Rect bottom = this.StaffSystemRects[n - 1].Rectangle;
            bottom.Bottom = this.Height;
        }

        /// <summary>
        /// Attempts to deserialise staff rect arrays from cache on hard drive. If manages
        /// to successfully deserialise objects, sets the arrays.
        /// </summary>
        /// <param name="filename">Full path to the pbm file for which the staff arrays are fetched</param>
        /// <returns>True if the deserialised files were read and deserialised, false otherwise.</returns>
        private bool TryDeserialiseStaffSystemRects(string pbmPath)
        {
            bool retVal = true;
            SystemRectangle[] staffRects;
            TextReader r = null;
            XmlSerializer s = new XmlSerializer(typeof(SystemRectangle[]));

            int dotPos = pbmPath.LastIndexOf('.');
            string pathNoExt = pbmPath.Substring(0, dotPos);

            string srXmlFile = pathNoExt + "_ss_rects.xml";

            try
            {
                r = new StreamReader(srXmlFile);
                staffRects = (SystemRectangle[])s.Deserialize(r);
                
                //Set the page fields of the deserialised rectangles
                foreach (SystemRectangle rect in staffRects)
                    rect.Page = this;

                this.StaffSystemRects = staffRects;
            }
            catch (Exception ex)
            {
                Logger.Error("Encountered an error while deserialising staff arrays");
                Logger.Error(ex.ToString());

                retVal = false;
            }
            finally
            {
                if (r != null)
                    r.Close();
            }

            return retVal;
        }

        /// <summary>
        /// Serialises the staff arrays into xml files in the application data folder specific to the user.
        /// </summary>
        private void SerialiseStaffSystemRects(string pbmPath)
        {
            XmlSerializer s = new XmlSerializer(typeof(SystemRectangle[]));

            int dotPos = pbmPath.LastIndexOf('.');
            string pathNoExt = pbmPath.Substring(0, dotPos);

            string ssXmlFile = pathNoExt + "_ss_rects.xml";

            TextWriter w = new StreamWriter(ssXmlFile);
            s.Serialize(w, this.StaffSystemRects);
            w.Close();
        }
    }
}
