using System.Xml;
using System.Xml.Serialization;

namespace Espresso.Viewer.Model
{
    /// <summary>
    /// A Rectangle represented by an upper left and lower right corners
    /// </summary>
    [XmlRoot("rectangle")]
    public class Rect
    {
        private EspressoPoint m_UpperLeft;

        /// <summary>
        /// Upper left corner of the rectangle
        /// </summary>
        [XmlElement("upperLeftCorner")]
        public EspressoPoint UpperLeftCorner
        {
            get { return m_UpperLeft; }
            set { m_UpperLeft = value; }
        }

        private EspressoPoint m_LowerRight;

        /// <summary>
        /// Lower right corner of the rectangle
        /// </summary>
        [XmlElement("lowerRightCorner")]
        public EspressoPoint LowerRightCorner
        {
            get { return m_LowerRight; }
            set { m_LowerRight = value; }
        }

        /// <summary>
        /// Creates a rectangle with corners at <see cref="upperLeft"/> and
        /// <see cref="lowerRight"/>
        /// </summary>
        /// <param name="upperLeft"></param>
        /// <param name="lowerRight"></param>
        public Rect(EspressoPoint upperLeft, EspressoPoint lowerRight)
        {
            this.UpperLeftCorner = upperLeft;
            this.LowerRightCorner = lowerRight;
        }

        /// <summary>
        /// Creates a rectangle with both corners at (0,0)
        /// </summary>
        public Rect() : this(new EspressoPoint(), new EspressoPoint()) { }

        /// <summary>
        /// Creates a new rectangle <see cref="xDim"/> wide and <see cref="yDim"/> high
        /// with upper left corner at (<see cref="xOrg"/>, <see cref="yOrg"/>)
        /// </summary>
        public Rect(int xOrg, int yOrg, int xDim, int yDim)
            :
            this(new EspressoPoint(xOrg, yOrg), new EspressoPoint(xOrg + xDim, yOrg + yDim)) { }

        /// <summary>
        /// Creates a copy of <see cref="rect"/>.
        /// </summary>
        /// <param name="rect">Rectangle to copy</param>
        public Rect(Rect rect) : this(rect.UpperLeftCorner, rect.LowerRightCorner) { }

        /// <summary>
        /// Try to get rid of this method
        /// </summary>
        /// <param name="xOrg"></param>
        /// <param name="yOrg"></param>
        /// <param name="xDim"></param>
        /// <param name="yDim"></param>
        public void setRect(int xOrg, int yOrg, int xDim, int yDim)
        {
            this.UpperLeftCorner.X = xOrg;
            this.UpperLeftCorner.Y = yOrg;
            this.LowerRightCorner.X = xDim + xOrg;
            this.LowerRightCorner.Y = yDim + yOrg;
        }

        /// <summary>
        /// X coordinate of the upper left corner
        /// </summary>
        [XmlIgnore]
        public int Left
        {
            get { return this.UpperLeftCorner.X; }
            set { this.UpperLeftCorner.X = value; }
        }

        /// <summary>
        /// Y coordinate of the upper left corner
        /// </summary>
        [XmlIgnore]
        public int Top
        {
            get { return this.UpperLeftCorner.Y; }
            set { this.UpperLeftCorner.Y = value; }
        }

        [XmlIgnore]
        public int Bottom
        {
            get { return this.LowerRightCorner.Y; }
            set { this.LowerRightCorner.Y = value; }
        }

        [XmlIgnore]
        public int Right
        {
            get { return this.LowerRightCorner.X; }
            set { this.LowerRightCorner.X = value; }
        }
        
        /// <summary>
        /// Width of the rectangle
        /// </summary>
        public int Width
        {
            get { return this.LowerRightCorner.X - this.UpperLeftCorner.X; }
        }

        /// <summary>
        /// Height of the rectangle
        /// </summary>
        public int Height
        {
            get { return this.LowerRightCorner.Y - this.UpperLeftCorner.Y; }
        }

        /// <summary>
        /// String representation of the rectangle
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            string message = string.Format("[{0},{1},{2},{3}]",
                               this.UpperLeftCorner.X,
                               this.UpperLeftCorner.Y,
                               this.LowerRightCorner.X,
                               this.LowerRightCorner.Y);
            return message;
        }
    }
}


