using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using Espresso.Viewer.Model;

namespace Espresso.Viewer.View
{
    public abstract class AbstractPainter :IPainter
    {
        protected Bitmap m_bitBuffer;
        protected Graphics g;

        //TODO: Added by JOJO, protected Document document;
        protected Document document;

        public virtual Document Document
        {
            get { return document; }
            set { document = value; }
        }

        private float m_TransitionComplete;

        /// <summary>
        /// The value representing how much of the transition is complete.
        /// </summary>
        public virtual float CompleteRatio
        {
            get { return m_TransitionComplete; }
            set { m_TransitionComplete = value; }
        }

        private Bitmap m_OldImage;

        public virtual Bitmap OldImage
        {
            get { return m_OldImage; }
            set
            {
                m_OldImage = value;
                this.ScaledOldImage = null;
            }
        }

        private Bitmap m_NewImage;

        public virtual Bitmap NewImage
        {
            get { return m_NewImage; }
            set
            {
                m_NewImage = value;
                this.ScaledNewImage = null;
            }
        }

        private Bitmap m_ScaledOldImage;

        protected Bitmap ScaledOldImage
        {
            get { return m_ScaledOldImage; }
            set { m_ScaledOldImage = value; }
        }

        private Bitmap m_ScaledNewImage;

        protected Bitmap ScaledNewImage
        {
            get { return m_ScaledNewImage; }
            set { m_ScaledNewImage = value; }
        }

        private Rectangle m_InvalidateRegion;
	
        /// <summary>
        /// The rectangle that should be updated. This should be set by the
        /// inheriting classes once the painting is done. The controller will
        /// then update just this region.
        /// </summary>
        //TODO This could probably be removed.
        public Rectangle InvalidRegion
        {
            get { return m_InvalidateRegion; }
            protected set { m_InvalidateRegion = value; }
        }

        public abstract string Name
        {
            get;
        }

        /// <summary>
        /// Resets the <see cref="InvalidateRegion"/>
        /// </summary>
        public void Reset()
        {
            this.InvalidRegion = new Rectangle(-1,-1,1,1);
        }
	
        /// <summary>
        /// Paints the transition between the <see cref="OldImage"/> and <see cref="NewImage"/>
        /// on <see cref="graphics"/>. The transition stage is specified by 
        /// <see cref="TransitionComplete"/>.
        /// </summary>
        public abstract void Paint(object sender, PaintEventArgs args);

        public AbstractPainter(Bitmap oldImage, Bitmap newImage)
        {
            this.OldImage = oldImage;
            this.NewImage = newImage;
        }

        public AbstractPainter() { }

        //TODO: ADDED BY JOJO, AbstractPainter(Document document) 
        public AbstractPainter(Bitmap oldImage, Bitmap newImage, Document document) 
        {
            this.OldImage = oldImage;
            this.NewImage = newImage;
            this.document = document;
        }

        /// <summary>
        /// Calculates the rectangle that the images should be drawn on.
        /// Currently tries to fit the whole page onto the control.
        /// Assumes the images are of equal width.
        /// </summary>
        /// <param name="control">The control that the images are drawn on.</param>
        /// <returns>The calculated rectangle</returns>
        protected Rectangle GetDrawRectangle(Control control)
        {
            float scaleFactor = (float)control.Height / (float)NewImage.Height;
            int scaledWidth = (int)(NewImage.Width * scaleFactor);
            int scaledHeight = (int)(NewImage.Height * scaleFactor);
            Point scaledOrigin = new Point((control.Width - scaledWidth) / 2, 0);

            Rectangle scaledRectangle = new Rectangle(scaledOrigin.X,
                                                      scaledOrigin.Y,
                                                      scaledWidth,
                                                      scaledHeight);
            return scaledRectangle;

            // TODO: JOJO
            //float scaleFactor = (float) (control.Width / 3) / (float) NewImage.Width;
            //int scaledWidth = (int)(NewImage.Width * scaleFactor);
            //int scaledHeight = (int)(NewImage.Height * scaleFactor);
            //Point scaledOrigin = new Point(0, (control.Height - scaledHeight) / 2);
            //Rectangle scaledRectangle = new Rectangle(scaledOrigin.X,
            //                                          scaledOrigin.Y,
            //                                          scaledWidth,
            //                                          scaledHeight);
            //return scaledRectangle;
        }

        // TODO: JOJO
        protected Rectangle GetDrawRectangle(Control control, int numDisplayedPages)
        {
            float scaleFactor = (float)(control.Width / numDisplayedPages) / (float)NewImage.Width;
            int scaledWidth = (int)(NewImage.Width * scaleFactor);
            int scaledHeight = (int)(NewImage.Height * scaleFactor);
            Point scaledOrigin = new Point(0, (control.Height - scaledHeight) / 2);
            Rectangle scaledRectangle = new Rectangle(scaledOrigin.X,
                                                      scaledOrigin.Y,
                                                      scaledWidth,
                                                      scaledHeight);
            return scaledRectangle;
        }

        // TODO: Added by Jojo
        protected List<Rectangle> GetDrawRectangleList(Control control, int numDisplayedPages)
        {
            List<Rectangle> rectangles = new List<Rectangle>();
            for(int i = 0; i < numDisplayedPages; i++) {
                float scaleFactor = (float)(control.Width / numDisplayedPages) / (float)NewImage.Width;
                int scaledWidth = (int)(NewImage.Width * scaleFactor);
                int scaledHeight = (int)(NewImage.Height * scaleFactor);
                Point scaledOrigin = new Point(i * scaledWidth, (control.Height - scaledHeight) / 2);
                Rectangle scaledRectangle = new Rectangle(scaledOrigin.X,
                                                          scaledOrigin.Y,
                                                          scaledWidth,
                                                          scaledHeight);
                rectangles.Add(scaledRectangle);
            }
            
            return rectangles;
        }

        /// <summary>
        /// Initialises the bit and graphics buffer. Sets the scaled image fields.
        /// </summary>
        protected void InitialiseBuffers(Control control)
        {
            this.m_bitBuffer = new Bitmap(control.Size.Width, control.Size.Height);
            g = Graphics.FromImage(this.m_bitBuffer);

            Rectangle scaled = GetDrawRectangle(control);
            
            this.ScaledNewImage = (Bitmap)NewImage.GetThumbnailImage(scaled.Width, scaled.Height, null, System.IntPtr.Zero);
            this.ScaledOldImage = (Bitmap)OldImage.GetThumbnailImage(scaled.Width, scaled.Height, null, System.IntPtr.Zero);
        }

        // TODO: ADDED BY JOJO
        protected void LoadNewImageSet(int i)
        {
            this.OldImage = Document.Pages[i].Bitmap;
            this.NewImage = Document.Pages[i].Bitmap;
        }
    }
}
