using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Windows.Forms;
using Espresso.Utils;

namespace Espresso.Viewer.View
{
    public class PageTurnTransitionPainter : AbstractPainter
    {
        //Page turn variables
        private double ALPHA
        {
            get
            {
                return (int)Preferences.Instance.GetProperty("PageFoldAngle");
            }
        }

        private double SIN;
        private double COS;
        private double TAN;
        private double V;

        private double MAX_HEIGHT
        {
            get
            {
                bool limit =  (bool)Preferences.Instance.GetProperty("LimitFoldMaxHeight");

                if (limit)
                    return 200;
                else
                    return 10000;
                //TODO Fix this hack
            }
        }

        public PageTurnTransitionPainter()
            : base(null, null)
        {
            Preferences.Instance.PreferencesChanged += delegate(PreferencesChangedEventArgs args)
            {
                if (args.Key.Equals("PageFoldAngle"))
                {
                    SetTrigonometry();
                }
            };

            SetTrigonometry();
        }

        private void SetTrigonometry()
        {
            SIN = Math.Sin((ALPHA / 180) * Math.PI);
            COS = Math.Cos((ALPHA / 180) * Math.PI);
            TAN = Math.Tan((ALPHA / 180) * Math.PI);
            V = SIN / (COS + 1);
        }

        public override void Paint(object sender, System.Windows.Forms.PaintEventArgs args)
        {
            Point A, B, C, D;
            Control control = (Control)sender;
            Rectangle drawRect = GetDrawRectangle(control);

            if (this.m_bitBuffer == null || this.m_bitBuffer.Size != control.Size
                || this.ScaledNewImage == null || this.ScaledOldImage == null)
            {
                InitialiseBuffers(control);
            }
            
            g.FillRectangle(Brushes.DarkGray, new Rectangle(new Point(0, 0), m_bitBuffer.Size));

            if (this.CompleteRatio == 0)
            {
                //Draw the white background
                g.FillRectangle(Brushes.White, drawRect);
                
                //Draw the complete old page
                Rectangle oldSrcRect = new Rectangle(0, 0, ScaledOldImage.Width, ScaledOldImage.Height);
                g.DrawImage(ScaledOldImage, drawRect, oldSrcRect, GraphicsUnit.Pixel);

                this.InvalidRegion = drawRect;
            }
            else if (this.CompleteRatio == 1)
            {
                //Draw the white background
                g.FillRectangle(Brushes.White, drawRect);

                //Draw the complete new page
                Rectangle newSrcRect = new Rectangle(0, 0, ScaledNewImage.Width, ScaledNewImage.Height);
                g.DrawImage(ScaledNewImage, drawRect, newSrcRect, GraphicsUnit.Pixel);

                this.InvalidRegion = drawRect;
            }
            else
            {
                double H = drawRect.Height;
                double W = drawRect.Width;
                double T = H * CompleteRatio;
                double X = drawRect.X;
                double Y = drawRect.Y;

                //A value by which the turning page should be offset down in order
                //to make it disappear completely when the transition is complete
                double OFFSET = ((1 - COS) * W) / SIN;
                double offRatio = OFFSET * (T / H);

                T += offRatio;

                Rectangle oldSrcRect = new Rectangle(0, 0, ScaledOldImage.Width, ScaledOldImage.Height);
                Rectangle newSrcRect = new Rectangle(0, 0, ScaledNewImage.Width, (int)T + 1);
                Rectangle newDestRect = new Rectangle((int)X, (int)Y, (int)W, (int)T + 1);

                //Need to do this because sometimes scaled images end up being null by the time
                // we get to this point in execution. Need to investigate into this more.
                if (this.m_bitBuffer == null || this.m_bitBuffer.Size != control.Size
                || this.ScaledNewImage == null || this.ScaledOldImage == null)
                {
                    InitialiseBuffers(control);
                }

                //Draw the white background
                g.FillRectangle(Brushes.White, drawRect);
                
                //Draw the complete old page
                g.DrawImage(ScaledOldImage, drawRect, oldSrcRect, GraphicsUnit.Pixel);
                
                //Draw part of the new image
                g.DrawImage(ScaledNewImage, newDestRect, newSrcRect, GraphicsUnit.Pixel);
                
                A = new Point((int)X, (int)T);

                if (T * COS < MAX_HEIGHT)
                {
                    D = new Point((int)(X + (T * SIN)), (int)(T * (1 + COS)));
                }
                else
                {
                    D = new Point((int) (X + MAX_HEIGHT*TAN), (int) (T + MAX_HEIGHT));
                }

                if (T / W <= V) //Triangle
                {
                    B = C = new Point((int)(X + T / V), 0);
                }
                else //Polygon
                {
                    B = new Point((int)(X + W), (int)((W * (COS - 1)) / SIN + T));

                    if (T * COS < MAX_HEIGHT)
                    {
                        C = new Point((int)(X + W * COS + T * SIN), (int)(W * (-SIN) + T * (COS + 1)));
                    }
                    else
                    {
                        C = new Point((int)(D.X + W * COS), (int)(D.Y - W * SIN));
                    }
                }

                Point[] pts = new Point[4] { A, B, C, D };
                g.FillPolygon(Brushes.White, pts, FillMode.Winding);
                g.DrawPolygon(new Pen(Brushes.Gray, 1), pts);

                if (T != 0)
                {
                    GraphicsPath path = GetShadowPath(A, B, C, D);

                    PathGradientBrush pgBrush = new PathGradientBrush(path);
                    pgBrush.CenterColor = Color.Gray;
                    

                    pgBrush.SurroundColors = new Color[] { Color.FromArgb(255, Color.Black), 
                                                           Color.FromArgb(255, Color.Black), 
                                                           Color.FromArgb(127, Color.White), 
                                                           Color.FromArgb(127, Color.White) };
                    Region r = new Region(path);
                    g.FillRegion(pgBrush, r);

                    path = GetSecondaryShadowPath(A, B, C, D);
                    pgBrush = new PathGradientBrush(path);
                    pgBrush.CenterColor = Color.FromArgb(127, Color.LightGray); 
                    pgBrush.SurroundColors = new Color[] { Color.FromArgb(127, Color.Gray), 
                                                           Color.FromArgb(127, Color.Gray), 
                                                           Color.FromArgb(127, Color.White), 
                                                           Color.FromArgb(127, Color.White) };
                    r = new Region(path);

                    g.FillRegion(pgBrush, r);
                }

                this.InvalidRegion = new Rectangle(new Point(0, 0), m_bitBuffer.Size);
            }

            args.Graphics.DrawImage(this.m_bitBuffer, new Point(0, 0));
        }

        /// <summary>
        /// Calculates the path that should enclose the shadow region
        /// </summary>
        /// <param name="A"></param>
        /// <param name="B"></param>
        /// <param name="C"></param>
        /// <param name="D"></param>
        /// <returns></returns>
        protected GraphicsPath GetShadowPath(Point A, Point B, Point C, Point D)
        {
            double l = 30;
            Point nA, nB, nC, nD;
            GraphicsPath retVal;
            nA = A;
            nB = B;
            nD = new Point((int)(A.X + l * SIN), (int)(A.Y + l * COS));
            
            if (B == C)
            {
                nC = B;
            }
            else
            {
                nC = new Point((int)(B.X + l * SIN), (int)(B.Y + l * COS));
            }

            retVal = new GraphicsPath(new Point[] { nA, nB, nC, nD }, new byte[] { 1, 1, 1, 1 });
            return retVal;
        }

        protected GraphicsPath GetSecondaryShadowPath(Point A, Point B, Point C, Point D)
        {
            double l = 30;
            Point nA, nB, nC, nD;
            GraphicsPath retVal;

            nA = A;
            nB = B;

            nD = new Point((int)(A.X), (int)((A.Y > l ? A.Y - l : 0)));
            
            if (B == C)
            {                
                nC = B;
            }
            else
            {
                nC = new Point((int)(B.X), (int)((B.Y > l ? B.Y - l : 0)));
            }

            retVal = new GraphicsPath(new Point[] { nA, nB, nC, nD }, new byte[] { 1, 1, 1, 1 });
            return retVal;
        }

        public override string Name
        {
            get { return "PageTurnPainter"; }
        }
    }
}
