using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;

namespace Espresso.Viewer.View
{
    public abstract class TransitionPainter :IPainter
    {
        protected Bitmap m_bitBuffer;
        protected Graphics g;

        private float m_TransitionComplete;

        /// <summary>
        /// The value representing how much of the transition is complete.
        /// </summary>
        public virtual float TransitionComplete
        {
            get { return m_TransitionComplete; }
            set { m_TransitionComplete = value; }
        }

        private Bitmap m_OldImage;

        public virtual Bitmap OldImage
        {
            get { return m_OldImage; }
            set
            {
                m_OldImage = value;
                this.ScaledOldImage = null;
            }
        }

        private Bitmap m_NewImage;

        public virtual Bitmap NewImage
        {
            get { return m_NewImage; }
            set
            {
                m_NewImage = value;
                this.ScaledNewImage = null;
            }
        }

        private Bitmap m_ScaledOldImage;

        public Bitmap ScaledOldImage
        {
            get { return m_ScaledOldImage; }
            protected set { m_ScaledOldImage = value; }
        }

        private Bitmap m_ScaledNewImage;

        public Bitmap ScaledNewImage
        {
            get { return m_ScaledNewImage; }
            protected set { m_ScaledNewImage = value; }
        }

        private Rectangle m_InvalidateRegion;
	
        /// <summary>
        /// The rectangle that should be updated. This should be set by the
        /// inheriting classes once the painting is done. The controller will
        /// then update just this region.
        /// </summary>
        //TODO This could probably be removed.
        public Rectangle InvalidRegion
        {
            get { return m_InvalidateRegion; }
            protected set { m_InvalidateRegion = value; }
        }

        public abstract string Name
        {
            get;
        }

        /// <summary>
        /// Resets the <see cref="InvalidateRegion"/>
        /// </summary>
        public void Reset()
        {
            this.InvalidRegion = new Rectangle(-1,-1,1,1);
        }
	
        /// <summary>
        /// Paints the transition between the <see cref="OldImage"/> and <see cref="NewImage"/>
        /// on <see cref="graphics"/>. The transition stage is specified by 
        /// <see cref="TransitionComplete"/>.
        /// </summary>
        public abstract void Paint(object sender, PaintEventArgs args);

        public TransitionPainter(Bitmap oldImage, Bitmap newImage)
        {
            this.OldImage = oldImage;
            this.NewImage = newImage;
        }

        /// <summary>
        /// Calculates the rectangle that the images should be drawn on.
        /// Currently tries to fit the whole page onto the control.
        /// Assumes the images are of equal width.
        /// </summary>
        /// <param name="control">The control that the images are drawn on.</param>
        /// <returns>The calculated rectangle</returns>
        protected Rectangle GetDrawRectangle(Control control)
        {
            float scaleFactor = (float)control.Height / (float)NewImage.Height;
            int scaledWidth = (int)(NewImage.Width * scaleFactor);
            int scaledHeight = (int)(NewImage.Height * scaleFactor);
            Point scaledOrigin = new Point((control.Width - scaledWidth) / 2, 0);

            Rectangle scaledRectangle = new Rectangle(scaledOrigin.X,
                                                      scaledOrigin.Y,
                                                      scaledWidth,
                                                      scaledHeight);
            return scaledRectangle;
        }

        /// <summary>
        /// Initialises the bit and graphics buffer. Sets the scaled image fields.
        /// </summary>
        protected void InitialiseBuffers(Control control)
        {
            this.m_bitBuffer = new Bitmap(control.Size.Width, control.Size.Height);
            g = Graphics.FromImage(this.m_bitBuffer);

            Rectangle scaled = GetDrawRectangle(control);
            
            this.ScaledNewImage = (Bitmap)NewImage.GetThumbnailImage(scaled.Width, scaled.Height, null, System.IntPtr.Zero);
            this.ScaledOldImage = (Bitmap)OldImage.GetThumbnailImage(scaled.Width, scaled.Height, null, System.IntPtr.Zero);
        }
    }
}
