using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using Espresso.Utils;

namespace Espresso.Viewer.View
{
    public class WipeTransitionPainter : AbstractPainter
    {
        public WipeTransitionPainter() : base(null, null) { }

        /// <summary>
        /// If TransitionCompleted is 0, the whole area is painted with the old image.
        /// If TransitionCompleted is 1, the whole area is painted with the new image.
        /// Otherwise just the part of the new image that needs to be drawn is updated 
        /// and <see cref="InvalidateRegion"/> is set to the updated part of the image
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        public override void Paint(object sender, PaintEventArgs args)
        {
            Control control = (Control)sender;
            Rectangle drawRect = GetDrawRectangle(control);

            if (this.m_bitBuffer == null || this.m_bitBuffer.Size != control.Size
               || this.ScaledNewImage == null || this.ScaledOldImage == null)
            {
                InitialiseBuffers(control);
            }

            g.FillRectangle(Brushes.DarkGray, new Rectangle(new Point(0, 0), m_bitBuffer.Size));

            if (this.CompleteRatio == 0)
            {
                //Draw the white background
                g.FillRectangle(Brushes.White, drawRect);

                //Draw the complete old page
                Rectangle oldSrcRect = new Rectangle(0, 0, ScaledOldImage.Width, ScaledOldImage.Height);
                g.DrawImage(ScaledOldImage, drawRect, oldSrcRect, GraphicsUnit.Pixel);

                this.InvalidRegion = drawRect;
            }
            else if (this.CompleteRatio == 1)
            {
                //Draw the white background
                g.FillRectangle(Brushes.White, drawRect);

                //Draw the complete new page
                Rectangle newSrcRect = new Rectangle(0, 0, ScaledNewImage.Width, ScaledNewImage.Height);
                g.DrawImage(ScaledNewImage, drawRect, newSrcRect, GraphicsUnit.Pixel);

                this.InvalidRegion = drawRect;
            }
            else
            {
                double H = drawRect.Height;
                double W = drawRect.Width;
                double X = drawRect.X;
                double Y = drawRect.Y;
                double T = H * CompleteRatio;

                //Need to do this because sometimes scaled images end up being null by the time
                // we get to this point in execution. Need to investigate into this more.
                if (this.m_bitBuffer == null || this.m_bitBuffer.Size != control.Size
                || this.ScaledNewImage == null || this.ScaledOldImage == null)
                {
                    InitialiseBuffers(control);
                }

                Rectangle oldSrcRect = new Rectangle(0, 0, ScaledOldImage.Width, ScaledOldImage.Height);
                Rectangle newSrcRect = new Rectangle(0, 0, ScaledNewImage.Width, (int)T + 1);
                Rectangle newDestRect = new Rectangle((int)X, (int)Y, (int)W, (int)T + 1); 
                
                //Draw the white background
                g.FillRectangle(Brushes.White, drawRect);

                //Draw the complete old page
                g.DrawImage(ScaledOldImage, drawRect, oldSrcRect, GraphicsUnit.Pixel);

                //Draw part of the new image
                g.DrawImage(ScaledNewImage, newDestRect, newSrcRect, GraphicsUnit.Pixel);

                g.DrawLine(new Pen(Brushes.Black, 2),
                           new Point(drawRect.Left, (int)T - 1),
                           new Point(drawRect.Right, (int)T - 1));
                
                //Calculate the region that needs to be invalidated or update the whole
                //screen if it has not been done before.
                if (this.InvalidRegion.X < 0)
                {
                    this.InvalidRegion = newDestRect;
                }
                else
                {
                    Rectangle prevInv = this.InvalidRegion;

                    Rectangle updateRegion = new Rectangle(prevInv.X,
                                                           prevInv.Bottom - 4,
                                                           prevInv.Width,
                                                           newDestRect.Height - prevInv.Bottom + 4);
                    this.InvalidRegion = updateRegion;
                }
            }

            args.Graphics.DrawImage(this.m_bitBuffer, new Point(0, 0));
        }

        public override string Name
        {
            get { return "WipePainter"; }
        }
    }
}
